/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: md5.c                                                       *
 *                                                                     *
 * Description:                                                        *
 * MD5 hash implementation.                                            *
 *                                                                     *
 * Comments:                                                           *
 *  Based on various old code and etc.                                 *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/sched.h>
#include <linux/file.h>
#include <linux/fs.h>

#include <asm/uaccess.h>

#include "vmmain.h"
#include "logging.h"
#include "md5.h"

static const unsigned char fillbuf[64] = { 0x80, 0 /* .. */};

static int md5_init_ctx(struct md5_ctx *ctx) {

    ctx->A = 0x67452301;
    ctx->B = 0xefcdab89;
    ctx->C = 0x98badcfe;
    ctx->D = 0x10325476;

    ctx->total[0] = ctx->total[1] = 0;
    ctx->buflen = 0;

    return 0;
}

static void *md5_read_ctx(struct md5_ctx *ctx, void *resbuf) {

    ((u_long *) resbuf)[0] = ctx->A;
    ((u_long *) resbuf)[1] = ctx->B;
    ((u_long *) resbuf)[2] = ctx->C;
    ((u_long *) resbuf)[3] = ctx->D;

    return resbuf;
}

static void *md5_finish_ctx (struct md5_ctx *ctx, void *resbuf) {

    u_long bytes = ctx->buflen;
    size_t pad;

    ctx->total[0] += bytes;
    if (ctx->total[0] < bytes) ++ctx->total[1];

    pad = bytes >= 56 ? 64 + 56 - bytes : 56 - bytes;
    memcpy(&ctx->buffer[bytes], fillbuf, pad);

    *(u_long *) &ctx->buffer[bytes + pad] = ctx->total[0] << 3;
    *(u_long *) &ctx->buffer[bytes + pad + 4] = (ctx->total[1] << 3) | (ctx->total[0] >> 29);

    md5_process_block (ctx->buffer, bytes + pad + 8, ctx);

    return md5_read_ctx (ctx, resbuf);
}

int QUICKCALL md5_stream(struct file *fp, void *resblock) {

    struct md5_ctx ctx;
    char *buffer;
    int sum, n;
    loff_t pos = 0;

    md5_init_ctx (&ctx);
    
    buffer = kmalloc(BLOCKSIZE + 72, GFP_KERNEL);
    if (!buffer)
        return -ENOMEM;

    while (1) {
        sum = 0;

        memset(buffer, 0, BLOCKSIZE + 72);
        while (1) {
            if ((n = fp->f_op->read(fp, (void *)buffer + sum, BLOCKSIZE - sum, &pos)) < 0)
                goto err;
           
            sum += n;

            if (sum == BLOCKSIZE) break;
            if (!n) goto eof;
        }
        md5_process_block (buffer, BLOCKSIZE, &ctx);
    }

eof:
    if (sum > 0) md5_process_bytes (buffer, sum, &ctx);

    md5_finish_ctx (&ctx, resblock);
    kfree(buffer);
    return 0;
    
err:
    kfree(buffer);
    return -EFAULT;
}

void* QUICKCALL md5_buffer(char *buffer, size_t len, void *resblock) {

    struct md5_ctx ctx;

    md5_init_ctx (&ctx);

    md5_process_bytes (buffer, len, &ctx);

    return md5_finish_ctx(&ctx, resblock);
}


int QUICKCALL md5_process_bytes(void *buffer, size_t len, struct md5_ctx *ctx) {

    size_t left, add;

    if (ctx->buflen != 0) {
        left = ctx->buflen;
        add = 128 - left > len ? len : 128 - left;

        memcpy (&ctx->buffer[left], buffer, add);
        ctx->buflen += add;

        if (ctx->buflen > 64) {
            md5_process_block (ctx->buffer, ctx->buflen & ~63, ctx);

            ctx->buflen &= 63;
            memcpy (ctx->buffer, &ctx->buffer[(left + add) & ~63], ctx->buflen);
        }

        buffer = (char *) buffer + add;
        len -= add;
    }

    if (len >= 64) {
        md5_process_block (buffer, len & ~63, ctx);
        buffer = (char *) buffer + (len & ~63);
        len &= 63;
    }

    if (len > 0) {
        left = ctx->buflen;

        memcpy (&ctx->buffer[left], buffer, len);
        left += len;
        if (left >= 64) {
            md5_process_block (ctx->buffer, 64, ctx);
            left -= 64;
            memcpy (ctx->buffer, &ctx->buffer[64], left);
        }
        ctx->buflen = left;
    }

    return 0;
}


int QUICKCALL md5_process_block(void *buffer, size_t len, struct md5_ctx *ctx) {

    u_long correct_words[16];
    const u_long *words = buffer;
    size_t nwords = len / sizeof (u_long);
    const u_long *endp = words + nwords;
    u_long A = ctx->A;
    u_long B = ctx->B;
    u_long C = ctx->C;
    u_long D = ctx->D;

    ctx->total[0] += len;
    if (ctx->total[0] < len) ++ctx->total[1];

    while (words < endp) {
        u_long *cwp = correct_words;
        u_long A_save = A;
        u_long B_save = B;
        u_long C_save = C;
        u_long D_save = D;

#define OP(a, b, c, d, s, T) \
    do { \
      a += FF (b, c, d) + (*cwp++ = *words) + T; \
      ++words; \
      a = rol (a, s); \
      a += b; \
    } while (0)

        /* Round 1.  */
        OP(A, B, C, D,  7, 0xd76aa478);
        OP(D, A, B, C, 12, 0xe8c7b756);
        OP(C, D, A, B, 17, 0x242070db);
        OP(B, C, D, A, 22, 0xc1bdceee);
        OP(A, B, C, D,  7, 0xf57c0faf);
        OP(D, A, B, C, 12, 0x4787c62a);
        OP(C, D, A, B, 17, 0xa8304613);
        OP(B, C, D, A, 22, 0xfd469501);
        OP(A, B, C, D,  7, 0x698098d8);
        OP(D, A, B, C, 12, 0x8b44f7af);
        OP(C, D, A, B, 17, 0xffff5bb1);
        OP(B, C, D, A, 22, 0x895cd7be);
        OP(A, B, C, D,  7, 0x6b901122);
        OP(D, A, B, C, 12, 0xfd987193);
        OP(C, D, A, B, 17, 0xa679438e);
        OP(B, C, D, A, 22, 0x49b40821);

#undef OP
#define OP(f, a, b, c, d, k, s, T) \
    do { \
      a += f (b, c, d) + correct_words[k] + T; \
      a = rol (a, s); \
      a += b; \
    } while (0)

        /* Round 2.  */
        OP(FG, A, B, C, D,  1,  5, 0xf61e2562);
        OP(FG, D, A, B, C,  6,  9, 0xc040b340);
        OP(FG, C, D, A, B, 11, 14, 0x265e5a51);
        OP(FG, B, C, D, A,  0, 20, 0xe9b6c7aa);
        OP(FG, A, B, C, D,  5,  5, 0xd62f105d);
        OP(FG, D, A, B, C, 10,  9, 0x02441453);
        OP(FG, C, D, A, B, 15, 14, 0xd8a1e681);
        OP(FG, B, C, D, A,  4, 20, 0xe7d3fbc8);
        OP(FG, A, B, C, D,  9,  5, 0x21e1cde6);
        OP(FG, D, A, B, C, 14,  9, 0xc33707d6);
        OP(FG, C, D, A, B,  3, 14, 0xf4d50d87);
        OP(FG, B, C, D, A,  8, 20, 0x455a14ed);
        OP(FG, A, B, C, D, 13,  5, 0xa9e3e905);
        OP(FG, D, A, B, C,  2,  9, 0xfcefa3f8);
        OP(FG, C, D, A, B,  7, 14, 0x676f02d9);
        OP(FG, B, C, D, A, 12, 20, 0x8d2a4c8a);

        /* Round 3.  */
        OP(FH, A, B, C, D,  5,  4, 0xfffa3942);
        OP(FH, D, A, B, C,  8, 11, 0x8771f681);
        OP(FH, C, D, A, B, 11, 16, 0x6d9d6122);
        OP(FH, B, C, D, A, 14, 23, 0xfde5380c);
        OP(FH, A, B, C, D,  1,  4, 0xa4beea44);
        OP(FH, D, A, B, C,  4, 11, 0x4bdecfa9);
        OP(FH, C, D, A, B,  7, 16, 0xf6bb4b60);
        OP(FH, B, C, D, A, 10, 23, 0xbebfbc70);
        OP(FH, A, B, C, D, 13,  4, 0x289b7ec6);
        OP(FH, D, A, B, C,  0, 11, 0xeaa127fa);
        OP(FH, C, D, A, B,  3, 16, 0xd4ef3085);
        OP(FH, B, C, D, A,  6, 23, 0x04881d05);
        OP(FH, A, B, C, D,  9,  4, 0xd9d4d039);
        OP(FH, D, A, B, C, 12, 11, 0xe6db99e5);
        OP(FH, C, D, A, B, 15, 16, 0x1fa27cf8);
        OP(FH, B, C, D, A,  2, 23, 0xc4ac5665);

        /* Round 4.  */
        OP(FI, A, B, C, D,  0,  6, 0xf4292244);
        OP(FI, D, A, B, C,  7, 10, 0x432aff97);
        OP(FI, C, D, A, B, 14, 15, 0xab9423a7);
        OP(FI, B, C, D, A,  5, 21, 0xfc93a039);
        OP(FI, A, B, C, D, 12,  6, 0x655b59c3);
        OP(FI, D, A, B, C,  3, 10, 0x8f0ccc92);
        OP(FI, C, D, A, B, 10, 15, 0xffeff47d);
        OP(FI, B, C, D, A,  1, 21, 0x85845dd1);
        OP(FI, A, B, C, D,  8,  6, 0x6fa87e4f);
        OP(FI, D, A, B, C, 15, 10, 0xfe2ce6e0);
        OP(FI, C, D, A, B,  6, 15, 0xa3014314);
        OP(FI, B, C, D, A, 13, 21, 0x4e0811a1);
        OP(FI, A, B, C, D,  4,  6, 0xf7537e82);
        OP(FI, D, A, B, C, 11, 10, 0xbd3af235);
        OP(FI, C, D, A, B,  2, 15, 0x2ad7d2bb);
        OP(FI, B, C, D, A,  9, 21, 0xeb86d391);

        A += A_save;
        B += B_save;
        C += C_save;
        D += D_save;
    }

    ctx->A = A;
    ctx->B = B;
    ctx->C = C;
    ctx->D = D;

    return 0;
}

