/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: timer.c                                                     *
 *                                                                     *
 * Description:                                                        *
 *  Internal RTC device based timer.                                   *
 *                                                                     *
 * Comments:                                                           *
 *  Documentation can be found at Models/HiResTimer.txt                *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/mc146818rtc.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/mm.h>

#include <asm/uaccess.h>

#include "logging.h"
#include "compat_procs.h"
#include "timer.h"

/* hires timer pollers */
static spinlock_t vmPollLock = SPIN_LOCK_UNLOCKED;
static LIST_HEAD(vmPollList);

/* current clock rate */
static unsigned vmTimerRate;

/* thread stuff  */
static compat_thread_id vmTimerThreadId = 0;
COMPAT_COMPLETION(vmTimerThreadCompl);


/* public ioctl() functions */
void vmTimerPoll(struct vmTimer *timer, unsigned countdown) {

    unsigned long flags;

    spin_lock_irqsave(&vmPollLock, flags);
    /* in case timer is rescheduled */
    list_del(&timer->list);
    INIT_LIST_HEAD(&timer->list);
    timer->countdown = countdown;
    list_add(&timer->list, &vmPollList);
    spin_unlock_irqrestore(&vmPollLock, flags);
}

void vmTimerKill(struct vmTimer *timer) {

    unsigned long flags;

    spin_lock_irqsave(&vmPollLock, flags);
    list_del(&timer->list);
    INIT_LIST_HEAD(&timer->list);
    spin_unlock_irqrestore(&vmPollLock, flags);
}


static int vmTimerThread(void *data) {

    struct file *fp = (struct file *)data;
    unsigned long buff;
    unsigned rate = 0;
    int ret;
    loff_t ppos;
    mm_segment_t rfs;
    unsigned long flags;
    struct list_head *l;

    compat_daemonize("vm-rtc");
    
    rfs = get_fs();
    set_fs(KERNEL_DS); /* allow rtc driver to do put_user */
    cap_raise(current->cap_effective, CAP_SYS_RESOURCE); /* allow high resolution interrupts */

    while (vmTimerRate > HZ) {

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 6)
        if (kthread_should_stop()) break;
#endif
        if (signal_pending(current)) break;

        /* change lock rate */
        if (rate != vmTimerRate) {
            rate = 64;
            while (rate < vmTimerRate && rate < 8192) {
                rate <<= 1;
            }

            if ((ret = fp->f_op->ioctl(fp->f_dentry->d_inode, fp, RTC_IRQP_SET, rate)) < 0) {
                Err("Can't set clock rate: %d", rate);
                goto err;
            }
        }

        /* sleep for IRQ */
        ppos = 0;
        ret = fp->f_op->read(fp, (void *)&buff, sizeof(buff), &ppos);
        if (ret < 0) {
            if (ret != -ERESTARTSYS) { /* return if sleep interrupted by signal */
                Err("Can't read from /dev/rtc: %d", ret);
                goto err;
            }
        }

        /* wake pollers */
        spin_lock_irqsave(&vmPollLock, flags);
        for (l = vmPollList.next; l != &vmPollList; ) {
            struct vmTimer *timer = list_entry(l, struct vmTimer, list);

            l = l->next;
            if (!timer->countdown || !--timer->countdown) {
                list_del(&timer->list);
                INIT_LIST_HEAD(&timer->list);
                wake_up_interruptible(&timer->wq);
            }
        }
        spin_unlock_irqrestore(&vmPollLock, flags);
    }

err:
    set_fs(rfs);
    filp_close(fp, current->files);
    compat_exit(vmTimerThreadCompl, 0);

    return 0; /* not reached */
}

/* timer setup */
int vmSetTimer(unsigned rate) {

    struct file *fp;
    int ret, cr;

    if (rate <= HZ) {
        /* yes, we do nothing. your poll() timeout will wake you up
        Warn("Too small rate %d, use system clock instead.", rate); */

        /* don't report error, it's looks very very ugly while rate && HZ = 1000
         * -- ms
         */
        return ((rate < HZ) ? -EINVAL : 0);
    }

    /* XXX need synchronization here */
    if (vmTimerThreadId)
        return 0;

    fp = filp_open("/dev/rtc", O_RDONLY, 0);
    if (IS_ERR(fp)) {
        Err("Can't open /dev/rtc: %d", PTR_ERR(fp));
        return -ENXIO;
    }

    /*
     * We are need to setup cup because /dev/rtc may be
     * alredy have interrupt rate > 64HZ.
     */
    if (!(cr = cap_raised(current->cap_effective, CAP_SYS_RESOURCE))) {
        cap_raise(current->cap_effective, CAP_SYS_RESOURCE);
    }
            
    if ((ret = fp->f_op->ioctl(fp->f_dentry->d_inode, fp, RTC_PIE_ON, 0)) < 0) {
        Err("Can't enable interrupts via /dev/rtc: %d\n", ret);
        filp_close(fp, current->files);
        return ret;
    }

    if (!cr) cap_lower(current->cap_effective, CAP_SYS_RESOURCE);

    vmTimerRate = rate;
    compat_start_thread(vmTimerThreadId, vmTimerThread, fp, "vm-rtc");

    if (!vmTimerThreadId) {
        Err("Can't start timer thread.");
        return -ENXIO;
    }

    return 0;
}

/* stop timer */
void vmStopTimer(void) {

    if (vmTimerThreadId) {
        compat_stop_thread(vmTimerThreadId);
        compat_wait_for_completion(vmTimerThreadCompl);
        vmTimerThreadId = 0;
    }
}
