/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: compat_procs.h                                              *
 *                                                                     *
 * Description:                                                        *
 *  Compatibility defines for work with processes from kernel.         *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#ifndef __COMPAT_PROCS_H__
#define __COMPAT_PROCS_H__

#include <linux/version.h>

/* kthread header */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 6)
# include <linux/kthread.h>
#endif

/* thread id */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 6)
# define compat_thread_id struct task_struct *
#else
# define compat_thread_id int
#endif

/* completion */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 9)
# define COMPAT_COMPLETION(x)   static DECLARE_COMPLETION(x);
#else
# define COMPAT_COMPLETION(x)   static struct semaphore x;
#endif

/* isAlive, test for alive pid */    
# if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 10)
static inline int compat_pid_alive(struct task_struct *p) {
# ifdef EXIT_DEAD
    return (p->state == EXIT_DEAD) ? 0 : 1;
# else
    return (p->state >= TASK_ZOMBIE) ? 0 : 1;
# endif
}
#else
# define compat_pid_alive(x) (pid_alive(x))
#endif

/* start thread */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 6)
# ifdef FREEZ_KTHREADS_PATCH
#  define compat_start_thread(id, func, data, name) {    \
     id = kthread_create(func, data, 0, "%s", name);           \
     if (!IS_ERR(id)) { wake_up_process(id); } else { id = 0; } }
# else
#  define compat_start_thread(id, func, data, name) {    \
     id = kthread_create(func, data, "%s", name);              \
     if (!IS_ERR(id)) { wake_up_process(id); } else { id = 0; } }
# endif /* FREEZ_KTHREADS_PATCH */
#else
# define compat_start_thread(id, func, data, name) {    \
    id = kernel_thread(func, data, 0); }
#endif
    
/* stop thread */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 6)
# define compat_stop_thread(x) {    \
    if (!IS_ERR(x)) {              \
        kthread_stop(x);           \
    }}
#else
# define compat_stop_thread(x)   kill_proc(x, SIGKILL, 1)
#endif

/* wait for completion */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 6)
# define compat_wait_for_completion(x) do {} while(0)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 9)
# define compat_wait_for_completion(x) wait_for_completion(&x);
#else
# define compat_wait_for_completion(x) {    \
    down(&x);                               \
    lock_kernel();                          \
    unlock_kernel(); }
#endif

/* daemonize */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 61)
# define compat_daemonize(x) { \
    daemonize();                \
    reparent_to_init(); }
#elif LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 6)
# define compat_daemonize(x) do {} while(0)
#else
# define compat_daemonize(x) { \
    daemonize("%s", x); }
#endif

/* exit */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 6)
# define compat_exit(compl, code)  return code;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 9)
# define compat_exit(compl, code) complete_and_exit(&compl, code);
#else
# define compat_exit(compl, code) {    \
    lock_kernel();                      \
    up(&compl);                         \
    compat_exit(code); }
#endif

/* need resched */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 3)
# define compat_need_resched() (current->need_resched)
#else
# define compat_need_resched() (need_resched())
#endif

#endif /* __COMPAT_PROCS_H__*/
