/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: device.c                                                    *
 *                                                                     *
 * Description:                                                        *
 *  /dev/vm-main generic operations                                    *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/miscdevice.h>
#include <linux/poll.h>

#include "vmmain.h"
#include "timer.h"
#include "intgate.h"
#include "logging.h"
#include "utils.h"

/* extern increment/decrement functions */
extern int appIncCounter(void);
extern void appDecCounter(void);


/* device stuff */
#define DEVICE_NAME "vm-main"

#ifndef MISC_DYNAMIC_MINOR
# define MISC_DYNAMIC_MINOR 250
#endif
#define VM_MAIN_MINOR MISC_DYNAMIC_MINOR

static struct miscdevice vmMainDev;

/* device status structure */
static struct vmDeviceState {
    int OpenCount;
} vmDeviceState;

/* basic file operations */
extern int vmIoctl(struct inode *, struct file *, unsigned int, unsigned long);

unsigned int vmPoll(struct file *, poll_table *);
int vmOpen(struct inode *, struct file *);
int vmClose(struct inode *, struct file *);

/* fops structure for /dev/vm-main */
static struct file_operations vm_fops = {
    owner:              THIS_MODULE,
    ioctl:              vmIoctl,
    poll:               vmPoll,
    open:               vmOpen,
    release:            vmClose,
};

/*
 * open() function
 */
int vmOpen(struct inode *inode, struct file *filp) {

    struct module_param_t* param;

#ifdef ENABLE_MODULES_PROTECT
    if (check_called_proc()) {
        return -EACCES;
    }
#endif

    /* try to increment counter */
    if (appIncCounter())
        return -EPERM;
    
    if (!(param = iDataOpen()))
        return -ENOMEM;

    filp->private_data = (void *)param;

    vmDeviceState.OpenCount++;

    vmGlobal.vmWset.wsCount++;

    return 0;
}

/*
 * close() function
 */
int vmClose(struct inode *inode, struct file *filp) {
    
    struct module_param_t *param = (struct module_param_t *)filp->private_data;
    
    if (param->MonAddr) {
        Warn("Application process unexpectly died");
        /* close idata struct */
        iDataClose(param);
    }

    /* reorder wset */
    vmGlobal.vmWset.wsCount--;
    calcWset();
    
    /* decrement counter in hypervisor */
    appDecCounter();
        
    /* check for locked pages for current task */
    vm_mmFreeCurrent(param->iData.bma);
    param->iData.bma = NULL;

    iDataFree(param);

    filp->private_data = NULL;

    vmDeviceState.OpenCount--;

    /* If driver not open, delete gates for all processors */
    if (!vmDeviceState.OpenCount)
        ig_CloseGates();
#if 0
    if (!vmDeviceState.OpenCount) /* kill timer thread */
        vmStopTimer();
#endif

    return 0;
}

/*
 * poll() function
 */
unsigned int vmPoll(struct file *filp, poll_table *wait) {

    struct module_param_t *param = (struct module_param_t *)filp->private_data;

    poll_wait(filp, &param->timer.wq, wait);
    
    if (param->timer.countdown == 0)
        return POLLIN | POLLRDNORM;

    return 0;
}


/* device init/deinit functions */
int vmDeviceInit(void) {

    char *devname = DEVICE_NAME;

    /* init device */
    memset(&vmMainDev, 0, sizeof(struct miscdevice));
    vmMainDev.minor = VM_MAIN_MINOR;
    vmMainDev.name = devname;
    vmMainDev.fops = &vm_fops;

    /* register it */
    if (misc_register(&vmMainDev)) {
        Err("vm-main failed to register /dev/%s device", DEVICE_NAME);
        return -ENOENT;
    }

    Log("Device /dev/%s registered with %d.%d", DEVICE_NAME, 10, vmMainDev.minor);
    vmDeviceState.OpenCount = 0;

    return 0;
}

void vmDeviceDeinit(void) {
    /* deregister device */
    misc_deregister(&vmMainDev);
}
