/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: intgate.c                                                   *
 *                                                                     *
 * Description:                                                        *
 *  Interrupt gate for direct calls to monitor.                        *
 *                                                                     *
 * Comments:                                                           *
 *  Based on IntGate from Diman@                                       *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/spinlock.h>
#include <asm/errno.h>
#include <linux/sched.h>

#include "hypercall.h"
#include "AtomicOps.h"
#include "intgate.h"
#include "logging.h"
#include "compat_smp.h"

/* far jump struct (ljmp) */
struct farjmp_t {
    unsigned int offset;
    unsigned short sel;
} PACKED;

/* gate number */
static unsigned int SetExc = 0;

/* is gate magic initialized ? */
static volatile int gateWasInit = 0;

/* global gate magic number */
static gate_magic_num_descr_t gateMagic;

/* counter for registered gate functions */
static volatile unsigned int monReqCount = 0;

/* array of registered function */
static unsigned int monitorReqFuncs[SIMUL_VM_MAX_COUNT] = {0};

/* gate functions for every cpu */
static struct {
    struct idt_descriptor_t *PrimaryIdt;
    struct idt_descriptor_t  SaveDesc;
} ProcInfo[MAX_PROCESSOR_NUMBER];

/* spinlock for generator */
static spinlock_t genLock = SPIN_LOCK_UNLOCKED;

static struct farjmp_t redirectPtr;

/* Find free gate */
int ig_FindFreeGate(void) {

    struct idt_descriptor_t *PrimaryIdt = ig_GetIdtPtr();
    int i;

    /* check all gates int IDT*/
    for (i = 0; i < 256; i++) {
        /* check for nonexistence */
        if (!PrimaryIdt[i].wSelector || !PrimaryIdt[i].cType || !PrimaryIdt[i].cPresent) {
            return i;
        }
    }

    return -1;
}

/* 
 * Interrupt handler 
 * 
 * Gate use fake function for avoid to use prologue/epilogue
 * because they are can't be disabled on x86 arhc, we are use this
 * little hack with global mark inside of fake function. 
 *                                                      -- ms
 */
void ig_GateHandler(void);
void fake_gate_handler(void) { 
    
    __asm__ __volatile__ (
        ".globl ig_GateHandler\n\t"
        "ig_GateHandler:\n\t"
        "cmpl %6, %%ebp\n\t"
        "jne  _v_konec\n\t"
        "cmpl %%eax, %0\n\t"
        "jne    _invalid_gate_call\n\t"
        "cmpl %%ebx, %1\n\t"
        "jne    _invalid_gate_call\n\t"
        "cmpl %%ecx, %2\n\t"
        "jne    _invalid_gate_call\n\t"
        "cmpl %%edx, %3\n\t"
        "jne    _invalid_gate_call\n\t"
        "\n\t"
        "cmpl %4, %%edi\n\t"
        "jae    _invalid_gate_call\n\t"
        "\n\t"
        "add $3, (%%esp)\n\t"
        "jmp *monitorReqFuncs(, %%edi, 4)\n\t" /* this made for SuSe gcc 3.3.3 --ms */
        "test %5, %%eax\n\t"
        "\n\t"
        "_invalid_gate_call:\n\t"
        "iret\n\t"
        "_v_konec:\n\t"
        "jmpl *%%ss:%7\n\t"
        : :
         "m" (gateMagic.regMagic.eax), 
         "m" (gateMagic.regMagic.ebx), 
         "m" (gateMagic.regMagic.ecx), 
         "m" (gateMagic.regMagic.edx),
         "m" (monReqCount),
         "m" (*monitorReqFuncs),
         "nN" (GATE_STATIC_ID),
         "m" (redirectPtr)
       : "%eax", "%ebx", "%ecx", "%edx", "%edi", "%esi", "memory"
    );
}

/* get magic number for gate */
void ig_GetGateMagic(gate_magic_num_descr_t *saveGateMagic) {
   
    int i;
    
    if (!gateWasInit) {
        spin_lock(&genLock);
        if (!gateWasInit) {
            for (i = 0; i < GATE_MAGIC_NUMBER_LEN; i++) {
                __asm__ __volatile__ (
                    "hlt\t\n"
                    "rdtsc\t\n"
                    "movb %%al, %0\t\n"
                    : "=r" (gateMagic.magic[i])
                    : : "%eax", "%edx"
                );
            }
            gateWasInit = 1;
        }
        spin_unlock(&genLock);
    }
    
    *saveGateMagic = gateMagic;
}

/* register monitor request func */
unsigned int ig_RegisterMonFunc(unsigned int funct) {

    int id;
    
    /* no empty records */
    if (monReqCount >= SIMUL_VM_MAX_COUNT)
        return GATE_INV_MON_ID;
    
    /* find empty record */
    for (id = monReqCount - 1; id >= 0; id--) {
        if (AtomicCompareSwap(&monitorReqFuncs[id], GATE_INV_REQ_FUNC, funct) == GATE_INV_REQ_FUNC) {
            /* found */
            return id;
        }
    }

    /* add record */
    id = AtomicInc((unsigned int*)&monReqCount);
    monitorReqFuncs[id] = funct;
    
    return id;
}

/* free monitor id */
void ig_FreeMonId(unsigned int id) {

    if (id < monReqCount)
        monitorReqFuncs[id] = GATE_INV_REQ_FUNC;
}

/* open gate */
void ig_OpenCurrGate_ipi(void *arg) {

    unsigned int uNumExc = *((unsigned int *)arg);
    struct idt_descriptor_t *PrimaryIdt = NULL;
    static struct idt_descriptor_t gateDescr;
    int cpu = smp_processor_id();

    if (!SetExc)
        SetExc = uNumExc;

    if (!(PrimaryIdt = ig_GetIdtPtr()))
        return;

    redirectPtr.sel = PrimaryIdt[SetExc].wSelector;
    redirectPtr.offset = ((PrimaryIdt[SetExc].wHighOffset <<  16) | (PrimaryIdt[SetExc].wLowOffset));
    
    /* save old */
    ProcInfo[cpu].PrimaryIdt = PrimaryIdt;
    ProcInfo[cpu].SaveDesc = PrimaryIdt[SetExc];

    /* create new descriptor */
    gateDescr.wSelector = ig_GetCsSelector();
    gateDescr.wHighOffset = ((unsigned int)ig_GateHandler) >> 16;
    gateDescr.wLowOffset = ((unsigned int)ig_GateHandler);
    gateDescr.cReserved = 0;
    gateDescr.cType = 0xe;
    gateDescr.cDpl = 0;
    gateDescr.cPresent = 1;

    /* set new handler */
    hypAssignInterrupt(SetExc, cpu, &gateDescr);
    
}


/* open gates */
int ig_OpenGates(unsigned int uNumExc) {

    return compat_smp_ipi_call(ig_OpenCurrGate_ipi, (void *)&uNumExc, 1, 1);
}

/* close gate */
void ig_CloseCurrGate_ipi(void UNUSED *arg) {

    int cpu = smp_processor_id();

    if (!SetExc)
        return;

    if (!ProcInfo[cpu].PrimaryIdt)
        return;

    /* set old handler */
    hypAssignInterrupt(SetExc, cpu, &ProcInfo[cpu].SaveDesc);

    ProcInfo[cpu].PrimaryIdt = NULL;
}

/* close all gates */ 
int ig_CloseGates(void) {

    compat_smp_ipi_call(ig_CloseCurrGate_ipi, NULL, 1, 1);

    SetExc = 0;

    return 0;
}

/* init IntGate */
void ig_InitIntGate(void) {

    int i;

    for (i = 0; i < MAX_PROCESSOR_NUMBER; i++) {
        ProcInfo[i].PrimaryIdt = NULL;
    }
}
