/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: intgate.h                                                   *
 *                                                                     *
 * Description:                                                        *
 *  Interrupt gate for direct calls to monitor.                        *
 *                                                                     *
 * Comments:                                                           *
 *  Based on IntGate from Diman@                                       *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#ifndef __INTGATE_H__
#define __INTGATE_H__

#ifndef PACKED
# define PACKED __attribute__((packed))
#endif

/* processors count */
#define MAX_PROCESSOR_NUMBER 32

/* defailt num */
#define DEFAULT_GATE_EXCEPTION 5 

#define GATE_STATIC_ID 0x90909090

/* max number of monitors working simultaniously */
#define SIMUL_VM_MAX_COUNT 256

/* total interrupts in system */
#define INTERRUPTS_COUNT 256

/* invalid monitor identifier */
#define GATE_INV_MON_ID (unsigned int) -1

/* invalid gate function */
#define GATE_INV_REQ_FUNC (unsigned int) -1

/* call gate macro */
#define CALL_FROM_GATE(except, gateInfo, userMsg)       \
        __asm__ __volatile__ (                          \
	    "pushal\n\t"				\
            "xorl %%ebp, %%ebp\n\t"                     \
            "pushl %%ebp\n\t"                           \
            "pushl %%ebp\n\t"                           \
            "movl %0, %%ebp\n\t"                        \
            "boundl %%ebp, (%%esp)\n\t"                 \
            "addl $8, %%esp\n\t"                        \
            "popal\n\t"                                 \
             : :                                        \
	      "nN" (GATE_STATIC_ID),			\
              "S" (userMsg),                            \
              "a" (gateInfo.magicNum.regMagic.eax),     \
              "b" (gateInfo.magicNum.regMagic.ebx),     \
              "c" (gateInfo.magicNum.regMagic.ecx),     \
              "d" (gateInfo.magicNum.regMagic.edx),     \
              "D" (gateInfo.directCall))
             
        

/* IDT descriptor description */
struct idt_descriptor_t {
    unsigned short wLowOffset;
    unsigned short wSelector;
    unsigned char cReserved;
    unsigned char cType: 5;
    unsigned char cDpl: 2;
    unsigned char cPresent: 1;
    unsigned short wHighOffset;
} PACKED;

/* Far call description struct */
struct far_call_ptr_t {
    unsigned int uOffset;
    unsigned short wSelector;
} PACKED;

/* IDT save/load struct */
struct idt_info_t {
    unsigned short wLimit;
    struct idt_descriptor_t *pBase;
} PACKED;

/* magic gate descriptor */
#define GATE_MAGIC_NUMBER_LEN 16

typedef union gate_magic_num_descr_t { 
    unsigned char magic[GATE_MAGIC_NUMBER_LEN];

    struct {
        unsigned int eax;
        unsigned int ebx;
        unsigned int ecx;
        unsigned int edx;
    } regMagic;
} gate_magic_num_descr_t;

/* gate reg structure */
struct intgate_reg_t {
    /* magic number for intgate */
    gate_magic_num_descr_t magicNum;

    /* direct call function pointer */
    unsigned int directCall;
};

/* get CS selector */
static inline unsigned short ig_GetCsSelector(void) {

    unsigned short wRetVal;

    __asm__ __volatile__ (
        "mov %%cs,%0"
        : "=r"(wRetVal)
    );
    return wRetVal;
}

/* get pointer to IDT */
static inline struct idt_descriptor_t *ig_GetIdtPtr(void) {

    struct idt_info_t IdtInfo;

    __asm__ __volatile__ (
        "sidt %0\n"
        : : "m"(IdtInfo)
        : "memory"
    );
    
    return IdtInfo.pBase;
}

/* get gate type*/
static inline unsigned char ig_GetGateType(unsigned int uNumExc) {

    struct idt_info_t IdtInfo;

    __asm__ __volatile__ (
        "sidt %0\n"
        : : "m"(IdtInfo)
        : "memory"
    );

    return IdtInfo.pBase[uNumExc].cType;
}

int ig_FindFreeGate(void);
int ig_OpenGates(unsigned int);
int ig_CloseGates(void);
void ig_InitIntGate(void);

void ig_FreeMonId(unsigned int);
void ig_GetGateMagic(gate_magic_num_descr_t *);
unsigned int ig_RegisterMonFunc(unsigned int);

#ifdef __IN_GATE__

#endif /* __IN_GATE__ */

#endif /* __INTGATE_H__ */
