/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: ioctls.c                                                    *
 *                                                                     *
 * Description:                                                        *
 *  ioctls requests handler.                                           *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/poll.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>

#include "logging.h"
#include "vmmain.h"
#include "ioctls.h"
#include "timer.h"
#include "intgate.h"
#include "vtx_init.h"
#include "svm_init.h"


extern unsigned int vmPoll(struct file *, poll_table *);
int vmMainIoctl(struct module_param_t *, unsigned, struct ioctl_request_t *);

/*
 * ioctl() function
 */
int vmIoctl(struct inode *inode, struct file *filp, unsigned int iocmd, unsigned long ioarg) {

    int retVal = -EINVAL;
    struct module_param_t *param = (struct module_param_t *)filp->private_data;
    struct ioctl_request_t request;
    unsigned timer;
    void *arg = (void *)ioarg;
    char __attribute__((unused)) sbuff[MAX_SALT_SIZE];

    if (!param || !arg)
        return retVal;

#ifdef ENABLE_MODULES_PROTECT
    if (iocmd != VM_CHECK_RANDOM_SALT && !param->saltPassed)
        return retVal;
#endif
        
    /* get ioctl() command */
    switch(iocmd) {
    case VM_SET_TIMER_RATE:
        if (!(retVal = get_user(timer, (unsigned *)arg))) {
            retVal = vmSetTimer(timer);
        }
        break;
    case VM_SET_TIMER_ALARM:
        if (!(retVal = get_user(timer, (unsigned *)arg))) {
            vmTimerPoll(&param->timer, timer);
        }
        break;
    case VM_MAIN_REQUEST:
        if (!(retVal = copy_from_user(&request, arg, sizeof(struct ioctl_request_t)))) {
            retVal = vmMainIoctl(param, request.CtlCode, &request);
            if (copy_to_user(arg, &request, sizeof(struct ioctl_request_t)))
                retVal = -EFAULT;
        }
        break;
    case VM_CHECK_RANDOM_SALT:
#ifdef ENABLE_MODULES_PROTECT
        if (!copy_from_user(&request, arg, sizeof(struct ioctl_request_t))) {
            if (!copy_from_user(sbuff, request.Buffer, ((request.InLength >= MAX_SALT_SIZE) ? MAX_SALT_SIZE - 1 : request.InLength))) {
                if (strlen(vmGlobal.salt) == request.InLength) {
                    if (!memcmp(vmGlobal.salt, sbuff, request.InLength)) {
                        retVal = 0;
                        param->saltPassed = 1;
                    }
                }
            }
        }
#else
        retVal = 0;
        param->saltPassed = 1;
#endif
        break;
    default:
        Err("vmIoctl: incorrect ioctl code (%p)", iocmd);
        break;
    }

    return retVal;
}

/*
 * VM_MAIN_REQUEST ioctl's command handler
 */
int vmMainIoctl(struct module_param_t *param, unsigned cmd, struct ioctl_request_t *request) {

    long ret = STATUS_UNSUCCESSFUL;
    void *arg = (void *)request->Buffer;
    int i;

    /* copy from/to variables */
    MEMORY_INFO memInfo;
    struct monitor_functions_def_t mFunc[MONFUNC_COUNT];
    struct pmm_initial_values_t pmmInitial;
    unsigned int uintVal;
	unsigned int uiInfoBuff[2];
    void *monArg = NULL;
    struct intgate_reg_t gateReg; 

    if (cmd > END_IOCTL || cmd < START_IOCTL)
        return -EINVAL; /* incorrect ioctl code */
    
    switch(cmd) {
    case IOCTL_PMM_LOCK_PAGE: /* lock pages region */
        if (param->iData.PmmData) {
            struct LOCK_PAGE_RESULT   LockPageResult;

            if( ( request->InLength < sizeof(struct LOCK_PAGE_RESULT) ) || 
		( request->OutLength < sizeof(struct LOCK_PAGE_RESULT) )
		)
            {
                break;
            }

            if (copy_from_user(&LockPageResult, arg, sizeof(struct LOCK_PAGE_RESULT)))
               break;

            if(LockPageResult.uRequiredSize != -1)
            {
		// error during previouse allocation
            }
            if (param->iData.PmmData->MonitorPmmBuff) {
                LockPageResult.uResult = mmLockPmmRegion(param->iData.PmmData,
                                                 param->iData.PmmData->MonitorPmmBuff->ApiMem,
                                                 param,
                                                 &LockPageResult);
            }

            if (copy_to_user(arg, &LockPageResult, sizeof(struct LOCK_PAGE_RESULT)))
               break;
            request->ReturnCount = sizeof(LockPageResult);
            ret = STATUS_SUCCESS;
        }
        break;
    case IOCTL_PMM_UNLOCK_PAGE: /* unlock pages region */
        if (param->iData.PmmData) {
            if (param->iData.PmmData->MonitorPmmBuff) {
                mmUnlockPmmRegion(param->iData.PmmData, param->iData.PmmData->MonitorPmmBuff->pPageIds, param);
                request->ReturnCount = 0;
                ret = STATUS_SUCCESS;
            }
        }
        break;
    case IOCTL_ALLOC_SHAREBUFF_FOR_PMM: /* allocate memory manager shared buffer */
        if (!get_user(uintVal, (unsigned int *)arg)) {
            param->iData.pmmShareBuff = mm_mon_allocMem(uintVal);
            if (param->iData.pmmShareBuff) {
                put_user((unsigned int *)param->iData.pmmShareBuff, (unsigned int *)arg);
                ret = STATUS_SUCCESS;
            }
        }
        break;
    case IOCTL_FREE_SHAREBUFF_FOR_PMM: /* free memory manager shared buffer */
        if (arg) {
            if (!get_user(uintVal, (unsigned int *)arg)) {
                mm_mon_freeMem((void *)uintVal);
                ret = STATUS_SUCCESS;
            }
        }
        break;
    case IOCTL_PMM_INIT: /* initialize memory manager */
        if (arg) {
            if (!copy_from_user(&pmmInitial, arg, sizeof(struct pmm_initial_values_t))) {
                if(!mmInitPMM(param, &pmmInitial))
                    ret = STATUS_SUCCESS;
            }
        }
        request->ReturnCount = 0;
        break;
    case IOCTL_PMM_UNINIT: /* deinitialize memory manager */
        if (param->iData.PmmData)
            mmDeinitPMM(param->iData.PmmData, param);
        
        param->iData.PmmData = NULL;
        request->ReturnCount = 0;
        ret = STATUS_SUCCESS;
        break;
    case IOCTL_PMM_SET_MAX_MEMORY_FOR_VM: /* setup overall limit in pages */
        if (!get_user(uintVal, (unsigned int *)arg)) {
            if ((uintVal * 256) != vmGlobal.vmWset.overallMemoryLimit) {
                vmGlobal.vmWset.overallMemoryLimit = uintVal * 256;
                calcWset();
            }
        }
        ret = STATUS_SUCCESS;
        break;
    case IOCTL_MODE_FEATURES: /* get processor features */
        if (arg)
		{

			__asm__ __volatile__ (
				"movl %%cr4, %%eax\n\t"
				"movl %%eax, %0\n\t"
				: : "m" (uintVal)
				: "%eax", "memory"
								 );

			uiInfoBuff[0] = uintVal;
			uiInfoBuff[1] = is_vtx_support() || is_svm_support();
			if (copy_to_user(arg, &uiInfoBuff, sizeof(uiInfoBuff)))
				break;
			else
				ret = STATUS_SUCCESS;
        }
        break;
#ifdef CALL_MONITOR_FROM_GATE
    case IOCTL_SET_INT_GATE: /* set int gate */
        if (!ig_OpenGates(DEFAULT_GATE_EXCEPTION)) {
            ret = STATUS_SUCCESS;
        }
        break;
#endif
    case IOCTL_REPRESENT_MEM: /* representToMonMem */
        if (!arg)
            break;

        if (copy_from_user(&memInfo, arg, sizeof(MEMORY_INFO)))
            break;

        if (mm_mon_representToMonMem(memInfo.m_MemMap.m_pMemPtr, memInfo.m_dwMemSize,
                                    &(memInfo.m_MemMap.m_pMonPtr), &(memInfo.m_MemMap.m_pMdlPtr),
                                    param)) {
            break;
        }

        /* setup monitor specific values */
        if (memInfo.m_MemType == SYS_MONBODY_MEMORY) {
            param->MonAddr = memInfo.m_MemMap.m_pMonPtr;
            param->MonSize = memInfo.m_dwMemSize;
        }

        if (copy_to_user(arg, &memInfo, sizeof(MEMORY_INFO)))
            break;

        ret = STATUS_SUCCESS;
        break;
    case IOCTL_RELEASE_MEM: /* releaseFromMonMem */
        if (!arg)
            break;

        if (copy_from_user(&memInfo, arg, sizeof(MEMORY_INFO)))
            break;

        mm_mon_releaseFromMonMem(memInfo.m_MemMap.m_pMonPtr, memInfo.m_MemMap.m_pMdlPtr, param);

        if (memInfo.m_MemType == SYS_MONBODY_MEMORY)
            param->MonAddr = NULL;

        ret = STATUS_SUCCESS;
        break;
    case IOCTL_SETUP_MON_FUNCTIONS: /* setup mon functions and finish monitor loading */
        if (!arg)
            break;

        if (copy_from_user(&mFunc, arg, sizeof(struct monitor_functions_def_t) * MONFUNC_COUNT))
            break;

        /* setup functions pointers */
        for (i = 0; i < MONFUNC_COUNT; i++)
            param->iData.MonitorFuncs[i] = (monitor_funct_t)mFunc[i].fId;

        /* initialize callbacks  */
        vmSetExports(param);

        /* Monitor open */
        if (param->iData.MonitorFuncs[MONFUNC_OPEN]) {
            ret = param->iData.MonitorFuncs[MONFUNC_OPEN](&param->drvInfo, 0, param);
        }

        if (!NT_SUCCESS(ret)) {
            iDataClose(param);
            break;
        }
        break;
    default: /* request for monitor, try to transmit */
        if (!param->MonAddr) {
            Err("vmMainIoctl: Monitor PE module not loaded, but [%d] try to send control request to monitor.", current->pid);
            /* Monitor is not loaded */
            ret = STATUS_UNSUCCESSFUL;
            break;
        }

        if (request->InLength > 0) {
            monArg = (void *)kmalloc(request->InLength, GFP_KERNEL);
            if (monArg) {
                if (copy_from_user(monArg, arg, request->InLength)) {
                    Err("Error in copy parameters for monitor");
                    kfree(monArg);
                    break;
                }
                request->Buffer = monArg;
            } else { break; }
        }

        request->GMemInfoBuf = param->drvInfo.memBuff;

        if (param->iData.MonitorFuncs[MONFUNC_CONTROL])
            ret = param->iData.MonitorFuncs[MONFUNC_CONTROL](request, 0, NULL);

        if (request->InLength > 0) {
            if (copy_to_user(arg, monArg, request->InLength)) {
                Err("Error in copy parameters from monitor");
                kfree(monArg);
                break;
            }
            kfree(monArg);
            request->Buffer = arg;
        }

        if (!NT_SUCCESS(ret))
            break;

        param->drvInfo.memBuff = request->GMemInfoBuf;

        switch (cmd) {
            case IOCTL_INIT_MONITOR:
                ret = STATUS_UNSUCCESSFUL;
                if (copy_from_user(&gateReg, arg, sizeof(struct intgate_reg_t)))
                    break;

                /* create magic key */
                ig_GetGateMagic(&gateReg.magicNum);

                /* get current monitor identifier */
                gateReg.directCall = param->iData.MonId = 
                    ig_RegisterMonFunc((unsigned int)param->iData.MonitorFuncs[MONFUNC_REQUEST]);

                if (gateReg.directCall != GATE_INV_MON_ID) {
                    /* ident valid */
                    request->ReturnCount = sizeof(struct intgate_reg_t);

                    if (copy_to_user(arg, &gateReg, sizeof(struct intgate_reg_t)))
                        break;
                } else {
                    /* tooo many monitors :) */
                    Err("vmMainIoctl: Error - too many monitor identifiers, no space available for new identifier (%u)", SIMUL_VM_MAX_COUNT);
                    request->ReturnCount = 0;
                    break;
                }
                ret = STATUS_SUCCESS;
                break;
            case IOCTL_QUIT:
                if (param->iData.MonitorFuncs[MONFUNC_CLOSE])
                    param->iData.MonitorFuncs[MONFUNC_CLOSE](param->drvInfo.memBuff, 0, NULL);

                iDataClose(param);
            break;
        }
        break;
    }
    return ret;
}
