/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: manager.c                                                   *
 *                                                                     *
 * Description:                                                        *
 *  Memory manager routines.                                           *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/list.h>
#include <linux/mm.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>

#include "vmmain.h"
#include "mm.h"
#include "ioctls.h"
#include "logging.h"

/* lock/unlock PMM region */
int mmLockPmmRegion(struct pmm_struct_t *pmm,
                    API_PMM_PAGE *apiPmm,
                    struct module_param_t *param,
                    struct LOCK_PAGE_RESULT *pLockPageResult
) {

    PRIMARY_PAGE_ID *ppId;
    struct pmm_region_t *pmmReg;
    int ret = -ENOMEM, i;

    if (!pmm || !apiPmm || !param){
        ret = pLockPageResult->uResult = LOCK_PAGE_STATUS_FAIL;
        pLockPageResult->uRequiredSize = apiPmm->ppID.uRegionSize;
        return -EINVAL;
    }
    
    /* save statistic */
    pmm->MonitorPmmBuff->uPagesRequested += apiPmm->ppID.uRegionSize;
    
    /* check for size */
    if (apiPmm->ppID.uRegionSize > MAX_MEMORY_BLOCK) {
        Err("Too large memory block requested: %d", apiPmm->ppID.uRegionSize);
        ret = pLockPageResult->uResult = LOCK_PAGE_STATUS_FAIL;
        pLockPageResult->uRequiredSize = apiPmm->ppID.uRegionSize;
        goto err; /* error */
    }
    
    ppId = &apiPmm->ppID;
    if (ppId->PrimaryDesc) {
        Err("Non Zero descriptor [%p]", ppId->PrimaryDesc);
        ret = pLockPageResult->uResult = LOCK_PAGE_STATUS_FAIL;
        pLockPageResult->uRequiredSize = apiPmm->ppID.uRegionSize;
        goto err;
    }

    pmmReg = kmem_cache_alloc(vmGlobal.vmPmms.pmmRegionsSlab, GFP_KERNEL);
    if (!pmmReg) {
        Err("Can't allocate memory for PMM region");
        ret = pLockPageResult->uResult = LOCK_PAGE_STATUS_FAIL;
	pLockPageResult->uRequiredSize = apiPmm->ppID.uRegionSize;
        goto err;
    }

    for(i = 0; i < MAX_MEMORY_BLOCK; i++) {
        pmmReg->pages[i] = NULL;
        pmmReg->vmas[i] = NULL;
    }

    down(&pmm->accSem); /* lock pmm access */

    pmmReg->regionSize = ppId->uRegionSize;
    pmmReg->virtualAddr = pmm->virtualBase + (ppId->uRegionBase * PAGE_SIZE);

    /* lock pages range */
    ret = pLockPageResult->uResult = 
     mmLockPages( pmmReg->regionSize,
                pmmReg->virtualAddr,
                1,
                0,
                pmmReg->pages,
                NULL,
                apiPmm->PhyPageIdxsw,
                param->iData.bma);
    if (IS_LOCK_FAILED(pLockPageResult->uResult)){
	pLockPageResult->uRequiredSize = apiPmm->ppID.uRegionSize;
        Err("mmLockRegion: Pages locking failed %x",pLockPageResult->uResult);
        goto err_free;
    }

    ppId->PrimaryDesc = (unsigned int)pmmReg;
    list_add(&pmmReg->list, &pmm->regions);

    /* update counters */
    pmm->stat.rCount ++;
    pmm->stat.pCount += pmmReg->regionSize;
    up(&pmm->accSem); /* unlock pmm access */
    
    return 0;

err_free:
    kmem_cache_free(vmGlobal.vmPmms.pmmRegionsSlab, pmmReg);
err:
    apiPmm->ppID.uRegionSize = -1;

    up(&pmm->accSem); /* unlock pmm access */

    return ret;
}

void mmUnlockPmmRegion(struct pmm_struct_t *pmm, PRIMARY_PAGE_ID *ppId, struct module_param_t *param) {

    struct pmm_region_t *pmmReg;
    int i, z;
    unsigned char tp;

    if (!pmm || !ppId || !param)
        return;

    if (!pmm->MonitorPmmBuff)
        return;

    down(&pmm->accSem); /* lock pmm access */
    if (pmm->MonitorPmmBuff->uBuffPagesCount > MAX_ADJUST_WS_PAGES) {
        Err("mmFreePmmRegion: Pages count incorrect value %d", pmm->MonitorPmmBuff->uBuffPagesCount);
        goto out;
    }

    for(i = 0; i < pmm->MonitorPmmBuff->uBuffPagesCount; i++) {
        pmmReg = (struct pmm_region_t *)ppId[i].PrimaryDesc;
        if (pmmReg) {
            for (z = 0; z < pmmReg->regionSize; z++) {
                if (pmmReg->pages[z]) {
                    /* make dirty for swap */
                    get_user(tp, (unsigned char *)pmmReg->virtualAddr + (z * PAGE_SIZE));
                    put_user(tp, (unsigned char *)pmmReg->virtualAddr + (z * PAGE_SIZE));

                    /* put page back */
                    vmPutPage(pmmReg->pages[z], param->iData.bma);
                }
            }

            list_del(&pmmReg->list);
            kmem_cache_free(vmGlobal.vmPmms.pmmRegionsSlab, pmmReg);
            pmm->stat.rCount --;
        }
    }
out:
    up(&pmm->accSem); /* unlock pmm access */
}

/* unlock all memory locked by PMM */
void mmFreeAllPmmRegions(struct pmm_struct_t *pmm, struct module_param_t *param) {

    struct pmm_region_t *pmmReg;
    struct list_head *q, *s;
    int i;

    if (!pmm)
        return;

    down(&pmm->accSem); /* lock pmm access */

    list_for_each_safe(q, s, &pmm->regions) {
        pmmReg = (struct pmm_region_t *)list_entry(q, struct pmm_region_t, list);
        if (pmmReg) {
            for (i = 0; i < pmmReg->regionSize; i++) {
                if (pmmReg->pages[i])
                    vmPutPage(pmmReg->pages[i], ((param) ? param->iData.bma : NULL));
            }
            list_del(&pmmReg->list);
            kmem_cache_free(vmGlobal.vmPmms.pmmRegionsSlab, pmmReg);
        }
    }
    up(&pmm->accSem); /* unlock pmm access */
}

/* initialize/deinitialize PMM */
int mmInitPMM(struct module_param_t *param, struct pmm_initial_values_t *pmm_initial) {

    struct pmm_struct_t *pmm;

    if (!param)
        return -EINVAL;
    
    pmm = (struct pmm_struct_t *)kmalloc(sizeof(struct pmm_struct_t), GFP_KERNEL);
    if (!pmm)
        return -ENOMEM;

    pmm->MonitorPmmBuff = (GUEST_PHY_MEM_MAN *)pmm_initial->PmmShareBuff;
    pmm->virtualBase = (void *)pmm_initial->GuestMemBase;

    /* save in param */
    param->iData.PmmData = pmm;
    param->iData.PmmData->MonitorPmmBuff->WSSpinLock = SPINLOCK_UNLOCKED;

    /* init regions list */
    INIT_LIST_HEAD(&pmm->regions);

    /* init vmalist */
    INIT_LIST_HEAD(&pmm->vmalist);

    sema_init(&pmm->accSem, 1); /* init pmm semaphore */

    /* add to list */
    down(&vmGlobal.vmPmms.pmmAccSem);
    list_add(&pmm->list, &vmGlobal.vmPmms.pmmList);
    up(&vmGlobal.vmPmms.pmmAccSem);

    /* calculate workingset */
    if(calcWset()) {
        down(&vmGlobal.vmPmms.pmmAccSem);
        list_del(&pmm->list);
        up(&vmGlobal.vmPmms.pmmAccSem);
    
        param->iData.PmmData = NULL;
        return -ENOMEM;
    }

    pmm->stat.pCount = 0;
    pmm->stat.rCount = 0;
    pmm->stat.tPages = pmm->MonitorPmmBuff->uGuestRamPages;
    pmm->stat.ownerPid = current->pid;

    vmGlobal.vmPmms.totalMem += pmm->stat.tPages;

    return 0;
}

void mmDeinitPMM(struct pmm_struct_t *pmm, struct module_param_t *param) {

    if (!pmm)
        return;

    /* remove pmm from list */
    down(&vmGlobal.vmPmms.pmmAccSem);

    /* increment pages */
    vmGlobal.vmPmms.totalMem -= pmm->stat.tPages;
    
    /* free all pages associated with this pmm */
    mmFreeAllPmmRegions(pmm, param);

    list_del(&pmm->list);

    kfree(pmm);

    up(&vmGlobal.vmPmms.pmmAccSem);
}

/* pmm managers initialization */
int QUICKCALL vmPmmsMainInit(void) {

    sema_init(&vmGlobal.vmPmms.pmmAccSem, 1);

    INIT_LIST_HEAD(&vmGlobal.vmPmms.pmmList);

    vmGlobal.vmPmms.pmmRegionsSlab = kmem_cache_create("vmRegions", sizeof(struct pmm_region_t), 0,
                                                       SLAB_HWCACHE_ALIGN, NULL, NULL);
    if (!vmGlobal.vmPmms.pmmRegionsSlab) {
        Err("Can't allocate vmRegions cache SLAB");
        return -1;
    }

    return 0;
}

/* deinit pmms */
void QUICKCALL vmPmmsMainDeinit(void) {

    struct list_head *q, *s;
    struct pmm_struct_t *pmm;

    /* check for pmms */
    list_for_each_safe(q, s, &vmGlobal.vmPmms.pmmList) {
        pmm = (struct pmm_struct_t *)list_entry(q, struct pmm_struct_t, list);
        if (pmm)
            mmDeinitPMM(pmm, NULL);
    }

    /* free slabcache */
    kmem_cache_destroy(vmGlobal.vmPmms.pmmRegionsSlab);
}
