/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: mm.h                                                        *
 *                                                                     *
 * Description:                                                        *
 *  Main memory works header.                                          *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#ifndef __MM_MM_H__
#define __MM_MM_H__

#include "vm_compiler.h"

#ifndef _HYPER_BINARY_PART_
#include <asm/semaphore.h>
#include <linux/mm.h>

#include "compat_memory.h"
#include "bma.h"
#include "PMM.h"
#include "ioctls.h"


/* cache operations */
#define vmGetPage(pg, bma) {     \
    get_page(pg);                \
    if (vm_mmGetPage(pg, bma)) { \
        put_page(pg);           }}

#define vmPutPage(pg, bma) { vm_mmPutPage(pg, bma); }

/* get/put pages */
#ifndef _HYPERVISOR_
  int  vm_mmGetPage(struct page *, struct bma_t *);
  void vm_mmPutPage(struct page *, struct bma_t *);
#else
  static inline int  vm_mmGetPage(struct UNUSED page *pg, struct UNUSED bma_t *bma) { return 0; }
  static inline void vm_mmPutPage(struct UNUSED page *pg, struct UNUSED bma_t *bma) { put_page(pg); }
#endif

void QUICKCALL vm_mmFreeCurrent(struct bma_t *);

/* lock/unlock pages */
void QUICKCALL mmUnlockPages(struct page **, int, struct bma_t *);
int QUICKCALL mmLockPages(int, void *, int, int, struct page **, struct vm_area_struct **, unsigned int *, struct bma_t *);

/* get pages wrapper */
int QUICKCALL mmGetPages(void *, int, struct page **, struct vm_area_struct **, struct bma_t *);

#endif /* _HYPER_BINARY_PART_ */

/* alloc_pages/free_pages */
void* STDCALL mmAllocPages(int);
void  STDCALL mmFreePages(void *, int);

/* get by addr */
struct page* STDCALL mmUGetPageByAddr(void *);
struct page* STDCALL mmKGetPageByAddr(void *);

#ifndef _HYPER_BINARY_PART_
/* pmm region structure */
struct pmm_region_t {
    /* list of regions */
    struct list_head list;
    /* virtual address of region */
    void *virtualAddr;
    /* size of region */
    unsigned int regionSize;
    /* pages in region */
    struct page *pages[MAX_MEMORY_BLOCK];
    /* pmm parent */
    struct vm_area_struct *vmas[MAX_MEMORY_BLOCK];
};

struct pmm_vma_list_t {
    struct list_head list;
    struct vm_area_struct *vma;
};

/* main PMM structure */
struct pmm_struct_t {
    /* list of PMM's */
    struct list_head list;
    /* list of PMM's */
    struct list_head regions;
    /* virtual address base of guest */
    void *virtualBase;
    /* pmm access semaphore */
    struct semaphore accSem;
    /* vma for this pmm */
    struct list_head vmalist;
    /* monitor shared buffer */
    GUEST_PHY_MEM_MAN *MonitorPmmBuff;
    /* statistic struct */
    struct {
        /* count of pages */
        unsigned long pCount;
        /* count of regions */
        unsigned long rCount;
        /* owner pid */
        pid_t ownerPid;
        /* total guest pages */
        unsigned long tPages;
    } stat;
};

#endif /* _HYPER_BINARY_PART_ */

enum PROS_PAGING_MODE
{
	PROS_PAGING_INVALID_MODE = -1,
	PROS_PAGING_STD,
	PROS_PAGING_PAE
};


/* pte entry */
typedef union {

    unsigned int cell;
	
    /* pte bits */
    struct {
		unsigned int    P : 1;
		unsigned int    RW : 1;
		unsigned int    US : 1;
		unsigned int    PWT : 1;
		unsigned int    PCD : 1;
		unsigned int    A : 1;
		unsigned int    D : 1;
		unsigned int    PS : 1;
		unsigned int    G : 1;
		unsigned int    ProgAvail : 3;
		unsigned int    Base : 20;
	} details PACKED;
} ALIGNED pte_entry_t;

/* PAE mode pte entry */
typedef union {

    unsigned long long cell;
	
    /* pte bits */
    struct {
		unsigned long long P : 1;
		unsigned long long RW : 1;
		unsigned long long US : 1;
		unsigned long long PWT : 1;
		unsigned long long PCD : 1;
		unsigned long long A : 1;
		unsigned long long D : 1;
		unsigned long long PS : 1;
		unsigned long long G : 1;
		unsigned long long ProgAvail : 3;
		unsigned long long Base : 24;
                unsigned long long Reserved2 : 27;
                unsigned long long NX : 1;                
	} details PACKED;
} ALIGNED pte_pae_entry_t;


/* unified pte/pde/pdtp entry for different modes */
typedef union {
    /* Non-PAE mode PTE */
    pte_entry_t		Std;
    /* PAE mode PTE */	
    pte_pae_entry_t	Pae;
} ALIGNED pte_unified_entry_t;

/* protect/unprotect page by RW */
void STDCALL mmProtectPageRw(unsigned int, int, unsigned int);
pte_unified_entry_t* STDCALL mmGetPagePte(unsigned int, unsigned int);

void STDCALL mmEnableExecution(unsigned int linAddr, unsigned int pages, unsigned int uPagingMode);

#if !defined(_HYPERVISOR_) && !defined(_HYPER_BINARY_PART_)
/* initialize/deinitialize PMM for each copy of application */
int mmInitPMM(struct module_param_t *, struct pmm_initial_values_t *);
void mmDeinitPMM(struct pmm_struct_t *, struct module_param_t *);

/* manage pmm regions */
int mmLockPmmRegion(struct pmm_struct_t *,
                    API_PMM_PAGE *,
                    struct module_param_t *,
                    struct LOCK_PAGE_RESULT *);
void mmUnlockPmmRegion(struct pmm_struct_t *, PRIMARY_PAGE_ID *, struct module_param_t *);

/* setup workingset initial values */
static inline void vmWsetSetup(void) {

    vmGlobal.vmWset.wsCount = 0;
    vmGlobal.vmWset.regPages = 0;
    vmGlobal.vmWset.vmsPages = 0;

    vmGlobal.vmWset.totalAllocated = 0;

    sema_init(&vmGlobal.vmWset.wsSem, 1);
}
#endif /* _HYPERVISOR_ && _HYPER_BINARY_PART_ */

/* wset functions */
int QUICKCALL calcWset(void);

/* main pmm's structure init/deinit */
int QUICKCALL vmPmmsMainInit(void);
void QUICKCALL vmPmmsMainDeinit(void);

// check page is grater 1gb
#define IS_PAGE_LARGE(phyPageIdx) (phyPageIdx >=0x100000)

#endif /* __MM_MM_H__*/
