/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: monmem.c                                                    *
 *                                                                     *
 * Description:                                                        *
 *  Monitor memory allocation/deallocation callbacks.                  *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <asm/io.h>

#include "vm_compiler.h"
#include "vmmain.h"
#include "mm.h"
#include "logging.h"

#include "compat_memory.h"

/* calculates pages count for size */
#define __S2P(size) (((size)+PAGE_SIZE-1)>>PAGE_SHIFT)

/* alloc/release aligned memory block */
void STD_INTERCOM *mm_mon_allocMem(long size) {

    if (size > 0)
        return vmalloc(size);

    return NULL;
}

void STD_INTERCOM mm_mon_freeMem(void *buff) {

    if (buff)
        vfree(buff);
}

/* alloc/release contig memory block */
void STD_INTERCOM *mm_mon_allocContigMem(long size) {

    if (size > 0)
        return mmAllocPages(__S2P(size));

    return NULL;
}

void STD_INTERCOM mm_mon_freeContigMem(void *buff, long size) {

    mmFreePages(buff, __S2P(size));
}

/*
 * Representation neccesary Application address area to Monitor
 * Memory lock and mapping
 */
long STD_INTERCOM mm_mon_representToMonMem(void *primaryAddr, long primarySize, void **monAddr, void **MDL, void *monCtx) {

    MON_MEM_DESC *mmDesc;
    int nCount = 0, pCount = (primarySize + PAGE_SIZE - 1) / PAGE_SIZE; /* count pages for size */
    int ret = -ENOMEM, i;
    struct module_param_t *param = (struct module_param_t *)monCtx;

    if (!primaryAddr || !param)
        return -EINVAL;

    /* allocate stuff */
    mmDesc = (MON_MEM_DESC *)kmalloc(sizeof(MON_MEM_DESC), GFP_KERNEL);
    if (!mmDesc)
        goto err;

    mmDesc->vmas = (struct vm_area_struct **)kmalloc(sizeof(struct vm_area_struct *) * pCount, GFP_KERNEL);
    if (!mmDesc->vmas)
        goto err_vmas;

    /* zero vmas */
    for(i = 0; i < pCount; i++)
        mmDesc->vmas[i] = NULL;

    mmDesc->pPages = (struct page **)kmalloc(sizeof(struct page *) * pCount, GFP_KERNEL);
    if (!mmDesc->pPages)
        goto err_pages;

    /* save addrs */
    *monAddr = primaryAddr;
    mmDesc->pMonAddr = (unsigned int)*monAddr;

    *MDL = (void *)mmDesc;
    mmDesc->bValid = 0;

    /* lock pages */
    nCount = mmGetPages(primaryAddr, pCount, mmDesc->pPages, mmDesc->vmas, param->iData.bma);
    if (nCount != pCount) {
        goto err_all;
    }

    mmDesc->bValid = 1; /* :) */
    mmDesc->uPages = nCount;

    return 0;
err_all:
    mmUnlockPages(mmDesc->pPages, nCount, param->iData.bma);
err_pages:
    kfree(mmDesc->pPages);
err_vmas:
    kfree(mmDesc->vmas);
err:
    kfree(mmDesc);

    Err("Error in mm_mon_representToMonMem.");

    return ret;
}

/*
 * Function cancels fixed memory representation to Monitor address area
 * Memory unlock and unmapping
 */
void STD_INTERCOM mm_mon_releaseFromMonMem(void *monAddr, void *MDL, void *monCtx) {

    MON_MEM_DESC *mmDesc;
    struct module_param_t *param = (struct module_param_t *)monCtx;

    if (!MDL || !param)
        return;

    mmDesc = (MON_MEM_DESC *)MDL;

    mmUnlockPages(mmDesc->pPages, mmDesc->uPages, param->iData.bma);

    /* cleanup */
    kfree(mmDesc->vmas);
    kfree(mmDesc->pPages);
    kfree(mmDesc);

}

/* Gather indexes of physical pages for virtual address area */
unsigned STD_INTERCOM mm_mon_getPagesIndex(unsigned long *page, int max_pages, void *vAddr, unsigned size, void *monCtx) {

    struct page **pages;
    int i, nCount, pCount;
    struct module_param_t *param = (struct module_param_t *)monCtx;

    if (!page || !vAddr)
        return 0;

    pCount = (size + PAGE_SIZE - 1) >> PAGE_SHIFT;

    if (pCount > max_pages)
        pCount = max_pages;

    // This path only for special contigues memory!
    if (!param)
    {
        for(i = 0; i < pCount; i++)
        {
            page[i] = virt_to_phys(vAddr + i * PAGE_SIZE) >> PAGE_SHIFT;
            if (!page[i])
                return 0;
        }
 
        return pCount;
    }
	
    /* alloc pages */
    pages = (struct page **)kmalloc(pCount * sizeof(struct pages *), GFP_KERNEL);
    if(!pages)
        goto err;

    /* lock phys pages */
    nCount = mmGetPages(vAddr , pCount, pages, NULL, param->iData.bma);
    if (nCount != pCount)
        goto err_free;

    /* gather pfn's for pages and save it in monitor pointer */
    for(i = 0; i < nCount; i++)
    {
        page[i] = page_to_pfn(pages[i]);
        //check if physical address is grater 1gb idx

        if(IS_PAGE_LARGE(page[i]))
        {
            goto err_free;
	}
    }

    kfree(pages);

    return nCount;

err_free:
    kfree(pages);

err:
    Err("Error in mm_mon_getPagesIndex.");
    return 0;
}
