/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: pages.c                                                     *
 *                                                                     *
 * Description:                                                        *
 *  Lowlevel functions for work with physical pages.                   *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/mm.h>
#include <linux/highmem.h>
#include <asm/pgtable.h>
#include <asm/uaccess.h>

#include "vmmain.h"
#include "mm.h"
#include "logging.h"
#include "compat_memory.h"

/* lock pages */
int QUICKCALL mmLockPages(int size, void *addr, int wr, int fr, struct page **pages, struct vm_area_struct **vmas, unsigned int *phidx, struct bma_t *bma) {

    int i;
    int pcnt = 0;
    struct vm_area_struct *vma;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 19)
    struct page *page, *pageT;
    unsigned int *iAddr, tp;
    int z;
#endif

    if (!addr || !pages) 
        return LOCK_PAGE_STATUS_FAIL;

    if (size < 0) 
        return LOCK_PAGE_STATUS_FAIL;

    if (vmas) {
        for(i = 0; i < size; i++) {
            vma = find_vma(current->mm, (unsigned int)addr + (i * PAGE_SIZE));
            if (vma) {
                vmas[i] = vma;
                vmas[i]->vm_flags |= VM_LOCKED;
                vmas[i]->vm_flags &= ~VM_IO;
            } else {
                Panic("NULL VMA!");
            }
        }
    }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 19)
    down_read(&current->mm->mmap_sem);
    pcnt = get_user_pages(current, current->mm, (unsigned int)addr, size,
                               wr, fr, pages, ((vmas) ? vmas : NULL));
    up_read(&current->mm->mmap_sem);

    if (pcnt != size) {
        for(i = 0; i < pcnt; i++) {
            put_page(pages[i]);
        }
        return LOCK_PAGE_STATUS_FAIL;
    } else {
        for(i = 0; i < size; i++) {
            if (phidx){
                phidx[i] = page_to_pfn(pages[i]);
                if(IS_PAGE_LARGE(phidx[i])) //>1GB
                {
                     int j;
                     for(j=0; j<i ; j++)
                     {
                         vmPutPage(pages[j], bma);
                     }
                     for(;j<size;j++)
                     {
                         put_page(pages[j]);
                     }
                     return LOCK_PAGE_STATUS_ONLY_HUGE;
                }
            }
            if (vm_mmGetPage(pages[i], bma))
		put_page(pages[i]); /* put twice referenced page */
        }
    }
#else
    for(i = 0; i < size; i++) {
        iAddr = (unsigned int *)((unsigned int)addr + i * PAGE_SIZE);
        if(get_user(tp, iAddr) < 0) {
            Err("vmLoockPages: get_user() failed for %p", iAddr);
            for(z = 0; z < i ; z++ ) {
                put_page(pages[z]);
            }
            return LOCK_PAGE_STATUS_FAIL;
        } else {
            put_user(tp, iAddr); /* korova */
            page = mmUGetPageByAddr((void *)iAddr);
            if (page) {

                vm_mmGetPage(page, bma);

                pageT = mmUGetPageByAddr((void *)iAddr);

                if(pageT && (pageT != page)) {
                    vm_mmPutPage(page, bma);
                    pcnt = 0;
                } else {
                    pcnt = 1;
                }
            } else {
                Err("mmLockPages: NULL pfn from addr %p", iAddr);
            }

            /* check for error. */
            if(!pcnt) {
                Err("mmLockPages: Unable to lock page");

                /* unlock all locked pages */
                for(z = 0; z < i ; z++ ) {
                    put_page(pages[z]);
                }
                return LOCK_PAGE_STATUS_FAIL;
            }

            pages[i] = page;
            if (phidx) 
            {
                phidx[i] = page_to_pfn(page);
                if(IS_PAGE_LARGE(phidx[i]))
                {
                    for(z = 0; z < i ; z++ ) {
                       vmPutPage(pages[z], bma);
                    }
                    return LOCK_PAGE_STATUS_ONLY_HUGE;
                }
            }
            if (vm_mmGetPage(pages[i], bma))
		put_page(pages[i]); /* put twice referenced page */

        }
    }
#endif

    return LOCK_PAGE_STATUS_SUCCESS;
}

/* unlock pages */
void QUICKCALL mmUnlockPages(struct page **pages, int cnt, struct bma_t *bma) {

    int i;

    if (!pages)
        return;

    for(i = 0; i < cnt; i++) {
        vmPutPage(pages[i], bma);
    }
}

/* get page from kernelspace address */
struct page* STDCALL mmKGetPageByAddr(void *addr) {
    
    compat_pgd_t * dir;
    pmd_t * pmd;
    pte_t * pte;

    dir = (compat_pgd_t *)pgd_offset_k((unsigned int)addr);

    if(dir && compat_pgd_present(*dir)) {
        pmd = pmd_offset(dir, (unsigned int)addr);
        if(pmd && pmd_present(*pmd)) {
            pte = pte_offset_kernel(pmd, (unsigned int)addr);
            if(pte && pte_present(*pte)) {
                return pfn_to_page(pte_pfn(*pte));
            }
        }
    }

    return NULL;
}

/* get page from userspace address */
struct page* STDCALL mmUGetPageByAddr(void *addr) {

    compat_pgd_t *dir;
    pmd_t *pmd;
    pte_t *pte;

    spin_lock(&current->mm->page_table_lock);

    dir = (compat_pgd_t *)pgd_offset(current->mm, (unsigned int)addr);

    if(dir && compat_pgd_present(*dir)) {
        pmd = pmd_offset(dir, (unsigned int)addr);
        if(pmd && pmd_present(*pmd)) {
            pte = pte_offset_map(pmd, (unsigned int)addr);
            if(pte && pte_present(*pte)) {
                spin_unlock(&current->mm->page_table_lock);
                return pte_page(*pte);
            }
        }
    }

    spin_unlock(&current->mm->page_table_lock);

    return NULL;
}

/* get_user_pages wrapper */
int QUICKCALL mmGetPages(void *toAddr, int pCount, struct page **pages, struct vm_area_struct **vmas, struct bma_t *bma) {

    struct page *pg = 0, UNUSED *pgtest;
    struct vm_area_struct *vma;
    unsigned char *addr;
    int i, gPages;
    char tp;
    int iRetVal;

    if (!pages || !toAddr)
        return -EINVAL;

    addr = (unsigned char *) toAddr;

    for(i = 0; i < pCount; i++) {
        if(vmas) {
            vma = find_vma(current->mm, (unsigned int)addr);

            if(vma) {
                /* disable page IO (swapping) */
                vma->vm_flags |= VM_LOCKED;
                vma->vm_flags &= ~VM_IO;
            }
            vmas[i] = vma;
        }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 19) /* Korova :) */
        get_user(tp, (char *)addr);
        put_user(tp, (char *)addr);
#endif
        gPages = 1;

        iRetVal = mmLockPages(1, addr, 1, 0, &pg, NULL, NULL, bma);
        if(IS_LOCK_FAILED(iRetVal)){
            gPages = -1;
        }

        /* check for kernelspace page */
        if(gPages < 1) {
            /* Kernel pages are looked here */
            *addr = *addr;

            pg = mmKGetPageByAddr(addr);
            if(pg) {
                vmGetPage(pg, bma);
                pgtest = mmKGetPageByAddr(addr);
                if (pg != pgtest) {
                    vmPutPage(pg, bma);
                    pg = NULL;
                }
            }
        }

        /* save or not */
        if(pg) {
            pages[i]= pg;
        } else {
            //Panic("NULL Page pointer for addr %p -> %p [%d] (probably pages sanity check not passed)",
            //              toAddr, addr + (pCount * PAGE_SIZE), pCount);
            if(i != 0)
            {
                mmUnlockPages(pages,i,bma);
            }
            return LOCK_PAGE_STATUS_FAIL; //iRetval
        }
        addr += PAGE_SIZE;
    }

    return i;
}


/* getting page PTE for non-PAE mode */
pte_entry_t* STDCALL mmGetPagePteStd(unsigned int linearAddr, unsigned int uCr3) {

    pte_entry_t *pDir, *pTbl; 
    struct page *pgD, *pgT;
    unsigned int tblIdx;
    unsigned int dirIdx;
   
    /* get indexes */
    tblIdx = linearAddr >> 12;
    dirIdx = tblIdx >> 10;
    tblIdx &= 0x3ff;
        
    pgD = pfn_to_page(uCr3 >> 12);
    if (!pgD)
        return NULL;
    
    /*
     * PD and PT can't be in highmem and we can
     * use page_address() instead of kmap and other magic code
     */
    pDir = (pte_entry_t *)page_address(pgD);
    if ((!pDir) || (pDir[dirIdx].details.P == 0)) {
        Err("Can't map page directory.");
        return NULL;
    }

    /* check for 4Mb page */
    if (pDir[dirIdx].details.PS) {
//	    Log(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 4MB Page !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
	    return NULL;
    }

    pgT = pfn_to_page(pDir[dirIdx].details.Base);
    if (!pgT) {
        return NULL;
    }
    
    
    pTbl = (pte_entry_t *)page_address(pgT);
   
    if ((!pTbl) || (pTbl[tblIdx].details.P == 0)) {
        Err("Can't map page table.");
        return NULL;
    }
    
    return &pTbl[tblIdx];
}

/* getting page PTE for PAE mode */
pte_pae_entry_t* STDCALL mmGetPagePtePae(unsigned int linearAddr, unsigned int uCr3) {

    pte_pae_entry_t *pDir, *pTbl, *pPdtp; 
    struct page *pgTp, *pgD, *pgT;
    unsigned int tblIdx;
    unsigned int dirIdx;
    unsigned int pdtpIdx;

    pdtpIdx = linearAddr >> 30;
    dirIdx = (linearAddr >> 21) & 0x1ff;
    tblIdx = (linearAddr >> 12) & 0x1ff;
	
    pgTp = pfn_to_page(uCr3 >> 12);
    if (!pgTp) {
        return NULL;
    }
    
    /*
     * PD and PT can't be in highmem and we can
     * use page_address() instead of kmap and other magic code
     */
    pPdtp = (pte_pae_entry_t *)(page_address(pgTp) + (uCr3 & 0xfe0));

    if ((!pPdtp) || (pPdtp[pdtpIdx].details.P == 0)) {
        Err("Can't map page directory pointer table.");
        return NULL;
    }
	
    pgD = pfn_to_page(pPdtp[pdtpIdx].details.Base);
    if (!pgD) {
        return NULL;
    }
    
    pDir = (pte_pae_entry_t *)page_address(pgD);
    if ((!pDir) || (pDir[dirIdx].details.P == 0)) {
        Err("Can't map page directory.");
        return NULL;
    }

    /* check for 4Mb page */
    if (pDir[dirIdx].details.PS) {
//	    Log(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> 2MB Page !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
	    return NULL;
    }

    pgT = pfn_to_page(pDir[dirIdx].details.Base);
    if (!pgT) {
        return NULL;
    }
    
    
    pTbl = (pte_pae_entry_t *)page_address(pgT);
   
    if ((!pTbl) || (pTbl[tblIdx].details.P == 0)) {
        Err("Can't map page table.");
        return NULL;
    }
    
    return &pTbl[tblIdx];
}


/* getting PTE for page */
pte_unified_entry_t* STDCALL mmGetPagePte(unsigned int linearAddr, unsigned int uPagingMode) {

    unsigned int cr3;
   
    if (!linearAddr)
        return NULL;
	
    /* read cr3 register */
    __asm__ __volatile__ (
        "movl %%cr3, %%eax\n\t"
        "movl %%eax, %0\n\t"
        : : "m" (cr3)
        : "%eax", "memory"
    );

    /* getting PTE for different modes */
    switch (uPagingMode)
    {
    case PROS_PAGING_STD:
	return (pte_unified_entry_t*)mmGetPagePteStd(linearAddr, cr3);
    case PROS_PAGING_PAE:
	return (pte_unified_entry_t*)mmGetPagePtePae(linearAddr, cr3);
    default:
	return NULL;
    }
}


/* getting page PTE for PAE mode */
void STDCALL mmClearPageNXPae(unsigned int linearAddr, unsigned int uCr3) {

    pte_pae_entry_t *pDir, *pTbl, *pPdtp; 
    struct page *pgTp, *pgD, *pgT;
    unsigned int tblIdx;
    unsigned int dirIdx;
    unsigned int pdtpIdx;
    
    pdtpIdx = linearAddr >> 30;
    dirIdx = (linearAddr >> 21) & 0x1ff;
    tblIdx = (linearAddr >> 12) & 0x1ff;
	
    pgTp = pfn_to_page(uCr3 >> 12);
    if (!pgTp) {
        return;
    }
    
    /*
     * PD and PT can't be in highmem and we can
     * use page_address() instead of kmap and other magic code
     */
    pPdtp = (pte_pae_entry_t *)(page_address(pgTp) + (uCr3 & 0xfe0));

    if ((!pPdtp) || (pPdtp[pdtpIdx].details.P == 0)) {
        Err("Can't map page directory pointer table.");
        return;
    }
	
    pgD = pfn_to_page(pPdtp[pdtpIdx].details.Base);
    if (!pgD) {
        return;
    }
    
    pDir = (pte_pae_entry_t *)page_address(pgD);
    if ((!pDir) || (pDir[dirIdx].details.P == 0)) {
        Err("Can't map page directory.");
        return;
    }

    
    pDir[dirIdx].details.NX = 0;
    
    /* check for 4Mb page */
    if (pDir[dirIdx].details.PS) {
        invalidate_page(linearAddr);
	return;
    }
    
    
    pgT = pfn_to_page(pDir[dirIdx].details.Base);
    if (!pgT) {
        return;
    }
    
    
    pTbl = (pte_pae_entry_t *)page_address(pgT);
   
    if ((!pTbl) || (pTbl[tblIdx].details.P == 0)) {
        Err("Can't map page table.");
        return;
    }
        
    pTbl[tblIdx].details.NX = 0;
    
    invalidate_page(linearAddr);
}




void STDCALL mmEnableExecution(unsigned int linAddr, unsigned int uPagesCnt, unsigned int uPagingMode)
{
    unsigned int cr3;
    unsigned int i;
   
    if(uPagingMode!= PROS_PAGING_PAE)
        return;
   
    /* read cr3 register */
    __asm__ __volatile__ (
        "movl %%cr3, %%eax\n\t"
        "movl %%eax, %0\n\t"
        : : "m" (cr3)
        : "%eax", "memory"
    );
    for(i = 0; i< uPagesCnt; i++)
    {
       mmClearPageNXPae(linAddr, cr3);
       linAddr = linAddr + PAGE_SIZE;
    }
}


/* alloc pages via alloc_pages */
void* STDCALL mmAllocPages(int cnt) {

    struct page *pg;
    int order;

    order = 0;
    while(cnt > (1 << order))
        order++;
    
    pg = alloc_pages(GFP_ATOMIC, order);
    if (pg == NULL)
        return NULL;

    if (compat_change_page_attr(pg, 1 << order, PAGE_KERNEL_NOCACHE)) {
        compat_global_flush_tlb();
        __free_page(pg);
        return NULL;
    }

    compat_global_flush_tlb();
    
    return page_address(pg);
}

/* free pages via free_pages */
void STDCALL mmFreePages(void *addr, int cnt) {

    struct page *pg;
    int order;

    order = 0;
    while(cnt > (1 << order))
        order++;

    if (!addr)
        return;

    pg = virt_to_page(addr);
    if (!pg)
        return;
    
/*    compat_change_page_attr(pg,  1 << order, PAGE_KERNEL);*/
    
    compat_global_flush_tlb();
    
    free_pages((unsigned long)addr, order);
}

/* protect or unprotect page on liner address */
void STDCALL mmProtectPageRw(unsigned int linAddr, int RW, unsigned int uPagingMode) {

    pte_unified_entry_t *pPte;

    /* get pte */
    if (!(pPte = mmGetPagePte(linAddr, uPagingMode)))
        return;
    
    
    /* set protection */
    if (RW) {
        pPte->Std.details.RW = 1; /* unprotect */
    } else {
        pPte->Std.details.RW = 0; /* protect */
    }
   
    /* invalidate page */
    invalidate_page(linAddr);
}
