/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: vmproc.c                                                    *
 *                                                                     *
 * Description:                                                        *
 *  proc file system interface realization.                            *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/mm.h>
#include <linux/proc_fs.h>

#include "vmmain.h"
#include "logging.h"

#define ROOT_DIR "parallels"

static struct proc_dir_entry *vmProcRoot;

static int vmProcMemRead(char *, char **, off_t, int, int *, void *);
static int vmProcStatRead(char *, char **, off_t, int, int *, void *);

static struct {
    char *name;
    read_proc_t *read;
    struct proc_dir_entry *entry;
} vmProcEntries[] = {
    {"mem", vmProcMemRead, NULL},
    {"stat", vmProcStatRead, NULL}
};

enum {
    PROC_ENTRY_MEM,
    PROC_ENTRY_STAT,
    PROC_ENTRY_MAX
};

/* entries readers */
static int vmProcMemRead(char *buffer, char **start, off_t offset, int length, int *eof, void *data) {

    struct list_head *q;
    struct pmm_struct_t *pmm;
    struct sysinfo si;

    down(&vmGlobal.vmWset.wsSem);

    si_meminfo(&si);

    length = sprintf(buffer, 
                      "VmsCount            : %d\n"
                      "BmaSizePerApp       : %lu Kb\n"
                      "OverallMemoryLimit  : %7lu P %5lu Mb\n"
                      "TotalMemoryUsed     : %7lu P %5lu Mb\n"
                      "VmsAvailable        : %7lu P %5lu Mb\n"
                      "MMUsed              : %7lu P %5lu Mb\n"
                      "EachWSLimit         : %7lu P %5lu Mb\n"
                      "TotalAllocated      : %7lu P %5lu Mb\n"
                      "AllVmsMemory        : %7lu P %5lu Mb\n",
                      vmGlobal.vmWset.wsCount,
                      (si.totalram / 8) / 1024,
                      vmGlobal.vmWset.overallMemoryLimit, vmGlobal.vmWset.overallMemoryLimit / 256,
                      vmGlobal.vmWset.allVmsUsed, vmGlobal.vmWset.allVmsUsed / 256,
                      vmGlobal.vmWset.vmsPages, vmGlobal.vmWset.vmsPages / 256,
                      vmGlobal.vmWset.regPages, vmGlobal.vmWset.regPages / 256,
                      vmGlobal.vmWset.eachWS, vmGlobal.vmWset.eachWS / 256,
                      vmGlobal.vmWset.totalAllocated, vmGlobal.vmWset.totalAllocated / 256,
                      vmGlobal.vmPmms.totalMem, vmGlobal.vmPmms.totalMem / 256);

    up(&vmGlobal.vmWset.wsSem);

    down(&vmGlobal.vmPmms.pmmAccSem);
    list_for_each(q, &vmGlobal.vmPmms.pmmList) {
        pmm = (struct pmm_struct_t *)list_entry(q, struct pmm_struct_t, list);

        down(&pmm->accSem);
        length += sprintf(buffer + length, 
                          "\nApplication        : %d\n"
                          " CurrentPages      : %7u P %5u Mb\n"
                          " MaxPages          : %7u P %5u Mb\n"
                          " UsedPages         : %7lu%%\n"
                          " RegionsCount      : %7lu\n"
                          " AveragePPR        : %7lu\n"
                          " VirtualBase       : %p\n", 
                          pmm->stat.ownerPid,
                          pmm->MonitorPmmBuff->uCurrentPages, pmm->MonitorPmmBuff->uCurrentPages / 256,
                          pmm->MonitorPmmBuff->uGuestRamPages, pmm->MonitorPmmBuff->uGuestRamPages / 256,
                          (long)(pmm->MonitorPmmBuff->uCurrentPages) / (pmm->MonitorPmmBuff->uGuestRamPages / 100),
                          (long)pmm->stat.rCount,
                          (long)(pmm->MonitorPmmBuff->uCurrentPages / 
                                 ((pmm->stat.rCount <= 0) ? 1 : pmm->stat.rCount)),
                          (void *)pmm->virtualBase);
        up(&pmm->accSem);
    }
    up(&vmGlobal.vmPmms.pmmAccSem);

    return length;
}

static int vmProcStatRead(char *buffer, char **start, off_t offset, int length, int *eof, void *data) {
    return sprintf(buffer, "GenericStatusReport\n");
}

/* 
 * remove proc entry(proc_dir_entry *chld, proc_dir_entry *parent)
 */
static inline void vm_remove_proc_entry(struct proc_dir_entry *chld, struct proc_dir_entry *par) {

    if (chld)
        remove_proc_entry(chld->name, par);
}

    struct list_head *q;
    struct pmm_struct_t *pmm;

/* init/deinit proc subsystem */
int vmProcInit(void) {

    int i;

    if (!(vmProcRoot = proc_mkdir(ROOT_DIR, &proc_root))) {
        Err("Failed to register proc root at %s", ROOT_DIR);
        return -1;
    }

    for(i = 0; i < PROC_ENTRY_MAX; i++) {
        if (!(vmProcEntries[i].entry = create_proc_read_entry(vmProcEntries[i].name, 
                                                              0, vmProcRoot, vmProcEntries[i].read, NULL))) {
            Err("Failed to register read entry at %s/%s", ROOT_DIR, vmProcEntries[i].name);
            goto err;
        }
    }

    return 0;

err:

    for(; i > 0; i--) {
        vm_remove_proc_entry(vmProcEntries[i].entry, vmProcRoot);
    }

    vm_remove_proc_entry(vmProcRoot, &proc_root);

    return -1;
}

void vmProcDeinit(void) {

    int i;

    for(i = 0; i < PROC_ENTRY_MAX; i++) {
        vm_remove_proc_entry(vmProcEntries[i].entry, vmProcRoot);
    }
    vm_remove_proc_entry(vmProcRoot, &proc_root);
}
