//
// Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.
//
// Author: Sergey Viktorov sergeyv@parallels.ru
//
// $Id: prlnet.c,v 1.6 2006/07/21 09:05:36 sergeyv Exp $
//
#include <linux/autoconf.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0)
#error "Kernels older than 2.2 are not supported"
#endif

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/list.h>
#include <linux/string.h>
#include <linux/capability.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/poll.h>

#include <linux/miscdevice.h>

#include <linux/skbuff.h>
#include <linux/if_ether.h>
#include <linux/netdevice.h>
#include <linux/wireless.h> /* defines WIRELESS_EXT */
#include <linux/etherdevice.h>
#include <linux/ip.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <net/sock.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,14)
#define net_device device
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,8)
#define netif_rx_ni(skb) netif_rx(skb)
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,4)
#include <asm/checksum.h>
#endif

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/pgtable.h>

#include "prlnet.h"


//
// from ApiNet.h
//
// compatibility crap
typedef unsigned char UCHAR, *PUCHAR;
typedef unsigned short USHORT;
typedef unsigned int UINT;

#define IO_NET_SIZE (256*1024)

typedef struct _NET_BUFFER
{
	// Virtual cable is plugged
	UINT	bCableUnplugged;
	// Network adapter able to receive packets
	UINT	bNetworkAdapterActive;
	// statistics
	UINT	uRecOverflows;
	UINT	uSendOverflows;

	// Head of buffer (first allocated packet)
	UINT	uHead;
	// Tail of buffer
	UINT	uTail;
	// Packet(s) present in buffer
	UINT	bRecPacketPresent;
	// statistics
	UINT	uRecPackets;

	// Head of buffer (first allocated packet)
	UINT	uSendHead;
	// Tail of buffer
	UINT	uSendTail;
	// Packet(s) present in send buffer
	UINT	bSendPacketPresent;
	// statistics
	UINT	uSentPackets;

	// align to 1Kb boundary here
	UCHAR	pad[1024 - 48];

	// data buffers allow 1.5Kb tailroom to ease wraparound
	UCHAR	aBuffer[IO_NET_SIZE + 1536];
	UCHAR	aSendBuffer[IO_NET_SIZE + 1536];
} __attribute__((packed)) NET_BUFFER;
typedef NET_BUFFER *PNET_BUFFER;

struct AssertIoNetSize {
	char assert[IO_NET_SIZE & (IO_NET_SIZE-1) ? -1 : 1];
};
struct AssertNetBufferStructure {
	char assert[sizeof(NET_BUFFER) != 4096 + IO_NET_SIZE * 2 ? -1 : 1];
};

typedef struct _NET_PACKET
{
	// eth header happens to be 14 bytes long
	// align IP payload on 16 byte boundary
	USHORT	uDataLen;
	// Packed data start
	UCHAR	aDataBuf[1];
} __attribute__((packed)) NET_PACKET;

struct AssertIPHeaderAlignment {
	char assert[(sizeof(NET_PACKET) - 1 + 14) & 15 ? -1 : 1];
};

// align packets to 16 byte boundary
#define FULL_PACKET_SIZE(uPacketDataLen) \
	((sizeof(NET_PACKET) - 1 + (uPacketDataLen) + 15) & ~15)


#ifndef structof
#define structof(p, s, f) ((s *)((char *)(p) - (int)&((s *)0)->f))
#endif
#ifndef lengthof
#define lengthof(a) (sizeof(a) / sizeof(*a))
#endif

#define HTONS(x) ((unsigned short)((x) << 8) | ((unsigned short)(x) >> 8))

//
// locking utilities
//
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
typedef struct wait_queue wait_queue_t, *wait_queue_head_t;

static __inline void
init_waitqueue_head(struct wait_queue **wq)
{
	*wq = 0;
}
static __inline void
init_waitqueue_entry(struct wait_queue *wq, struct task_struct *task)
{
	wq->task = task;
}
#endif

static __inline unsigned
prlnet_xadd(volatile unsigned *dst, unsigned delta)
{
	unsigned oldval;
	__asm__ __volatile__ (
		"lock xadd %0, %2\n\t"
		: "=a"(oldval)
		: "0"(delta), "m"(*dst)
		: "memory"); // compiler memory barrier
	return oldval;
}

static __inline unsigned
prlnet_swp(volatile unsigned *dst, unsigned swap)
{
	unsigned oldval;
	__asm__ __volatile__ (
		"xchg %2, %1\n\t"
		: "=a"(oldval)
		: "m"(*dst), "r"(swap)
		: "memory"); // compiler memory barrier
	return oldval;
}

static __inline unsigned
prlnet_cas(volatile unsigned *dst, unsigned compare, unsigned swap)
{
	unsigned oldval;
	__asm__ __volatile__ (
		"lock cmpxchg %3, %2\n\t"
		: "=a"(oldval)
		: "0"(compare), "m"(*dst), "r"(swap)
		: "memory"); // compiler memory barrier
	return oldval;
}

static __inline unsigned long long
prlnet_read64(volatile unsigned long long *src)
{
	unsigned long long val;
	__asm__ __volatile__ (
		"movl %%eax, %%ebx\n\t"
		"movl %%edx, %%ecx\n\t"
		"lock cmpxchg8b %1\n\t"
		: "=&A"(val)
		: "m"(*src)
		: "ebx", "ecx");
	return val;
}

static __inline unsigned long long
prlnet_cas64(volatile unsigned long long *dst,
	unsigned long long compare, unsigned long long swap)
{
	unsigned long long oldval;
	__asm__ __volatile__ (
		"lock cmpxchg8b %2\n\t"
		: "=A"(oldval)
		: "0"(compare), "m"(*dst), "b"((unsigned)swap), "c"((unsigned)(swap >> 32))
		: "memory"); // compiler memory barrier
	return oldval;
}


typedef struct prlnet_rwlock {
	volatile unsigned lock;
#define PRLNET_LOCK_HAS_WAITQUEUE 1
#define PRLNET_LOCK_WRITE_LOCKED 2
#define PRLNET_LOCK_ADD_READER 4
	wait_queue_head_t wq;
} prlnet_rwlock_t;

static __inline void
prlnet_rwlock_init(prlnet_rwlock_t *l)
{
	l->lock = 0;
	init_waitqueue_head(&l->wq);
}

// out-of-line slow part
static int
prlnet_rwlock_wait(prlnet_rwlock_t *l, unsigned prev)
{
	wait_queue_t wait;
	unsigned curr;

	init_waitqueue_entry(&wait, current);
	current->state = TASK_INTERRUPTIBLE;
	add_wait_queue(&l->wq, &wait);
	curr = prlnet_cas(
		&l->lock, prev, prev | PRLNET_LOCK_HAS_WAITQUEUE);
	if (curr != prev) {
		current->state = TASK_RUNNING;
		remove_wait_queue(&l->wq, &wait);
		return 0;
	}
	schedule();
	remove_wait_queue(&l->wq, &wait);
	if (signal_pending(current))
		return -ERESTARTSYS;
	return 1;
}

static __inline int
prlnet_read_lock(prlnet_rwlock_t *l)
{
	unsigned prev, curr;

	for (;;) {
		prev = l->lock;
		for (; !(prev & PRLNET_LOCK_WRITE_LOCKED); ) {
			curr = prlnet_cas(
				&l->lock, prev, prev + PRLNET_LOCK_ADD_READER);
			if (curr == prev)
				return 0;
			prev = curr;
		}
		if (prlnet_rwlock_wait(l, prev) < 0)
			return -ERESTARTSYS;
	}
}

static __inline void
prlnet_read_unlock(prlnet_rwlock_t *l)
{
	unsigned prev, curr, next;

	prev = l->lock;
	for (;;) {
//		if ((prev & PRLNET_LOCK_WRITE_LOCKED) || !(prev >> 2)) {
//			BUG();
//		}
		next = prev - PRLNET_LOCK_ADD_READER;
		if (next == PRLNET_LOCK_HAS_WAITQUEUE)
			next = 0;
		curr = prlnet_cas(&l->lock, prev, next);
		if (curr == prev) {
			if (curr == (PRLNET_LOCK_HAS_WAITQUEUE
					| PRLNET_LOCK_ADD_READER))
				wake_up(&l->wq);
			break;
		}
		prev = curr;
	}
}

static __inline int
prlnet_write_lock(prlnet_rwlock_t *l)
{
	unsigned prev, curr;

	for (;;) {
		prev = l->lock;
		for (; prev == 0; ) {
			curr = prlnet_cas(
				&l->lock, prev, PRLNET_LOCK_WRITE_LOCKED);
			if (curr == prev)
				return 0;
			prev = curr;
		}
		if (prlnet_rwlock_wait(l, prev) < 0)
			return -ERESTARTSYS;
	}
}

static __inline void
prlnet_write_unlock(prlnet_rwlock_t *l)
{
	unsigned prev, curr;

	prev = l->lock;
	for (;;) {
//		if (!(prev & PRLNET_LOCK_WRITE_LOCKED) || (prev >> 2)) {
//			BUG();
//		}
		curr = prlnet_cas(&l->lock, prev, 0);
		if (curr == prev) {
			if (curr == (PRLNET_LOCK_HAS_WAITQUEUE
					| PRLNET_LOCK_WRITE_LOCKED))
				wake_up(&l->wq);
			break;
		}
		prev = curr;
	}
}

//
// networking protocols
// these could have been taken from BSD headers but
// we'd better have the same definitions for all platforms
//
#define H2NS(x) \
	((unsigned short)((x) << 8) | ((unsigned short)(x) >> 8))

typedef struct eth_header_s {
	unsigned char dst[6];
	unsigned char src[6];
	unsigned short protocol;
#define ETH_PROTOCOL_IPV4 0x800
#define ETH_PROTOCOL_ARP 0x806
} __attribute__ ((packed)) eth_header_t;

typedef struct arp_header_s {
	unsigned short hwspace;
#define ARP_HWSPACE_ETH 1
#define ARP_HWSPACE_802 6
	unsigned short ptspace;
#define ARP_PTSPACE_ETH_IPV4 ETH_PROTOCOL_IPV4
	unsigned char hwaddrlen;
	unsigned char ptaddrlen;
	unsigned short opcode;
#define ARP_OPCODE_REQUEST 1
#define ARP_OPCODE_REPLY 2
// RARP opcodes
#define ARP_OPCODE_RREQUEST 3
#define ARP_OPCODE_RREPLY 4
	// srchwaddr
	// srcptaddr
	// dsthwaddr
	// dstptaddr
} __attribute__ ((packed)) arp_header_t;

typedef struct arp_eth_ipv4_s {
	arp_header_t head;
	unsigned char srchwaddr[6];
	unsigned srcptaddr;
	unsigned char dsthwaddr[6];
	unsigned dstptaddr;
} __attribute__ ((packed)) arp_eth_ipv4_t;

typedef struct ipv4_header_s {
	unsigned char version_headerlen;
#define IPV4_VERSION(h) \
	((h)->version_headerlen >> 4)
#define IPV4_HEADERLEN(h) \
	(((h)->version_headerlen & 0x0f) * sizeof(unsigned))
	unsigned char tos;
	unsigned short length;
	unsigned short id;
	unsigned short frag;
#define IPV4_FRAG_DF 0x4000
#define IPV4_FRAG_MF 0x2000
	unsigned char ttl;
	unsigned char protocol;
#define IPV4_PROTOCOL_TCP 6
#define IPV4_PROTOCOL_UDP 17
	unsigned short cksum;
	unsigned srcaddr;
	unsigned dstaddr;
	// options
	// payload
} __attribute__ ((packed)) ipv4_header_t;

typedef struct udp_header_s {
	unsigned short srcport;
	unsigned short dstport;
	unsigned short length;
	unsigned short cksum;
	// payload
} __attribute__ ((packed)) udp_header_t;

typedef struct tcp_header_s {
	unsigned short srcport;
	unsigned short dstport;
	unsigned seq;
	unsigned ack;
	unsigned char dataoff;
#define TCP_DATAOFF(h) \
	(((h)->dataoff >> 4) * sizeof(unsigned))
	unsigned char flags;
#define TCP_FLAGS_FIN 1
#define TCP_FLAGS_SYN 2
#define TCP_FLAGS_RST 4
#define TCP_FLAGS_PSH 8
#define TCP_FLAGS_ACK 16
#define TCP_FLAGS_URG 32
	unsigned short window;
	unsigned short cksum;
	unsigned short urgent;
	// options
	// payload
} __attribute__ ((packed)) tcp_header_t;


#define UDP_PORT_DHCPSRV 67
#define UDP_PORT_DHCPCLI 68

typedef struct dhcp_header_s {
	unsigned char op;
#define DHCP_OP_REQUEST 1
#define DHCP_OP_REPLY 2
	unsigned char htype;
#define DHCP_HTYPE_ETH 1
	unsigned char hlen;
	unsigned char hops;
	unsigned xid;
	unsigned short secs;
	unsigned short flags;
#define DHCP_FLAGS_BCAST 0x8000
	unsigned ciaddr;
	unsigned yiaddr;
	unsigned siaddr;
	unsigned giaddr;
	unsigned char chaddr[16];
	char sname[64];
	char file[128];
// allow BOOTP patching too
//	unsigned magic;
//#define DHCP_MAGIC 0x63538263
} __attribute__ ((packed)) dhcp_header_t;

#define DHCP_OPT_PAD 0
#define DHCP_OPT_END 255
#define DHCP_OPT_CLIENTID 61


//
// internal structures
//
// NET_BUFFERs heads/tails are misnamed, follows the convention
typedef union shadow_tail_u {
	struct {
		unsigned tail;
		unsigned contenders;
	} s;
	unsigned long long u;
} shadow_tail_t;

typedef struct virtual_if {
	struct list_head by_hw;
	struct hardware_if *hw;
	char mac[6];
	unsigned ip[4];
	int iip;
	volatile unsigned sendloop;
#define PRLNET_SENDLOOP_STOPPED 0
#define PRLNET_SENDLOOP_RUNNING 1
#define PRLNET_SENDLOOP_RESTART 2
	shadow_tail_t tail;
	NET_BUFFER *netbuf;
	wait_queue_head_t rdq;
} virtual_if_t;

typedef struct hardware_if {
	struct list_head list;
	char ifname[16];
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct sock *sk;
#endif
	struct packet_type proto;
//	struct notifier_block notifier;
	char mac[6];
	short wireless;
	struct list_head vis;
	unsigned logidx;
	void *skblog[512];
} hardware_if_t;

static struct list_head hardware_ifs;

//
// this read-write lock protects our if lists.
// clients may sleep, so we can't just use kernel rw(spin)locks
// global lock is ok since it's written very rarely
//
static prlnet_rwlock_t prlnet_lock;

//
// IP to vi mapping for wireless bridging
//
static __inline virtual_if_t*
get_vi_addr(hardware_if_t *hw, unsigned ip)
{
	struct list_head *l;

	for (l = hw->vis.next; l != &hw->vis; l = l->next) {
		virtual_if_t *vi = structof(l, virtual_if_t, by_hw);

		if (vi->ip[0] == ip || vi->ip[1] == ip
				|| vi->ip[2] == ip || vi->ip[3] == ip)
			return vi;
	}
	return 0;
}

static __inline void
set_vi_addr(hardware_if_t *hw, virtual_if_t *vi, unsigned ip)
{
	if (vi->ip[vi->iip & 3] != ip) {
		struct list_head *l;

		for (l = hw->vis.next; l != &hw->vis; l = l->next) {
			virtual_if_t *v = structof(l, virtual_if_t, by_hw);

			if (v->ip[0] == ip) v->ip[0] = 0;
			if (v->ip[1] == ip) v->ip[1] = 0;
			if (v->ip[2] == ip) v->ip[2] = 0;
			if (v->ip[3] == ip) v->ip[3] = 0;
		}
		vi->ip[++vi->iip & 3] = ip;
	}
}

static __inline int
maccmp(const void *l, const void *r)
{
	// IA32 only
	return *(unsigned *)l != *(unsigned *)r
		|| ((unsigned short *)l)[2] != ((unsigned short *)r)[2];
}

//
// packet handling
//
static int
vi_recv(virtual_if_t *vi, const char *buf, int len, const char *mac)
{
	NET_BUFFER *netbuf = vi->netbuf;
	NET_PACKET *pkt;
	shadow_tail_t prev, next, curr;

	if (!netbuf->bNetworkAdapterActive || netbuf->bCableUnplugged)
		return -1;
	if (len > 1514)
		return -2;
	// shadow_tail provides concurrency safety
	curr.u = prlnet_read64(&vi->tail.u);
	for (;;) {
		next.s.tail = curr.s.tail;
		next.s.contenders = curr.s.contenders + 1;
		prev.u = prlnet_cas64(&vi->tail.u, curr.u, next.u);
		if (prev.u == curr.u)
			break;
		curr.u = prev.u;
	}
	curr.u = next.u;
	pkt = 0;
	for (;;) {
		next.s.tail = curr.s.tail + FULL_PACKET_SIZE(len);
		if (next.s.tail - netbuf->uHead > IO_NET_SIZE)
			goto commit;
		next.s.contenders = curr.s.contenders;
		prev.u = prlnet_cas64(&vi->tail.u, curr.u, next.u);
		if (prev.u == curr.u)
			break;
		curr.u = prev.u;
	}
	pkt = (NET_PACKET *)(netbuf->aBuffer + (curr.s.tail & (IO_NET_SIZE-1)));
	pkt->uDataLen = len;
	memcpy(pkt->aDataBuf, buf, len);
	if (mac) {
		eth_header_t *eth = (eth_header_t *)pkt->aDataBuf;
		arp_eth_ipv4_t *arp = (arp_eth_ipv4_t *)(eth + 1);

		memcpy(eth->dst, mac, sizeof(eth->dst));
		if (eth->protocol == H2NS(ETH_PROTOCOL_ARP)
				&& len >= sizeof(*eth) + sizeof(*arp)
				&& (arp->head.hwspace == H2NS(ARP_HWSPACE_ETH)
					|| arp->head.hwspace == H2NS(ARP_HWSPACE_802))
				&& arp->head.ptspace == H2NS(ARP_PTSPACE_ETH_IPV4)
				&& arp->head.hwaddrlen == sizeof(arp->srchwaddr)
				&& arp->head.ptaddrlen == sizeof(arp->srcptaddr)) {
			if (maccmp(arp->dsthwaddr, vi->hw->mac) == 0)
				memcpy(arp->dsthwaddr, mac, sizeof(arp->dsthwaddr));
		}
	}
	curr.u = next.u;
commit:
	for (;;) {
		next.s.tail = curr.s.tail;
		next.s.contenders = curr.s.contenders - 1;
		if (next.s.contenders == 0) {
			*(volatile unsigned *)&netbuf->uTail = curr.s.tail;
			*(volatile unsigned *)&netbuf->bRecPacketPresent = 1;
		}
		prev.u = prlnet_cas64(&vi->tail.u, curr.u, next.u);
		if (prev.u == curr.u)
			break;
		curr.u = prev.u;
	}
	// wake readers (incurs some unavoidable? spinlock overhead)
	if (next.s.contenders == 0)
		wake_up_interruptible(&vi->rdq);
	return pkt ? 0 : -3;
}

static int
hw_recv(hardware_if_t *hw, const char *buf, int len)
{
	struct list_head *l;
	eth_header_t *eth = (eth_header_t *)buf;

	// deduce mac address from the packet
	if (hw->wireless) {
		arp_eth_ipv4_t *arp = (arp_eth_ipv4_t *)(eth + 1);
		unsigned srcaddr = 0, dstaddr = 0;

		if (eth->protocol == H2NS(ETH_PROTOCOL_IPV4)) {
			ipv4_header_t *ipv4 = (ipv4_header_t *)(eth + 1);
			if (len < sizeof(*eth) + sizeof(*ipv4))
				return -3;
			srcaddr = ipv4->srcaddr;
			dstaddr = ipv4->dstaddr;
		}
		else if (eth->protocol == H2NS(ETH_PROTOCOL_ARP)) {
			if (len >= sizeof(*eth) + sizeof(*arp)
					&& (arp->head.hwspace == H2NS(ARP_HWSPACE_ETH)
						|| arp->head.hwspace == H2NS(ARP_HWSPACE_802))
					&& arp->head.ptspace == H2NS(ARP_PTSPACE_ETH_IPV4)
					&& arp->head.hwaddrlen == sizeof(arp->srchwaddr)
					&& arp->head.ptaddrlen == sizeof(arp->srcptaddr)) {
				srcaddr = arp->srcptaddr;
				dstaddr = arp->dstptaddr;
			}
		}
		if (maccmp(eth->src, hw->mac) == 0) {
			// ignore loopback packets
			if (srcaddr && get_vi_addr(hw, srcaddr))
				return -2;
		}
		if (maccmp(eth->dst, hw->mac) == 0) {
			virtual_if_t *vi;

			if (dstaddr == 0 || dstaddr == 0xffffffff) {
				// pseudo broadcast
				for (l = hw->vis.next; l != &hw->vis; l = l->next) {
					virtual_if_t *vi = structof(l, virtual_if_t, by_hw);

					vi_recv(vi, buf, len, vi->mac);
				}
				return 2;
			}
			if (!(vi = get_vi_addr(hw, dstaddr)))
				return -1;
			vi_recv(vi, buf, len, vi->mac);
			return 0;
		}
	}
	if (eth->dst[0] && (eth->dst[0] & 0x30) != 0x20) {
		// broadcasts/multicasts
		for (l = hw->vis.next; l != &hw->vis; l = l->next) {
			virtual_if_t *vi = structof(l, virtual_if_t, by_hw);

			if (maccmp(eth->src, vi->mac) != 0)
				vi_recv(vi, buf, len, 0);
		}
		return 0;
	}
	for (l = hw->vis.next; l != &hw->vis; l = l->next) {
		virtual_if_t *vi = structof(l, virtual_if_t, by_hw);

		if (maccmp(eth->dst, vi->mac) == 0) {
			vi_recv(vi, buf, len, 0);
			return 1;
		}
	}
	return -1;
}


static __inline struct sk_buff*
hw_alloc_skb(hardware_if_t *hw, const char *buf, int len)
{
	// len includes ethernet header (+ eth payload alignment)
	struct sk_buff *skb = dev_alloc_skb(len + 2);
	if (!skb)
		return 0;
	// pad to align ethernet payload on 16 bytes
	skb_reserve(skb, 2);
	memcpy(skb_put(skb, len), buf, len);
	skb->dev = hw->proto.dev;
	skb->mac.raw = skb->data;
	skb->nh.raw = skb->data + sizeof(skb->mac);
	// prepare to filter our own packets
	hw->skblog[
		*(unsigned short *)skb->head =
			 prlnet_xadd(&hw->logidx, 1) % lengthof(hw->skblog)] =
				skb->head;
	return skb;
}

static int
hw_send(hardware_if_t *hw, virtual_if_t *visrc, const char *buf, int len)
{
	struct list_head *l;
	struct sk_buff *skbout, *skbin;
	eth_header_t *eth = (eth_header_t *)buf;

// latest madwifi drivers advertise a header of 88
//	if (len > hw->proto.dev->mtu + hw->proto.dev->hard_header_len
//			|| len < hw->proto.dev->hard_header_len)
	if (len > 1500 + sizeof(*eth) || len < sizeof(*eth))
		return -1;
	// broadcasts/multicasts
	if (eth->dst[0] && (eth->dst[0] & 0x30) != 0x20) {
		for (l = hw->vis.next; l != &hw->vis; l = l->next) {
			virtual_if_t *vi = structof(l, virtual_if_t, by_hw);

			if (vi != visrc)
				vi_recv(vi, buf, len, 0);
		}
	}
	else {
		for (l = hw->vis.next; l != &hw->vis; l = l->next) {
			virtual_if_t *vi = structof(l, virtual_if_t, by_hw);

			if (maccmp(eth->dst, vi->mac) == 0) {
				vi_recv(vi, buf, len, 0);
				return 0;
			}
		}
	}
	skbout = hw_alloc_skb(hw, buf, len);
	if (!skbout)
		return -1;
	// send broadcasts both up and down
	if (eth->dst[0] && (eth->dst[0] & 0x30) != 0x20)
		skbin = hw_alloc_skb(hw, buf, len);
	else if (maccmp(eth->dst, hw->mac) == 0) {
		skbin = skbout;
		skbout = 0;
	}
	else
		skbin = 0;
	if (skbin) {
		skbin->protocol = eth_type_trans(skbin, skbin->dev);
		netif_rx_ni(skbin);
	}
	if (skbout && hw->wireless) {
		eth = (eth_header_t *)skbout->mac.raw;
		memcpy(eth->src, hw->mac, sizeof(eth->src));
		if (eth->protocol == H2NS(ETH_PROTOCOL_IPV4)) {
			ipv4_header_t *ipv4 = (ipv4_header_t *)(eth + 1);
			udp_header_t *udp;
			dhcp_header_t *dhcp;

			if (len < sizeof(*eth) + sizeof(*ipv4))
				goto output;
			// notice VM IP address advertisement
			if (ipv4->srcaddr != 0 && ipv4->srcaddr != 0xffffffff)
				set_vi_addr(hw, visrc, ipv4->srcaddr);
			if (ipv4->protocol != IPV4_PROTOCOL_UDP
					|| (ipv4->frag & ~H2NS(IPV4_FRAG_DF)) != 0
					|| len < sizeof(*eth) + IPV4_HEADERLEN(ipv4)
						+ sizeof(*udp) + sizeof(*dhcp))
				goto output;
			udp = (udp_header_t *)((char *)ipv4 + IPV4_HEADERLEN(ipv4));
			dhcp = (dhcp_header_t *)(udp + 1);
			if (udp->dstport == H2NS(UDP_PORT_DHCPSRV)
					&& dhcp->op == DHCP_OP_REQUEST
					&& dhcp->htype == DHCP_HTYPE_ETH
					&& dhcp->hlen == sizeof(eth->src)
					// allow BOOTP patching too
					//&& dhcp->magic == DHCP_MAGIC
					&& (dhcp->flags & H2NS(DHCP_FLAGS_BCAST)) == 0) {
				dhcp->flags |= H2NS(DHCP_FLAGS_BCAST);
				// update UDP checksum if present
				if (udp->cksum) {
					unsigned u = (udp->cksum ^ 0xffff) + H2NS(DHCP_FLAGS_BCAST);
					udp->cksum = ~(u + (u >> 16));
				}
			}
		}
		else if (eth->protocol == H2NS(ETH_PROTOCOL_ARP)) {
			// ARP reply is sent to hwaddr contained in ARP request
			// patch ARP packets to make things look OK for outside world
			arp_eth_ipv4_t *arp = (arp_eth_ipv4_t *)(eth + 1);

			if (len >= sizeof(*eth) + sizeof(*arp)
					&& (arp->head.hwspace == H2NS(ARP_HWSPACE_ETH)
						|| arp->head.hwspace == H2NS(ARP_HWSPACE_802))
					&& arp->head.ptspace == H2NS(ARP_PTSPACE_ETH_IPV4)
					&& arp->head.hwaddrlen == sizeof(arp->srchwaddr)
					&& arp->head.ptaddrlen == sizeof(arp->srcptaddr)) {
				memcpy(arp->srchwaddr, hw->mac, sizeof(arp->srchwaddr));
				// notice VM IP address advertisement
				if (arp->srcptaddr != 0 && arp->srcptaddr != 0xffffffff)
					set_vi_addr(hw, visrc, arp->srcptaddr);
			}
		}
	}
	if (skbout) {
output:
		skbout->pkt_type = PACKET_OUTGOING;
		dev_queue_xmit(skbout);
	}
	return 1;
}

static int
vi_send(virtual_if_t *vi)
{
	hardware_if_t *hw = vi->hw;
	NET_BUFFER *netbuf = vi->netbuf;
	UINT head;

restart:
	if (prlnet_swp(&vi->sendloop, PRLNET_SENDLOOP_RUNNING)
			!= PRLNET_SENDLOOP_RESTART) {
		// unexpected
	}
	for (head = netbuf->uSendHead; head != netbuf->uSendTail; ) {
		NET_PACKET *pkt;

		pkt = (NET_PACKET *)(netbuf->aSendBuffer + (head & (IO_NET_SIZE-1)));
		hw_send(hw, vi, pkt->aDataBuf, pkt->uDataLen);
		head += FULL_PACKET_SIZE(pkt->uDataLen);
	}
	netbuf->uSendHead = head;
	if (prlnet_cas(&vi->sendloop, PRLNET_SENDLOOP_RUNNING,
			PRLNET_SENDLOOP_STOPPED) != PRLNET_SENDLOOP_RUNNING)
		goto restart;
	return 0;
}


//
// protocol hook
//
static int
prlnet_recv(struct sk_buff *skb, struct net_device *dev, struct packet_type *pt)
{
	hardware_if_t *hw = structof(pt, hardware_if_t, proto);

	// we rely on dev_add_pack()/dev_remove_pack()
	// to guarantee hw immutability here (including hw->vis list)
	if (skb->pkt_type != PACKET_LOOPBACK) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,4)
		int do_csum = skb->pkt_type == PACKET_OUTGOING
			&& skb->ip_summed == CHECKSUM_HW;
		// this is slightly inefficient
		// applies to host packets only
		if (do_csum || skb_is_nonlinear(skb)) {
			int len = skb->data - skb->mac.raw + skb->len;
			char *buf = kmalloc(2 + len, GFP_ATOMIC);
			if (buf) {
				if (do_csum) {
					// can't use skb_copy_and_csum_dev(skb, buf + 2) here
					// because skb->h == skb->nh. opensource rules!
					int hoff = sizeof(struct ethhdr) + skb->nh.iph->ihl * 4;
					// IP header checksum is done by the kernel
					skb_copy_bits(skb,
						skb->mac.raw - skb->data, buf + 2, hoff);
					// skb->csum is an offset from true skb->h.raw
					*(unsigned short *)(buf + 2 + hoff + skb->csum) =
						csum_fold(
							// kernel does conceptual header too
							skb_copy_and_csum_bits(skb,
								skb->mac.raw + hoff - skb->data,
								buf + 2 + hoff, len - hoff, 0));
				}
				else {
					skb_copy_bits(skb,
						skb->mac.raw - skb->data, buf + 2, len);
				}
				hw_recv(hw, buf + 2, len);
			}
			kfree(buf);
			dev_kfree_skb(skb);
			return 0;
		}
#endif
		// filter our own packets
		if (prlnet_cas(
				(unsigned *)(hw->skblog +
					*(unsigned short *)skb->head % lengthof(hw->skblog)),
				(unsigned)skb->head, 0) != (unsigned)skb->head) {
			// undo possibly popped off hard_heder_len
			hw_recv(hw, skb->mac.raw, skb->data - skb->mac.raw + skb->len);
		}
	}
	dev_kfree_skb(skb);
	return 0;
}

//static int
//prlnet_notify(struct notifier_block *nb, unsigned long msg, void *dev)
//{
//	hardware_if_t *hw = structof(nb, hardware_if_t, notifier);
//
//	(void)hw;
//	switch (msg) {
//	case NETDEV_UP:
//	case NETDEV_DOWN:
//	case NETDEV_REGISTER:
//	case NETDEV_UNREGISTER:
//	default:
//		break;
//	}
//	return 0;
//}

// the caller must hold prlnet_lock
static int
hw_bind(hardware_if_t **phw, const char *ifname)
{
	struct list_head *l;
	hardware_if_t *hw;

	// check if we have ifname already bound
	for (l = hardware_ifs.next; l != &hardware_ifs; l = l->next) {
		hw = structof(l, hardware_if_t, list);
		if (strcmp(hw->ifname, ifname) == 0) {
			*phw = hw;
			return 0;
		}
	}
	hw = kmalloc(sizeof(*hw), GFP_KERNEL);
	if (!hw)
		return -ENOMEM;
	memset(hw, 0, sizeof(*hw));
	INIT_LIST_HEAD(&hw->list);
	INIT_LIST_HEAD(&hw->vis);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	hw->sk =
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,5)
		sk_alloc(PF_NETLINK, GFP_KERNEL, 1, 0);
#else
		sk_alloc(0, GFP_KERNEL, 1);
#endif
	if (!hw->sk) {
		kfree(hw);
		return -ENOMEM;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,65)
	sock_valbool_flag(hw->sk, SOCK_DEAD, 1);
#else
	hw->sk->dead = 1;
#endif
	hw->proto.data = hw->sk;
#endif

//	// listen to interface state changes
//	hw->notifier.notifier_call = prlnet_notify;
//	if (register_netdevice_notifier(&hw->notifier) < 0)
//		goto nodev;

	strncpy(hw->ifname, ifname, sizeof(hw->ifname));
	hw->proto.dev =
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,15)
		dev_get_by_name(ifname);
#else
		// this is actually dangerous
		dev_get(ifname);
#endif
	if (!hw->proto.dev) {
//		unregister_netdevice_notifier(&hw->notifier);
//nodev:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		sk_free(hw->sk);
#endif
		kfree(hw);
		return -ENODEV;
	}
	memcpy(hw->mac, hw->proto.dev->dev_addr, sizeof(hw->mac));
#ifdef CONFIG_NET_RADIO
	if (hw->proto.dev->get_wireless_stats)
		hw->wireless = 1;
#if WIRELESS_EXT > 12
	if (hw->proto.dev->wireless_handlers)
		hw->wireless = 1;
#endif
#endif
	if (!hw->wireless)
		dev_set_promiscuity(hw->proto.dev, +1);

	list_add(&hw->list, &hardware_ifs);
	// regiter protocol
	hw->proto.func = prlnet_recv;
	hw->proto.type = HTONS(ETH_P_ALL);
	dev_add_pack(&hw->proto);
	*phw = hw;
	return 0;
}

// the caller must hold prlnet_lock
static void
hw_unbind(hardware_if_t *hw)
{
	dev_remove_pack(&hw->proto);
	if (!hw->wireless)
		dev_set_promiscuity(hw->proto.dev, -1);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,15)
	dev_put(hw->proto.dev);
#endif
	hw->proto.dev = 0;
//	unregister_netdevice_notifier(&hw->notifier);
	list_del(&hw->list);
	INIT_LIST_HEAD(&hw->list);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	sk_free(hw->sk);
#endif
	kfree(hw);
}


//
// file operations
//
static int
prlnet_open(struct inode *inode, struct file *filp)
{
	virtual_if_t *vi;

	vi = kmalloc(sizeof(*vi), GFP_USER);
	if (!vi)
		return -ENOMEM;
	memset(vi, 0, sizeof(*vi));
	INIT_LIST_HEAD(&vi->by_hw);
	init_waitqueue_head(&vi->rdq);
	vi->netbuf = vmalloc(
		(sizeof(*vi->netbuf) + PAGE_SIZE-1) & ~(PAGE_SIZE-1));
	if (!vi->netbuf) {
		kfree(vi);
		return -ENOMEM;
	}
	memset(vi->netbuf, 0,
		(sizeof(*vi->netbuf) + PAGE_SIZE-1) & ~(PAGE_SIZE-1));
	filp->private_data = vi;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	MOD_INC_USE_COUNT;
#endif
	return 0;
}

static int
prlnet_release(struct inode *inode, struct file *filp)
{
	virtual_if_t *vi = filp->private_data;

	filp->private_data = 0;
	// release() can't be interrupted
	for (; prlnet_write_lock(&prlnet_lock) < 0; );
	if (vi->hw) {
		dev_remove_pack(&vi->hw->proto);
		list_del(&vi->by_hw);
		dev_add_pack(&vi->hw->proto);
		if (list_empty(&vi->hw->vis))
			hw_unbind(vi->hw);
	}
	prlnet_write_unlock(&prlnet_lock);
	vfree(vi->netbuf);
	kfree(vi);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}

static int
prlnet_ioctl(struct inode *inode, struct file *filp,
	unsigned int cmd, unsigned long arg)
{
	virtual_if_t *vi = filp->private_data;
	hardware_if_t *hw;
	prlnet_bind_t bind;
	int r;

	switch (cmd) {
	case PRLNET_IOCBIND:
		r = copy_from_user(&bind, (void *)arg, sizeof(bind));
		if (r)
			return r;
		// null terminate ifname
		bind.ifname[sizeof(bind.ifname) - 1] = '\0';

		r = prlnet_write_lock(&prlnet_lock);
		if (r < 0)
			return r;
		r = hw_bind(&hw, bind.ifname);
		if (r >= 0 && vi->hw != hw) {
			if (vi->hw) {
				dev_remove_pack(&vi->hw->proto);
				list_del(&vi->by_hw);
				dev_add_pack(&vi->hw->proto);
				if (list_empty(&vi->hw->vis))
					hw_unbind(vi->hw);
			}
			memcpy(vi->mac, bind.mac, sizeof(vi->mac));
			vi->hw = hw;
			// disable packet reception while we modify hw->vis list
			dev_remove_pack(&hw->proto);
			list_add(&vi->by_hw, &hw->vis);
			dev_add_pack(&hw->proto);
		}
		prlnet_write_unlock(&prlnet_lock);
		if (r < 0)
			return r;
		break;
	case PRLNET_IOCKICKSEND:
		if (!vi->hw)
			return -EINVAL;
		r = prlnet_read_lock(&prlnet_lock);
		if (r < 0)
			return r;
		if (prlnet_swp(&vi->sendloop, PRLNET_SENDLOOP_RESTART)
				== PRLNET_SENDLOOP_STOPPED)
			vi_send(vi);
		prlnet_read_unlock(&prlnet_lock);
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

// Oh, holy compatibility crap...
static struct page*
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,1)
prlnet_nopage(struct vm_area_struct *vma, unsigned long addr, int *type)
#else
prlnet_nopage(struct vm_area_struct *vma, unsigned long addr, int write)
#endif
{
	virtual_if_t *vi = vma->vm_private_data;
#ifdef NOPAGE_SIGBUS
	struct page *page = NOPAGE_SIGBUS;
#else
	struct page *page = 0;
#endif
	unsigned long offset = addr - vma->vm_start
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,26)
		+ (vma->vm_pgoff << PAGE_SHIFT);
#else
		+ vma->vm_offset;
#endif
#ifdef pte_offset
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *pte;
#endif

	if (offset >= ((sizeof(*vi->netbuf) + PAGE_SIZE-1) & ~(PAGE_SIZE-1)))
		return page;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,1)
	if (type)
		*type = VM_FAULT_MINOR;
#endif
#ifdef pte_offset
	addr = (unsigned long)vi->netbuf + offset;
	spin_lock(&init_mm.page_table_lock);
	pgd = pgd_offset(&init_mm, addr);
	pmd = pmd_offset(pgd, addr);
	pte = pte_offset(pmd, addr);
	page = pte_page(*pte);
	spin_unlock(&init_mm.page_table_lock);
#else
	page = vmalloc_to_page((char *)vi->netbuf + offset);
#endif
	get_page(page);
	return page;
}

static struct vm_operations_struct prlnet_vmops = {
	.nopage = prlnet_nopage
};

static int
prlnet_mmap(struct file *filp, struct vm_area_struct *vma)
{
	virtual_if_t *vi = filp->private_data;

	// only allow mmaping full netbuf
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,26)
	if (vma->vm_pgoff
#else
	if (vma->vm_offset
#endif
		|| vma->vm_end - vma->vm_start
			!= ((sizeof(*vi->netbuf) + PAGE_SIZE-1) & ~(PAGE_SIZE-1)))
		return -EINVAL;
#ifdef VM_RESERVED
	vma->vm_flags |= VM_RESERVED | VM_SHARED;
#endif
	vma->vm_ops = &prlnet_vmops;
	vma->vm_private_data = vi;
	return 0;
}

static unsigned int
prlnet_poll(struct file *filp, poll_table *wait)
{
	virtual_if_t *vi = filp->private_data;

	if (!vi->hw)
		return POLLERR;
	poll_wait(filp, &vi->rdq, wait);
	if (vi->netbuf->bRecPacketPresent)
		return POLLIN | POLLRDNORM;
	return 0;
}

static ssize_t
prlnet_read(struct file *filp, char *ubuf, size_t len, loff_t *off)
{
	virtual_if_t *vi = filp->private_data;
	wait_queue_t wait;
	NET_BUFFER *netbuf = vi->netbuf;
	UINT head;
	int r;

	if (!vi->hw)
		return -EINVAL;
	// XXX eliminate bNetworkAdapterActive
	netbuf->bNetworkAdapterActive = 1;
	init_waitqueue_entry(&wait, current);
	add_wait_queue(&vi->rdq, &wait);
	for (;;) {
		current->state = TASK_INTERRUPTIBLE;
		for (head = netbuf->uHead; head != netbuf->uTail; ) {
			NET_PACKET *pkt;

			pkt = (NET_PACKET *)(netbuf->aBuffer + (head & (IO_NET_SIZE-1)));
			current->state = TASK_RUNNING;
			remove_wait_queue(&vi->rdq, &wait);
			if (pkt->uDataLen > len)
				return -EMSGSIZE;
			r = copy_to_user(ubuf, pkt->aDataBuf, pkt->uDataLen);
			if (r < 0)
				return r;
			r = pkt->uDataLen;
			head += FULL_PACKET_SIZE(pkt->uDataLen);
			*(volatile UINT *)&netbuf->uHead = head;
			return r;
		}
		if (filp->f_flags & O_NONBLOCK) {
			remove_wait_queue(&vi->rdq, &wait);
			return -EAGAIN;
		}
		schedule();
		if (signal_pending(current)) {
			remove_wait_queue(&vi->rdq, &wait);
			return -ERESTARTSYS;
		}
	}
}

static ssize_t
prlnet_write(struct file *filp, const char *ubuf, size_t len, loff_t *off)
{
	virtual_if_t *vi = filp->private_data;
	char *buf;
	int r;

	if (!vi->hw)
		return -EINVAL;
	if (len > 1514)
		return -EMSGSIZE;
	// ok, don't need top efficiency for dhcpd
	buf = kmalloc(len, GFP_USER);
	if (!buf)
		return -ENOMEM;
	r = copy_from_user(buf, ubuf, len);
	if (r < 0)
		goto out;
	r = prlnet_read_lock(&prlnet_lock);
	if (r < 0)
		goto out;
	hw_send(vi->hw, vi, buf, len);
	r = len;
	prlnet_read_unlock(&prlnet_lock);
out:
	kfree(buf);
	return r;
}

static struct file_operations prlnet_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	.owner = THIS_MODULE,
#endif
	.open = prlnet_open,
	.release = prlnet_release,
	.ioctl = prlnet_ioctl,
	.mmap = prlnet_mmap,
	.poll = prlnet_poll,
	.read = prlnet_read,
	.write = prlnet_write
};
// 0 for dynamic assignment
//static int prlnet_major = 0;

static struct miscdevice prlnet_miscdev = {
#ifdef MISC_DYNAMIC_MINOR
	.minor = MISC_DYNAMIC_MINOR,
#else
	.minor = 250,
#endif
	.name = "vm-bridge",
	.fops = &prlnet_fops
};

int __init
init_module()
{
	int r;

	INIT_LIST_HEAD(&hardware_ifs);
	prlnet_rwlock_init(&prlnet_lock);

//	r = register_chrdev(prlnet_major, "prlnet", &prlnet_fops);
	r = misc_register(&prlnet_miscdev);
	if (r < 0)
		return r;
//	if (!prlnet_major)
//		prlnet_major = r;
	return 0;
}

#ifndef __exit
#define __exit
#endif

void __exit
cleanup_module()
{
//	unregister_chrdev(prlnet_major, "prlnet");
	misc_deregister(&prlnet_miscdev);
}
