/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2005 Parallels Inc. All Rights Reserved.
//		 http://www.parallelssoft.com
//
// MODULE: VirtualNIC.c
//
// DESCRIPTION:
//		Driver for guest-host only networking based on loopback 
//		standard linux driver
//
// DEVELOPER: Tony Zelenoff 
//
/////////////////////////////////////////////////////////////////////////////
#include <linux/autoconf.h>
#include <linux/version.h>

#ifdef CONFIG_MODVERSIONS
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,60)
 #define MODVERSIONS
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
  #include <linux/modversions.h>
 #else
  #include <config/modversions.h>
 #endif
#endif
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/socket.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/in.h>
#include <linux/init.h>
#include <linux/module.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/if_ether.h>	/* For the statistics structure. */
#include <linux/if_arp.h>	/* For ARPHRD_ETHER */

#if 0
extern int hypervisorPresentInSystem(void);
#endif

/////////////////////////////////////////////////////////////////////////////
//
// Constants which required for Linux Driver execution
// (from kernel version 2.1.18)  
//
/////////////////////////////////////////////////////////////////////////////
#ifndef THIS_MODULE
  #define THIS_MODULE (&__this_module)
#endif

#ifdef MODULE_AUTHOR
MODULE_AUTHOR("Parallels");
#endif
#ifdef MODULE_DESCRIPTION
MODULE_DESCRIPTION("Transparen ethernet bridge");
#endif

// Version
//static const char *version = "VMVirtualNIC.c:v1.01 VirtualNIC driver from Parallels Inc. http://www.parallelssoft.com/";

// static member for driver 
struct net_device* dev;

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: VNICOpen
//
// DESCRIPTION:
//              Start virtual card
//
// PARAMETERS:
//		dev  [in] - device to set MAC
//
// RETURN VALUE:
//		0 - successful, < 0 - not successful
//
/////////////////////////////////////////////////////////////////////////////
static int VNICOpen(struct net_device *dev)
{ 
  // Start device
  netif_start_queue(dev);

  return 0; 
};

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: VNICClose
//
// DESCRIPTION:
//              Stop device
//
// PARAMETERS:
//		dev  [in] - device to set MAC
//
// RETURN VALUE:
//		0 - successful, < 0 - not successful
//
/////////////////////////////////////////////////////////////////////////////
static int VNICClose(struct net_device *dev)
{ 
  // Stop device
  netif_stop_queue(dev);
 
  return 0; 
};

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: VNICSetMAC
//
// DESCRIPTION:
//              Set MAC address
//
// PARAMETERS:
//		dev  [in] - device to set MAC
//		pVSA [in] - pointer to socket data
//
// RETURN VALUE:
//		0 - successful, < 0 - not successful
//
/////////////////////////////////////////////////////////////////////////////
static int VNICSetMAC(struct net_device *dev, void *pVSA)
{
   struct sockaddr *pSA = pVSA;

   if ( netif_running(dev) )
      return -EBUSY;

   memcpy(dev->dev_addr, pSA->sa_data, dev->addr_len);

   return 0;
}

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: VNICXmit
//
// DESCRIPTION:
//              Emulate send to physical network
//
// PARAMETERS:
//		skb [in] - network packet
//		dev [in] - device to send to
//
// RETURN VALUE:
//		0 - successful, < 0 - not successful
//
/////////////////////////////////////////////////////////////////////////////
static int VNICXmit(struct sk_buff *skb, struct net_device *dev)
{
	struct net_device_stats *stats = (struct net_device_stats *)dev->priv;

	stats->rx_bytes += skb->len;
	stats->tx_bytes += skb->len;
	stats->rx_packets++;
	stats->tx_packets++;

	kfree_skb(skb);

	netif_start_queue(dev);

	return(0);
}

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: VNICGetStats
//
// DESCRIPTION:
//              Get device statistics
//
// PARAMETERS:
//		dev [in] - statistics for which device
//
// RETURN VALUE:
//		Statistics
//
/////////////////////////////////////////////////////////////////////////////
static struct net_device_stats *VNICGetStats(struct net_device *dev)
{
	return (struct net_device_stats *)dev->priv;
}

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: VNICInit
//
// DESCRIPTION:
//              Initialize module
//
// PARAMETERS:
//
// RETURN VALUE:
//		0 - success, <0 error
//
/////////////////////////////////////////////////////////////////////////////
int __init VNICInit(void)
{
        // MAC address is hardwired to that value as in windows version
        unsigned char station_addr[6] = { 0x00, 0x01, 0x23, 0x45, 0xBE, 0xEF };
#if 0
    hypervisorPresentInSystem(); /* fake call to hypervisor -- ms */
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
        dev = alloc_etherdev(sizeof(struct net_device_stats));
#else
	dev = init_etherdev(0, 0);
	// May be it needed ?
	ether_setup(dev);
	dev->priv = kmalloc(sizeof(struct net_device_stats), GFP_KERNEL);
#endif	        
        // Set name 
	dev_alloc_name(dev, "vnic%d");
	// Set MAC address
	memcpy(dev->dev_addr, station_addr, ETH_ALEN);
	memset(dev->priv, 0, sizeof(struct net_device_stats));

	// Set functions
	dev->hard_start_xmit = &VNICXmit;
	dev->get_stats = &VNICGetStats;
	dev->open = &VNICOpen;
	dev->stop = &VNICClose;
	dev->set_mac_address = &VNICSetMAC;

	SET_MODULE_OWNER(dev);

	// Register network device
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	register_netdev(dev);
#endif

	return(0);
};

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: VNICCleanup
//
// DESCRIPTION:
//              Deinitialize module
//
// PARAMETERS:
//
// RETURN VALUE:
//
/////////////////////////////////////////////////////////////////////////////
static void __exit VNICCleanup(void)
{
	struct net_device_stats* ndstat;

        if (dev)
	{
	   ndstat = dev->priv;

           unregister_netdev(dev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	   free_netdev(dev);
#else
	   kfree(ndstat);
	   dev->priv = NULL;
	   kfree(dev);
#endif
	}
}

// Module Init and Exit procedures
module_init(VNICInit);
module_exit(VNICCleanup);
