/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: device.c                                                    *
 *                                                                     *
 * Description:                                                        *
 *  /dev/hypervisor generic operations                                 *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/miscdevice.h>
#include <linux/spinlock.h>
#include <asm/semaphore.h>

#include "utils.h"

#include "hypervisor.h"

#include "hypidtemul.h"
#include "logging.h"
#include "vtx_init.h"
#include "svm_init.h"

#include "compat_smp.h"

/* device name */
#define DEVICE_NAME "hypervisor"

#ifndef MISC_DYNAMIC_MINOR
# define MISC_DYNAMIC_MINOR 250
#endif

#define HYPERVISOR_MINOR MISC_DYNAMIC_MINOR

/* hypervisor device struct */
static struct miscdevice hypDevice;

/* /dev/hypervisor fops functions */
extern int hypIoctl(struct inode *, struct file *, unsigned int, unsigned long);
int hypOpen(struct inode *, struct file *);
int hypClose(struct inode *, struct file *);

/* file operations for /dev/hypervisor */
static struct file_operations hypFops = {
    .owner =      THIS_MODULE,
    .open =     hypOpen,
    .ioctl =	hypIoctl,
    .release =    hypClose,
};

static int openCount = 0;
static struct semaphore deviceLock;
static int req;
static int unloadBlocked = 0;

/* enable/disable unload */
void hypEnableUnload(void) {

    if (unloadBlocked) {
#ifdef MOD_DEC_USE_COUNT
        MOD_DEC_USE_COUNT;
#else
        module_put(&__this_module);
#endif
    }
}

void hypDisableUnload(void) {

    if (!unloadBlocked) {
#ifdef MOD_INC_USE_COUNT
        MOD_INC_USE_COUNT;
#else
        __module_get(&__this_module);
#endif
    }
}

struct compat_smp_fn_s {
    void QUICKCALL (*fn)(void *arg);
    void *arg;
};

static void hypSmpWrap(void *arg) {
    struct compat_smp_fn_s *fn = arg;
    fn->fn(fn->arg);
}

int vtSupport = 0;

static void hypCheckVmxMode(void) {
    if (is_svm_support()) {
	vtSupport = 1;
    }
    else if (!is_vtx_support()) {
		vtSupport = 0;
    } else {
        if (is_vtx_active() && is_vtx_support_on())
			vtSupport = -1;
        else
			vtSupport = 1;
    }
}

/* open() call */
int hypOpen(struct inode *ino, struct file *fp) {

#ifdef ENABLE_MODULES_PROTECT
    if (check_called_proc()) {
        return -EACCES;
    }
#endif

    down(&deviceLock);

    req = IPI_REGISTER_IDT;

    if (!openCount) {
        if (!unloadBlocked) {
            struct compat_smp_fn_s fn = { hypIpiHandler, &req };

            hypState.initError = 0;

            /* deinit emul flags */
            hypState.idtEmul.inited = 0;

	    /* allocate memory for idts */
	    if (hypAllocateIdt())
		goto err;

            /* register idt */
            if (compat_smp_ipi_call(hypSmpWrap, &fn, 1, 1))
                goto err;

            /* if error on init detected */
            if (hypState.initError)
                goto err;
        }

        /* init emulation */
        hypIdtEmulInit();

	    /* check vmx mode before enable vmx :) */
    	hypCheckVmxMode();

	    /* Initialize for work monitor with VT-x mode. */
    	vtx_support_on();
	
	   /* Initialize for work monitor with SVM mode */
	svm_support_on();
    }

    openCount++;

    up(&deviceLock);

    return 0;


err:
    up(&deviceLock);
    
    return -EFAULT;
}

/* close() call */
int hypClose(struct inode *ino, struct file *fp) {
    
    /* int cpu, record; */

    /* lock device */
    down(&deviceLock);

    req = IPI_DEREGISTER_IDT;
    openCount--;

    if (openCount < 0) {
        openCount = 0;
        goto out;
    }

    if (!openCount) {
    
        /* check for unload */
        if (hypState.idtEmul.blocked) {
            hypDisableUnload();
            unloadBlocked = 1;
            goto out;
        }
        
        if (!unloadBlocked) {
            struct compat_smp_fn_s fn = { hypIpiHandler, &req };

	    /* Deinitialize for work monitor with VT-x mode. */
	    vtx_support_off();
	    
	    /* Deinitialize for work minitor with SVM mode. */
	    svm_support_off();

            if (compat_smp_ipi_call(hypSmpWrap, &fn, 1, 1))
                goto err;
	
	    /* Freeing memory from IDT */
	    hypFreeIdt();
        }
    }
    
out:
    up(&deviceLock);
    return 0;

err:
    up(&deviceLock);
    return -EFAULT;
}

/* init hypervisor device */
int QUICKCALL hypDeviceInit(void) {

    char *devname = DEVICE_NAME;

    /* fill device information struct */
    memset(&hypDevice, 0, sizeof(hypDevice));

    hypDevice.minor = HYPERVISOR_MINOR;
    hypDevice.name = devname;
    hypDevice.fops = &hypFops;

    /* register device */
    if (misc_register(&hypDevice)) {
        Err("Unable to register /dev/%s device", DEVICE_NAME);
        return -ENOENT;
    }

    Log("Device /dev/%s registered with %d.%d", DEVICE_NAME, 10, hypDevice.minor);

    sema_init(&deviceLock, 1);

    return 0;
}

/* deinit hypervisor device */
void QUICKCALL hypDeviceDeinit(void) {
    /* deregister device */
    misc_deregister(&hypDevice);
}

int QUICKCALL hypCountProcessors(void) {

    hypState.cpuCount = 0;

    if (!openCount) {
        struct compat_smp_fn_s fn = { hypIpiHandler, &req };

        req = IPI_COUNT_CPUS;
        if (compat_smp_ipi_call(hypSmpWrap, &fn, 1, 1))
            goto err;
    }
    
    return 0;

err:
    return -EFAULT;
}
