/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: hypmain.c                                                   *
 *                                                                     *
 * Description:                                                        *
 *  Hypervisor main file.                                              *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#define MAIN_MODULE_FILE
#define MODULE_DESCRIPTION_STRING "Parallels hypervisor"

#include "vmmodule.h"

#include <linux/spinlock.h>
#include <asm/errno.h>

#include "mm.h"
#include "hypervisor.h"
#include "hypidtemul.h"
#include "logging.h"

/* applications counter and control stuff */
int appCounter = 0;
spinlock_t appLock = SPIN_LOCK_UNLOCKED;
int driverLocked = 0;

/* fake function for other modules
 * which tell about hypervisor driver is loaded
 * else other modules would be loaded, because symbol
 * will be unresolved -- ms 
 */
int hypervisorPresentInSystem(void) { return 0; }

/* app counter increment */
int appIncCounter(void) {

    if (driverLocked)
        return -EPERM;
    
    spin_lock(&appLock);
    
    appCounter++;
    
    spin_unlock(&appLock);

    return 0;
}

/* app counter decrement */
void appDecCounter(void) {

    spin_lock(&appLock);
    
    if (--appCounter < 0)
        Panic("appCounter is less then zero!");
    
    spin_unlock(&appLock);
}

/* module init function */ 
int init_module(void) {

    /* set internal flags */
    hypState.initError = 0;
    hypState.hypFp = NULL;
    hypState.idtEmul.inited = 0;
    hypState.idtEmul.blocked = 0;
    
    /* allocate flag */
    if (!(hypInHandlerFlag = (unsigned int *)kmalloc(sizeof(unsigned int) * MAX_PROCESSOR_NUMBER, GFP_KERNEL)))
        return -1;
	
    memset( hypInHandlerFlag, 0, sizeof(unsigned int) * MAX_PROCESSOR_NUMBER );
    
    /* init internal data structures */
    if (hypInternalInit())
        return -1;
    
    /* initialize hypervisor device */
    if (hypDeviceInit())
        return -1;
    
    /* initialize vmstates */
    if (hypVmStateInit())
        return -1;
    
    /* count number of cpus */
    if (hypCountProcessors())
        return -1;
    
    /* Allocate idt emulation structures */
    if (hypIdtEmulAlloc())
        return -1;

    Log("%s Hypervisor %s-%d initialized.", TARGET_NAME, TARGET_VERSION, VER_BUILD);
    
    return 0;
}

/* cleanup module function */
void cleanup_module(void) {

    /* free idt emul structs */
    hypIdtEmulFree();
    
    /* deinit vmstates */
    hypVmStateDeinit();
    
    /* deinit internal structs */
    hypInternalDeinit();
    
    /* deinit device */
    hypDeviceDeinit();
   
    /* free flag */
    kfree(hypInHandlerFlag);
    
    Log("%s Hypervisor %s-%d exited.", TARGET_NAME, TARGET_VERSION, VER_BUILD);
}

/* export symbols for other modules */
EXPORT_SYMBOL(appIncCounter);
EXPORT_SYMBOL(appDecCounter);
EXPORT_SYMBOL(hypervisorPresentInSystem);

/* little portion of wrappers for binary part of the driver */
int QUICKCALL hyper_wrap_smp_processor_id(void) {
    return smp_processor_id();
}

unsigned int QUICKCALL hyper_wrap_page_to_pfn(struct page *pg) {
    return page_to_pfn(pg);
}
