/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: hypvmstate.c                                                *
 *                                                                     *
 * Description:                                                        *
 *  Interface beetwen hypervisor and VM.                               *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>

#include "logging.h"
#include "hypervisor.h"

/* typical declaration of hyper switch lables */
#define DECLARE_HYPER_SWITCH(SUFFIX)	\
		void _HyperSwitchStartLabel_ ## SUFFIX (void);

/* declaration of unified hyper switcher */
DECLARE_HYPER_SWITCH(Unified)


/* getting pointer to start label depending on primary OS model */
HYPER_SWITCH_BUFF* getHostSwitchAreaPtr(void)
{
	switch (hypState.uPrOsPagingMode)
	{
	case PROS_PAGING_STD:
	case PROS_PAGING_PAE:
	default:
		return	(HYPER_SWITCH_BUFF*) _HyperSwitchStartLabel_Unified;
	}
}


/* allocate vmState struct */
void STDCALL *hypVMStateAllocate(void) {

    struct page *pg;
    struct hyp_vmstate_t *vmState;
    int i;
    unsigned int	uHyperSwitchAreaLen;
    HYPER_SWITCH_BUFF	*pHostSwitchBuff;
    HYPER_SWITCH_BUFF	*pSrcHostSwitchBuff;
	
    pSrcHostSwitchBuff = getHostSwitchAreaPtr();
    
    /* allocate vmState */
    if (!(vmState = kmem_cache_alloc(hypState.vmStateCache, GFP_ATOMIC)))
        return NULL;
   
    /* alloc buffer */
    if (!(vmState->stateBuff = (void *)mmAllocPages(VM_ACTIVATOR_PAGES)))
        goto err;

    memset(vmState->stateBuff, 0, (VM_ACTIVATOR_PAGES * PAGE_SIZE));

	pHostSwitchBuff = vmState->stateBuff;
	
    /* copy HyperSwitch buffer code */
    uHyperSwitchAreaLen = pSrcHostSwitchBuff->uHyperSwitchAreaLen;
    memcpy(pHostSwitchBuff, pSrcHostSwitchBuff, uHyperSwitchAreaLen);
    
    /* set activation function */
    vmState->vmActivate = (void STDCALL (*)(void))
		((unsigned int)pHostSwitchBuff + pHostSwitchBuff->uOffs_RunVM);
                
    mmEnableExecution(vmState->vmActivate,1,hypState.uPrOsPagingMode);

    /* set idt descriptor */
    vmState->vmIdtd = (struct idt_descriptor_t *)
    			(vmState->stateBuff + VM_IDT_STATE_PAGE_OFFSET);
    
    /* get physical addresses for shared buffer */
    for(i = 0; i < VM_ACTIVATOR_PAGES; i++) {
        pg = virt_to_page((vmState->stateBuff + (i * PAGE_SIZE)));
        if (pg) {
            vmState->vmStatePhyPages[i] = page_to_pfn(pg);
        } else  {
            Err("Null page for internal buffer!");
            goto err_all;
        }
    }
    
    /* break idtr */
    vmState->vmIdtr.wLimit = -1;
    vmState->vmIdtr.pBase = NULL;
   
    /* append new vmState to list 
     * We are in interrupt context with disabled interrupts,
     * no locking needed.
     */
    list_add(&vmState->list, &hypState.vmStateList);
    
    return vmState;

err_all:
    mmFreePages(vmState->stateBuff, VM_ACTIVATOR_PAGES);
    vmState->stateBuff = 0;
    
err:
    kmem_cache_free(hypState.vmStateCache, vmState);

    return NULL;
}

/* free vmState struct */
int STDCALL hypVMStateFree(struct hyp_vmstate_t *vmState) {
    
    if (!vmState)
        return -EINVAL;

    /* free buffer */
    mmFreePages(vmState->stateBuff, VM_ACTIVATOR_PAGES);
    vmState->stateBuff = 0;

    /* remove from list */
    list_del(&vmState->list);

    /* free entry from cache */
    kmem_cache_free(hypState.vmStateCache, vmState);

    return 0;
}

/* Getting primary OS paging mode */
unsigned int QUICKCALL getPrOsPagingMode(void)
{
    unsigned int uiCr4Value;

    __asm__ __volatile__ (
        "movl %%cr4, %%eax\n\t"
        "movl %%eax, %0\n\t"
        : : "m" (uiCr4Value)
        : "%eax", "memory"
    );

    #ifndef MODE_FEATURES_PAE_FLAG
	#define MODE_FEATURES_PAE_FLAG	0x20
    #endif

    if ( uiCr4Value & MODE_FEATURES_PAE_FLAG )
	return PROS_PAGING_PAE;

    return PROS_PAGING_STD;
}


/* initialize vmstates list, cache and etc */
int QUICKCALL hypVmStateInit(void) {

    /* init semaphore */
    sema_init(&hypState.vmsListAccSem, 1);

    /* create slabcache */
    if (!(hypState.vmStateCache = kmem_cache_create(VMSTATE_SLABNAME, sizeof(struct hyp_vmstate_t), 0,
                                                    SLAB_HWCACHE_ALIGN, NULL, NULL)))
        return -ENOMEM;

    /* init list */
    INIT_LIST_HEAD(&hypState.vmStateList);

    /* detecting primary OS paging mode */
    hypState.uPrOsPagingMode = getPrOsPagingMode();
    if ( hypState.uPrOsPagingMode == PROS_PAGING_INVALID_MODE )
	return -EFAULT;

    return 0;
}

/* deinitialize vmstates stuff */
void QUICKCALL hypVmStateDeinit(void) {

    struct list_head *q, *s;
    struct hyp_vmstate_t *vmState;

    /* check for pmms */
    list_for_each_safe(q, s, &hypState.vmStateList) {
        vmState = (struct hyp_vmstate_t *)list_entry(q, struct hyp_vmstate_t, list);
        if (vmState) {
            /* free here, and show warn */
            Warn("NonEmpty VMState list on exit, free record [0x%x]", vmState);
            hypVMStateFree(vmState);
        }
    }

    /* free slabcache */
    kmem_cache_destroy(hypState.vmStateCache);

    return;
}
