/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: svm_init.c                                                  *
 *                                                                     *
 * Description:                                                        *
 *  Variables and functions for work with AMD SVM.                     *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: vigor                                                       *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/vmalloc.h>
#include <asm/processor.h>
#include <asm/io.h>

#include "hypervisor.h"
#include "vm_compiler.h"
#include "logging.h"
#include "compat_smp.h"
#include "mm.h"
#include "svm_init.h"


/* Control structure with all data for work with SVM. */
typedef struct __TAG_SVM_CTRL {
    /* Flags of success exit */
    int valid;
    /* Number of processor in system */
    u32 num_cpu;
    /* List of flags for SVME bit in EFER */
    int svme[SVM_MAX_CPU_NUMBER];
} SVM_CTRL, *PSVM_CTRL;

/* Global variable for save control structure. */
static PSVM_CTRL g_p_svm_ctrl = NULL;

/* Stub for SMP work. */
struct compat_smp_fn_s {
    void QUICKCALL (*fn)(void *arg);
    void *arg;
};

/* Stub for SMP work. */
static void svm_smp_wrap(void *arg) {
    struct compat_smp_fn_s *fn = arg;
    fn->fn(fn->arg);
}

/*
 *
 * FUNCTION: is_svm_support_on()
 *
 * DESCRIPTION:
 *   Check of SVM active
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 */
int QUICKCALL is_svm_support_on(void) {

    u32 i;

    /* If not allocated control structure */
    if (!g_p_svm_ctrl)
        return -EFAULT;

    /* Only for valid state */
    for (i = 0; i < g_p_svm_ctrl->num_cpu; i++)
    {
	/* Check state of CPU */
	if (!g_p_svm_ctrl->svme[i])
	{
	    return -EFAULT;
	}
    }

    return g_p_svm_ctrl->valid ? 0 : -EFAULT;
}


/*
 *
 * FUNCTION: svm_support_on()
 *
 * DESCRIPTION:
 *  Initialize for work monitor with SVM mode.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 */
int QUICKCALL svm_support_on(void) {
    int svm_support;
    unsigned long eflags;    
    struct compat_smp_fn_s fn = {svm_enter, NULL};
    
    /* Save IF state and disable hardware interrupts */
    local_irq_save(eflags);

    /* Get of SVM support */
    svm_support = is_svm_support();
    
    /* If need, enable hardware interrupts */
    local_irq_restore(eflags);

    /* Check of present SVM support */
    if (!svm_support)
        return -EFAULT;

    /* Allocate memory for control structure */
    g_p_svm_ctrl = (PSVM_CTRL)kmalloc(sizeof(SVM_CTRL), GFP_KERNEL);
    if (!g_p_svm_ctrl) {
        Err("kmalloc() failed.");
        return -ENOMEM;
    }

    /* Initialze memory for control structure */
    memset(g_p_svm_ctrl, 0, sizeof(SVM_CTRL));

    /* Get number of processors in system */
    g_p_svm_ctrl->num_cpu = hypState.cpuCount;
    if (g_p_svm_ctrl->num_cpu > SVM_MAX_CPU_NUMBER) {
        Err("Number of CPUs is greater then %d.", SVM_MAX_CPU_NUMBER);
        /* Release all resources */
        svm_support_off();
        return -EFAULT;
    }

    /* Loop for all processors in system */
    if (compat_smp_ipi_call(svm_smp_wrap, &fn, 1, 1)) {
        Err("compat_smp_ipi_call() failed.");
        /* Release all resources */
        svm_support_off();
        return -EFAULT;
    }

    /* Set flag of success exit */
    g_p_svm_ctrl->valid = 1;

    Log("svm_support_on(): SVM mode activated");

    return 0;
}


/*
 *
 * FUNCTION: svm_support_off()
 *
 * DESCRIPTION:
 *  Deinitialize for work monitor with SVM mode.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 */
int QUICKCALL svm_support_off(void) {
    
    struct compat_smp_fn_s fn = { svm_leave, NULL };

    /* If not allocated control structure */
    if (!g_p_svm_ctrl)
        return -EFAULT;
    
    /* Only for valid state */
    if (g_p_svm_ctrl->valid) {
        /* Loop for all processors in system */
        if (compat_smp_ipi_call(svm_smp_wrap, &fn, 1, 1)) {
            Err("compat_smp_ipi_call() failed.");
        }
    }

    /* Release memory for control structure */
    kfree(g_p_svm_ctrl);
    g_p_svm_ctrl = NULL;

	Log("svm_support_off(): SVM mode deactivated.");

    return 0;
}

/*
 *
 * FUNCTION: svm_enter()
 *
 * DESCRIPTION:
 *  Activate SVM mode for current processor.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 */
void QUICKCALL svm_enter(void *p) {
    
    int cpu;
    unsigned long eflags;

    /* Get number of processor */
    cpu = smp_processor_id();

    /* Save IF state and disable hardware interrupts */
    local_irq_save(eflags);

    /* Check of present SVM support */
    if (is_svm_support()) {
        /* Set enabled mode for SVM operations in monitor EFER */
        g_p_svm_ctrl->svme[cpu] = svm_enable();
    }

    /* If need, enable hardware interrupts */
    local_irq_restore(eflags);
}


/*
 *
 * FUNCTION: svm_leave()
 *
 * DESCRIPTION:
 *  Deactivate SVM mode for current processor.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 */
void QUICKCALL svm_leave(void *p) {
    
    int cpu;
    unsigned long eflags;

    /* Get number of processor */
    cpu = smp_processor_id();

    /* Save IF state and disable hardware interrupts */
    local_irq_save(eflags);

    /* Check of present SVM support */
    if (is_svm_support()) {
        /* If current processor enabled for SVM operations in EFER */
        if (g_p_svm_ctrl->svme[cpu]) {
            /* Disabled mode for SVM operations in host EFER */
            svm_disable();

            /* Clear flag for SVME bit in EFER */
            g_p_svm_ctrl->svme[cpu] = 0;
        }
    }

    /* If need, enable hardware interrupts */
    local_irq_restore(eflags);
}
