/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2006 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: svm_init.h                                                  *
 *                                                                     *
 * Description:                                                        *
 *  Definitions and structures for work with AMD SVM.                  *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: vigor                                                       *
 * ------------------------------------------------------------------- *
 */

#ifndef __SVM_INIT_H__
#define __SVM_INIT_H__

#include "vm_compiler.h"

#include <asm/processor.h>


/* Maximum number of support CPU for work in SVM mode. */
#define SVM_MAX_CPU_NUMBER		(32)


/* Definition of AMD CPUID function to get processor signature */
#define CPUID_PROCESSOR_SIGNATURE	(0x80000001)


/* Definition of bit for SVM support in ecx after CPUID with 0x80000001 fn. */
#define CPUID_SVM			(1 << 2)


/* Definition of Extended Feature Enables (EFER) control register */
#define MSR_IA32_EFER			(0xC0000080)


/* Definition of bit for control SVM instructions in EFER */
#define EFER_SVME			(1 << 12)



/*
 *
 * FUNCTION: is_svm_support()
 *
 * DESCRIPTION:
 *  Check for support SVM instructions.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *  True for proccesor with possible support SVM.
 *
 * REMARK:
 *  Hardware interrupts should be disabled.
 *
 */
static __inline__ int is_svm_support(void) {
    
    /* Check SVM bit in CPUID extended features */    
    return cpuid_ecx(CPUID_PROCESSOR_SIGNATURE) & CPUID_SVM;
}


/*
 *
 * FUNCTION: svm_enable()
 *
 * DESCRIPTION:
 *  Enable all SVM instructions for CPU.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *  True for previously disabled SVM mode in current CPU.
 *
 * REMARK:
 *  Hardware interrupts should be disabled.
 *
 */
static __inline__ int svm_enable(void) {
    
    u32 h, efer = 0;

    /* Get current EFER value */
    rdmsr(MSR_IA32_EFER, efer, h);
	
	/* If already set SVME bit in EFER */
    if (efer & EFER_SVME)
        return 0;

    /* Set SVME bit in EFER */
    efer |= EFER_SVME;
	
    /* Set new EFER value */
    wrmsr(MSR_IA32_EFER, efer, h);

    return 1;
}


/*
 *
 * FUNCTION: svm_disable()
 *
 * DESCRIPTION:
 *  Disable all SVM instructions for CPU.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *  True for previously enabled SVM mode in current CPU.
 *
 * REMARK:
 *  Hardware interrupts should be disabled.
 *
 */
static __inline__ int svm_disable(void) {
    
    u32 h, efer = 0;

    /* Get current EFER value */
    rdmsr(MSR_IA32_EFER, efer, h);

    /* If already clear SVME bit in EFER */
    if (!(efer & EFER_SVME))
        return 0;

    /* Clear SVME bit in EFER */
    efer &= ~EFER_SVME;
	
    /* Set new EFER value */
    wrmsr(MSR_IA32_EFER, efer, h);

    return 1;
}


/* Check of SVM active */
int QUICKCALL is_svm_support_on(void);

/* Initialize for work monitor with SVM mode. */
int QUICKCALL svm_support_on(void);

/* Deinitialize for work monitor with SVM mode. */
int QUICKCALL svm_support_off(void);

/* Enter SVM mode for current processor. */
void QUICKCALL svm_enter(void *);

/* Leave SVM mode for current processor. */
void QUICKCALL svm_leave(void *);

#endif /* __SVM_INIT_H__ */
