/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: vtx_init.c                                                  *
 *                                                                     *
 * Description:                                                        *
 *  Variables and functions for work with Intel VT-x.                  *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: diman                                                       *
 * ------------------------------------------------------------------- *
 */

#include "vmmodule.h"

#include <linux/vmalloc.h>
#include <asm/processor.h>
#include <asm/io.h>

#include "hypervisor.h"
#include "vm_compiler.h"
#include "logging.h"
#include "compat_smp.h"
#include "mm.h"
#include "vtx_init.h"


/* Control structure with all data for work with VT-x. */
typedef struct __TAG_VTX_CTRL {
    /* Flags of success exit */
    int valid;
    /* Number of processor in system */
    u32 num_cpu;
    /* VT-x revision id */
    u32 vmcs_revision;
    /* Pointer to VMXON region pages */
    PVMCS_REGION p_vmxon_regions;
    /* List of physical addresses for VMXON region pages */
    u64 vmxon_pages[VTX_MAX_CPU_NUMBER];
    /* List of flags for VMXE bit in CR4 */
    int vmxe[VTX_MAX_CPU_NUMBER];
    /* List of flags for correct execution VMXON instruction */
    int vmxon[VTX_MAX_CPU_NUMBER];
} VTX_CTRL, *PVTX_CTRL;

/* Global variable for save control structure. */
static PVTX_CTRL g_p_vtx_ctrl = NULL;

/* Stab for SMP work. */
struct compat_smp_fn_s {
    void QUICKCALL (*fn)(void *arg);
    void *arg;
};

/* Stab for SMP work. */
static void vtx_smp_wrap(void *arg) {
    struct compat_smp_fn_s *fn = arg;
    fn->fn(fn->arg);
}

/*
 *
 * FUNCTION: is_vtx_support_on()
 *
 * DESCRIPTION:
 *   Check of VT-x active
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 */
int QUICKCALL is_vtx_support_on(void) {

    u32 i;

    /* If not allocated control structure */
    if (!g_p_vtx_ctrl)
        return -EFAULT;

    /* Only for valid state */
    for (i = 0; i < g_p_vtx_ctrl->num_cpu; i++)
    {
	/* Check state of CPU */
	if (!g_p_vtx_ctrl->vmxe[i] || !g_p_vtx_ctrl->vmxon[i])
	{
	    return -EFAULT;
	}
    }

    return g_p_vtx_ctrl->valid ? 0 : -EFAULT;
}


/*
 *
 * FUNCTION: vtx_support_on()
 *
 * DESCRIPTION:
 *  Initialize for work monitor with VT-x mode.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 */
int QUICKCALL vtx_support_on(void) {
    int vtx_support;
    unsigned long eflags;
    u32 h, vmcs_revision = 0;
    struct compat_smp_fn_s fn = {vtx_activate, NULL};
    
#ifdef VTX_ALWAYS_ON_OFF
    return 0;
#endif

    /* Save IF state and disable hardware interrupts */
    local_irq_save(eflags);

    /* Get of VT-x support */
    vtx_support = is_vtx_support();
    if (vtx_support) {
        /* Get VT-x revision identifier */
        rdmsr(VMX_BASIC, vmcs_revision, h);
    }

    /* If need, enable hardware interrupts */
    local_irq_restore(eflags);

    /* Check of present VT-x support */
    if (!vtx_support)
        return -EFAULT;

    /* Allocate memory for control structure */
    g_p_vtx_ctrl = (PVTX_CTRL)kmalloc(sizeof(VTX_CTRL), GFP_KERNEL);
    if (!g_p_vtx_ctrl) {
        Err("kmalloc() failed.");
        return -ENOMEM;
    }

    /* Initialze memory for control structure */
    memset(g_p_vtx_ctrl, 0, sizeof(VTX_CTRL));
    g_p_vtx_ctrl->vmcs_revision = vmcs_revision;

    /* Get number of processors in system */
    g_p_vtx_ctrl->num_cpu = hypState.cpuCount;
    if (g_p_vtx_ctrl->num_cpu > VTX_MAX_CPU_NUMBER) {
        Err("Number of CPUs is greater then %d.", VTX_MAX_CPU_NUMBER);
        /* Release all resources */
        vtx_support_off();
        return -EFAULT;
    }

    /* Allocate memory for VMXON regions */
    g_p_vtx_ctrl->p_vmxon_regions = (PVMCS_REGION)mmAllocPages(g_p_vtx_ctrl->num_cpu);
    if (!g_p_vtx_ctrl->p_vmxon_regions) {
        Err("kmalloc() failed.");
        /* Release all resources   1 */
        vtx_support_off();
        return -ENOMEM;
    }

    /* Initialze memory for all VMXON regions */
    memset(g_p_vtx_ctrl->p_vmxon_regions, 0, g_p_vtx_ctrl->num_cpu * PAGE_SIZE);

    /* Loop for all processors in system */
    if (compat_smp_ipi_call(vtx_smp_wrap, &fn, 1, 1)) {
        Err("compat_smp_ipi_call() failed.");
        /* Release all resources */
        vtx_support_off();
        return -EFAULT;
    }

    /* Set flag of success exit */
    g_p_vtx_ctrl->valid = 1;

    Log("vtx_support_on(): VT-x mode activated, VMCS revision id %x.", g_p_vtx_ctrl->vmcs_revision);

    return 0;
}


/*
 *
 * FUNCTION: vtx_support_off()
 *
 * DESCRIPTION:
 *  Deinitialize for work monitor with VT-x mode.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 */
int QUICKCALL vtx_support_off(void) {
    
    struct compat_smp_fn_s fn = { vtx_deactivate, NULL };

#ifdef VTX_ALWAYS_ON_OFF
    return 0;
#endif

    /* If not allocated control structure */
    if (!g_p_vtx_ctrl)
        return -EFAULT;

    /* Only for valid state */
    if (g_p_vtx_ctrl->valid) {
        /* Loop for all processors in system */
        if (compat_smp_ipi_call(vtx_smp_wrap, &fn, 1, 1)) {
            Err("compat_smp_ipi_call() failed.");
        }
    }

    /* Release memory for VMXON regions */
    if (g_p_vtx_ctrl->p_vmxon_regions)
        mmFreePages(g_p_vtx_ctrl->p_vmxon_regions, g_p_vtx_ctrl->num_cpu);

    /* Release memory for control structure */
    kfree(g_p_vtx_ctrl);
    g_p_vtx_ctrl = NULL;

	Log("vtx_support_off(): VT-x mode deactivated.");

    return 0;
}

/*
 *
 * FUNCTION: vtx_support_off()
 *
 * DESCRIPTION:
 *  Activate VMX root mode for current processor.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 */
void QUICKCALL vtx_activate(void *p) {
    
    int cpu;
    unsigned long eflags;

    /* Get number of processor */
    cpu = smp_processor_id();

    /* Save IF state and disable hardware interrupts */
    local_irq_save(eflags);

    /* Check of present VT-x support */
    if (is_vtx_support()) {
        /* Set enabled mode for VT-x operations in monitor CR4 */
        g_p_vtx_ctrl->vmxe[cpu] = vtx_enable();

        /* Get of physical address for VMXON region page */
        g_p_vtx_ctrl->vmxon_pages[cpu] = virt_to_phys(&g_p_vtx_ctrl->p_vmxon_regions[cpu]);

        /* Set VT-x revision identifier in VMXON region page */
        g_p_vtx_ctrl->p_vmxon_regions[cpu].details.vmcs_revision = g_p_vtx_ctrl->vmcs_revision;

        /* Switch current CPU to VMX root mode */
        g_p_vtx_ctrl->vmxon[cpu] = VMXON(g_p_vtx_ctrl->vmxon_pages[cpu]);
    }

    /* If need, enable hardware interrupts */
    local_irq_restore(eflags);
}


/*
 *
 * FUNCTION: vtx_support_off()
 *
 * DESCRIPTION:
 *  Deactivate VMX root mode for current processor.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 */
void QUICKCALL vtx_deactivate(void *p) {
    
    int cpu;
    unsigned long eflags;

    /* Get number of processor */
    cpu = smp_processor_id();

    /* Save IF state and disable hardware interrupts */
    local_irq_save(eflags);

    /* Check of present VT-x support */
    if (is_vtx_support()) {
        /* If current processor in VMX root mode */
        if (g_p_vtx_ctrl->vmxon[cpu]) {
            /* Execute VMXON for disable possible #UD exception */
            VMXON(g_p_vtx_ctrl->vmxon_pages[cpu]);
        
            /* Turn off VMX root mode */
            VMXOFF();

            /* Clear flag for current processor */
            g_p_vtx_ctrl->vmxon[cpu] = 0;
        }

        /* If current processor enabled for VT-x operations in CR4 */
        if (g_p_vtx_ctrl->vmxe[cpu]) {
            /* Disabled mode for VT-x operations in host CR4 */
            vtx_disable();

            /* Clear flag for VMXE bit in CR4 */
            g_p_vtx_ctrl->vmxe[cpu] = 0;
        }
    }

    /* If need, enable hardware interrupts */
    local_irq_restore(eflags);
}
