/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: vtx_init.h                                                  *
 *                                                                     *
 * Description:                                                        *
 *  Definitions and structures for work with Intel VT-x.               *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: diman                                                       *
 * ------------------------------------------------------------------- *
 */

#ifndef __VTX_INIT_H__
#define __VTX_INIT_H__

#include "vm_compiler.h"

#include <asm/processor.h>


/* Maximum number of support CPU for work in VT-x mode. */
#define VTX_MAX_CPU_NUMBER      (32)

/* Definition of bit for VT-x support in CPUID. */
#define CPUID_VMX               (1 << 5)

/* Definition of bit for control VT-x instructions in CR4. */
#define CR4_VMXE                (1 << 13)

/* Definitions for check after VT-x instructions. */
#define EFLAGS_CF               (1 << 0)
#define EFLAGS_ZF               (1 << 6)
#define VTX_EFLAGS_FAILED       (EFLAGS_CF | EFLAGS_ZF)

/* Definitions of control MSR for VT-x. */
#define IA32_FEATURE_CONTROL    (0x0000003a)
#define FEATURE_CONTROL_LOCK    (1 << 0)
#define FEATURE_CONTROL_VMXON   (1 << 2)
#define FEATURE_CONTROL_VALID   (FEATURE_CONTROL_LOCK | FEATURE_CONTROL_VMXON)


/* Definitions of indexes special MSR for VT-x. */
#define VMX_BASIC               (1152)


/*
 *
 * UNION: VMCS_REGION
 *
 * DESCRIPTION:
 *  Type of Virtual Machine Control Structure.
 *
 */
typedef union __TAG_VMCS_REGION {
    struct {
        /* Contain the VMCS revision identifier.
         * These identifiers enable software to avoid using a VMCS region
         * formatted for one processor on a processor that uses a different format. */
        u32 vmcs_revision;
        /* A logical processor writes a non-zero value into these bytes
         * if a VMX abort occurs. Software may also write into this field. */
        u32 vmx_abort_indicator;
        /* VMCS data (implementation-specific format). */
        u8 data[1];
    } PACKED details;
    u8 cells[PAGE_SIZE];
} VMCS_REGION;
typedef VMCS_REGION* PVMCS_REGION;


/*
 *
 * FUNCTION: is_vtx_support()
 *
 * DESCRIPTION:
 *  Check for support VT-x instructions.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *  True for proccesor with possible support VT-x.
 *
 * REMARK:
 *  Hardware interrupts should be disabled.
 *
 */
static __inline__ int is_vtx_support(void) {
    
    u32 feature_ctl, h;

    /* Check VMX bit in CPUID extended features */
    if (!(cpuid_ecx(1) & CPUID_VMX))
        return 0;

    /* Get CPU features control MSR value */
    rdmsr(IA32_FEATURE_CONTROL, feature_ctl, h);

    /* If set only lock bit, VMXON hardware disabled */
    if ((feature_ctl & FEATURE_CONTROL_VALID) == FEATURE_CONTROL_LOCK)
        return 0;

    /* If VMXON only software disabled, set valid bits for VMXON enable */
    if ((feature_ctl & FEATURE_CONTROL_VALID) != FEATURE_CONTROL_VALID)
        wrmsr(IA32_FEATURE_CONTROL, feature_ctl | FEATURE_CONTROL_VALID, h);

    return 1;
}


/*
 *
 * FUNCTION: is_vtx_active()
 *
 * DESCRIPTION:
 *  Check for support VT-x instructions.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *  True for proccesor with active VT-x.
 *
 * REMARK:
 *  Hardware interrupts should be disabled.
 *
 */
static __inline__ int is_vtx_active(void) {
    
    /* Check set VMXE bit in CR4 */
    return (read_cr4() & CR4_VMXE) ? 1 : 0;
}


/*
 *
 * FUNCTION: set_cr4_features()
 *
 * DESCRIPTION:
 *  Set CR4 bits for internal Linux structures.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 * 
 * REMARK:
 *  Hardware interrupts should be disabled.
 *
 */
static __inline__ int set_cr4_features(u32 mask) {
    
    u32 *address = (u32*)CR4_FEATURES_ADDR;

    /* exported symbol not found */
    if (!address)
        return 0;

    *address |= mask;

    return 1;
}


/*
 *
 * FUNCTION: clear_cr4_features()
 *
 * DESCRIPTION:
 *  Clear CR4 bits for internal Linux structures.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 * REMARK:
 *  Hardware interrupts should be disabled.
 *
 */
static __inline__ int clear_cr4_features(u32 mask) {
    
    u32 *address = (u32*)CR4_FEATURES_ADDR;

    /* exported symbol not found */
    if (!address)
        return 0;

    *address &= ~mask;

    return 1;
}


/*
 *
 * FUNCTION: vtx_enable()
 *
 * DESCRIPTION:
 *  Enable all VT-x instructions for CPU.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *  True for previously disabled VT-x mode in current CPU.
 *
 * REMARK:
 *  Hardware interrupts should be disabled.
 *
 */
static __inline__ int vtx_enable(void) {
    
    u32 cr4;

    /* Get current CR4 value */
    cr4 = read_cr4();

    /* If already set VMXE bit in CR4 */
    if (cr4 & CR4_VMXE)
        return 0;

    /* Set new CR4 value */
    write_cr4(cr4 | CR4_VMXE);

    return 1;
}


/*
 *
 * FUNCTION: vtx_disable()
 *
 * DESCRIPTION:
 *  Disable all VT-x instructions for CPU.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *  True for previously enabled VT-x mode in current CPU.
 *
 * REMARK:
 *  Hardware interrupts should be disabled.
 *
 */
static __inline__ int vtx_disable(void) {
    
    u32 cr4;

    /* Get current CR4 value */
    cr4 = read_cr4();

    /* If already clear VMXE bit in CR4 */
    if (!(cr4 & CR4_VMXE))
        return 0;

    /* Set new CR4 value */
    write_cr4(cr4 & ~CR4_VMXE);

    return 1;
}

/*
 *
 * FUNCTION: VMXOFF()
 *
 * DESCRIPTION:
 *  This instruction leaves VT-x operation.
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 * REMARK:
 *  Hardware interrupts should be disabled.
 *
 */
static __inline__ void VMXOFF(void) {
    
    /* Execute VT-x instruction */
    __asm__ __volatile__ (
        ".byte 0x0f,0x01,0xc4\n\t"    /* vmxoff */
        : : : "memory"
    );
}


/*
 *
 * FUNCTION: VMXON()
 *
 * DESCRIPTION:
 *  This instruction takes a single 64-bit source operand that is
 *  in memory. It causes a logical processor to enter VMX root
 *  operation and to use the memory referenced by the operand to
 *  support VMX operation.
 *
 * PARAMETERS:
 *  uVmcsAddr - Physical address for VMXON region page.
 *
 * RETURN VALUE:
 *  True for correct execution.
 *
 * REMARK:
 *  Hardware interrupts should be disabled.
 *
 */
static __inline__ int VMXON(u64 vmxon_area) {
    
    unsigned long eflags;

    /* Execute VT-x instruction and get register of flags */
    __asm__ __volatile__ (
        ".byte 0xf3,0x0f,0xc7,0x30\n\t"   /* vmxon qword ptr [eax] */
        : : "a"(&vmxon_area) 
        : "memory"
    );

    /* Get value of flags register */
#ifdef local_save_flags
    local_save_flags(eflags);
#else
    save_flags(eflags);
#endif

    /* Check for correct execution */
    if (eflags & VTX_EFLAGS_FAILED)
        return 0;

    return 1;
}

/* Check of VT-x active */
int QUICKCALL is_vtx_support_on(void);

/* Initialize for work monitor with VT-x mode. */
int QUICKCALL vtx_support_on(void);

/* Deinitialize for work monitor with VT-x mode. */
int QUICKCALL vtx_support_off(void);

/* Activate VMX root mode for current processor. */
void QUICKCALL vtx_activate(void *);

/* Deactivate VMX root mode for current processor. */
void QUICKCALL vtx_deactivate(void *);

#endif /* __VTX_INIT_H__ */
