/* ------------------------------------------------------------------- *
 *                                                                     *
 *   Copyright (C) 2005 Parallels Software Inc. All Rights Reserved.   *
 *                      http://www.parallels.com                       *
 *                                                                     *
 * ------------------------------------------------------------------- *
 *                                                                     *
 * Module: x86.h                                                       *
 *                                                                     *
 * Description:                                                        *
 *  Some structures and defines for x86 arch.                          *
 *                                                                     *
 * Comments:                                                           *
 *                                                                     *
 * Author: ms                                                          *
 * ------------------------------------------------------------------- *
 */

#ifndef __X86_H__
#define __X86_H__

#include "vm_compiler.h"

/* load idtr with new value  */
#define do_lidt(addr)       \
    __asm__ __volatile__ (  \
        "cli\n\t"           \
        "lidtl %0\n\t"      \
        "sti\n\t"           \
        : : "m" (addr)      )

/* read idtr */
#define do_sidt(addr)       \
    __asm__ __volatile__ (  \
        "sidtl %0\n\t"   \
        : "=m" (addr))

/* cli instr */
#define do_cli()            \
    __asm__ __volatile__ (  \
        "cli\n\t"           )

/* str instr */
#define do_sti()            \
    __asm__ __volatile__ (  \
        "sti\n\t"           )

/* exceptions */
#define EXCEPTION_DE    0      /* Divide Error (fault) */
#define EXCEPTION_DB    1      /* Debug (fault/trap) */
#define EXCEPTION_NMI   2      /* NMI */ 
#define EXCEPTION_BP    3      /* Breakpoint (trap) */
#define EXCEPTION_OF    4      /* Overflow (trap) */
#define EXCEPTION_BR    5      /* BOUND (fault) */
#define EXCEPTION_UD    6      /* UnDefined opcode (fault) */
#define EXCEPTION_NM    7      /* No Math coprocessor (fault) */
#define EXCEPTION_DF    8      /* Double Fault (abort,EC=0) */
#define EXCEPTION_FPU   9      /* FPU segment exception */ 
#define EXCEPTION_TS    10     /* Task Segment (fault,EC) */
#define EXCEPTION_NP    11     /* Not Present (fault,EC) */
#define EXCEPTION_SS    12     /* Stack Segment (fault,EC) */
#define EXCEPTION_GP    13     /* General Protection (fault,EC) */
#define EXCEPTION_PF    14     /* Page Fault (fault,EC) */
#define EXCEPTION_RES   15     /* Reserved */ 
#define EXCEPTION_MF    16     /* Math coprocessor Fault (fault) */
#define EXCEPTION_AC    17     /* Align Check (fault,EC=0) */
#define EXCEPTION_MC    18     /* Machine Check (abort,EC) */
#define EXCEPTION_XF    19     /* Xmm Fault (fault) */

/* x86cpu flags */
typedef union {
    unsigned int cell;
    /* enum for easy access */
    struct {
        unsigned int    CF : 1;
        unsigned int    Reserved_1 : 1;
        unsigned int    PF : 1;
        unsigned int    R3 : 1;
        unsigned int    AF : 1;
        unsigned int    R5 : 1;
        unsigned int    ZF : 1;
        unsigned int    SF : 1;
        unsigned int    TF : 1;
        unsigned int    IF : 1;
        unsigned int    DF : 1;
        unsigned int    OF : 1;
        unsigned int    IOPL : 2;
        unsigned int    NT : 1;
        unsigned int    Reserved_15 : 1;
        unsigned int    RF : 1;
        unsigned int    VM : 1;
        unsigned int    AC : 1;
        unsigned int    VIF : 1;
        unsigned int    VIP : 1;
        unsigned int    ID : 1;
        unsigned int    Reserved_22_31 : 10;
    } details PACKED;
} reg_flags_t;

/* hypercall handler stack */
struct hypercall_context_t {
    unsigned int ds;
    unsigned int es;
    unsigned int edi;
    unsigned int esi;
    unsigned int ebp;
    unsigned int espInvalid;
    unsigned int ebx;
    unsigned int edx;
    unsigned int ecx;
    unsigned int eax;
    unsigned int eip;
    unsigned int uCs;
    reg_flags_t flags;
} PACKED ; 

/* exception handler stack struct */
struct int_handler_stack_t {
    unsigned int    uGs;
    unsigned int    uFs;
    unsigned int    uDs;
    unsigned int    uEs;
    unsigned int    uEdi;
    unsigned int    uEsi;
    unsigned int    uEbp;
    unsigned int    uEspInvalid;
    unsigned int    uEbx;
    unsigned int    uEdx;
    unsigned int    uEcx;
    unsigned int    uEax;
    unsigned int    uProcessorNumber;
    unsigned int    uIntNo;
    unsigned int    uInternalFlags;
    unsigned int    uRedirectionEip;
    unsigned int    uErrorNo;
    unsigned int    uEip;
    unsigned int    uCs;
    reg_flags_t     Eflags;
    unsigned int    uEsp;
    unsigned int    uSs;
} PACKED ;

/* handler instrs */
#define INSTR_PUSH  0x68    /* push */
#define INSTR_JMP   0xe9    /* jmp */

/* selector structure */
typedef union {
    unsigned short cell;

    struct {
        unsigned short rpl: 2;
        unsigned short ti: 1;
        unsigned short index: 13; 
    } details PACKED;
} selector_t;

/* descriptor structure */
struct descriptor_t {
    unsigned short limitLo;
    unsigned short baseLo;
    unsigned char  baseCentral;
    unsigned short type : 5;
    unsigned short dpl : 2;
    unsigned short p : 1;
    unsigned short limitHi4 : 4;
    unsigned short avl : 1;
    unsigned short reserved : 1;
    unsigned short db : 1;
    unsigned short g : 1;
    unsigned char  baseHi;
};

/* int handler info */
struct int_handlers_info_t {
    unsigned int handlersPhyPage;
    unsigned int hypervisorHandlersPageVA;
    unsigned int handlerException;
    unsigned int handlerExceptionWE;
    unsigned int interruptHandler;
    unsigned int vmJmpTableVA;
    unsigned int vmJmpTablePage;
};

/* descriptor length */
enum {
    DESCR_DEFSIZE_16 = 0,
    DESCR_DEFSIZE_32
};

/* DPL levels */
enum {
    DESCR_DPL0 = 0,
    DESCR_DPL1,
    DESCR_DPL2,
    DESCR_DPL3,
};

/* present/nonpresent descriptor flag */
enum {
    DESCR_NOT_PRESENT = 0,
    DESCR_PRESENT,
};

/* far jump struct (ljmp) */
struct farjmp_t {
    unsigned int offset;
    unsigned short sel;
}PACKED;

/* gdt descriptor */
struct gdt_descr_t { 
    unsigned short Limit;
    unsigned int   Base;
} PACKED;

/* segments and gate descriptors types */
enum {
    SSGD_RESERVED0       = 0x0,
    SSGD_16BIT_TSS_AVAIL = 0x1,
    SSGD_LDT             = 0x2,
    SSGD_16BIT_TSS_BUSY  = 0x3,
    SSGD_16BIT_CALL_GATE = 0x4,
    SSGD_TASK_GATE       = 0x5,
    SSGD_16BIT_INT_GATE  = 0x6,
    SSGD_16BIT_TRAP_GATE = 0x7,
    SSGD_RESERVED8       = 0x8,
    SSGD_32BIT_TSS_AVAIL = 0x9,
    SSGD_RESERVED_10     = 0xa,
    SSGD_32BIT_TSS_BUSY  = 0xb,
    SSGD_32BIT_CALL_GATE = 0xc,
    SSGD_RESERVED_13     = 0xd,
    SSGD_32BIT_INT_GATE  = 0xe,
    SSGD_32BIT_TRAP_GATE = 0xf,
};

/* build descriptors */
#define GATHER_INTERRUPT_GATE_DESCRIPTOR(Descr, S, O, I_P, DPL, D) {    \
        Descr.wSelector = (S);                                          \
        Descr.wHighOffset = (O) >> 16;                                  \
        Descr.wLowOffset = (O) & 0xffff;                                \
        Descr.cReserved = 0;                                            \
        Descr.cType = D ? SSGD_32BIT_INT_GATE : SSGD_16BIT_INT_GATE;    \
        Descr.cDpl = (DPL);                                             \
        Descr.cPresent = (I_P);                                         }

#define GATHER_TRAP_GATE_DESCRIPTOR(Descr, S, O, I_P, DPL, D) {         \
        Descr.wSelector = (S);                                          \
        Descr.wHighOffset = (O) >> 16;                                  \
        Descr.wLowOffset = (O) & 0xffff;                                \
        Descr.cReserved = 0;                                            \
        Descr.cType = D ? SSGD_32BIT_TRAP_GATE : SSGD_32BIT_TRAP_GATE;  \
        Descr.cDpl = (DPL);                                             \
        Descr.cPresent = (I_P);                                         }


#endif /* __X86_H__ */
