/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2005 Parallels Inc. All Rights Reserved.
//		 http://www.parallelssoft.com
//
// Module: ApiNet.h
//
// Description: Ring buffer description
//
/////////////////////////////////////////////////////////////////////////////
#ifndef API_NET_H
#define API_NET_H

// Packing structures with 1 byte 
#if !defined(_LINUX_) && !defined(_BSD_)
#pragma pack(push, save_pack, 1)
#ifndef STRUCT_NAME
#define STRUCT_NAME(x) x
#endif
#else
#define STRUCT_NAME(x) __attribute__((packed)) x
#endif


// Ring buffer for packets maximum size
#define IO_NET_SIZE (256*1024)

/////////////////////////////////////////////////////////////////////////////
//
// STRUCTURE: NET_BUFFER
//
// DESCRIPTION:
//		Ring buffer for packet reception description
//
// COMMENT:
//		Struct of ring buffer:
//		|<---------------------------------------- NET_BUFFER ------------------------------------------->|
//		|                                                                                                 |
//		|             |<--------------------------------- NET_BUFFER.aBuffer ---------------------------->|
//		|             |                                                                                   |
//		+-------------+--------------------------------+-----+--------------------------------+-----------+
//		| Info about  |        Received packet         | ... |        Received packet         | Free      |
//		| ring buffer.+---------------+----------------+ ... +---------------+----------------+           +
//		| (aBuffer)   | PACKET_HEADER | Packet(data)   | ... | PACKET_HEADER | Packet(data)   |  area     |
//		+-------------+---------------+----------------+-----+---------------+----------------+-----------+
//		              ^
//		              0
//
/////////////////////////////////////////////////////////////////////////////
typedef struct _NET_BUFFER
{
	// Virtual cable is plugged
	UINT	bCableUnplugged;
	// Which packets to receive
	// broad/multicasts always pass currently 
	UINT	uRecFilter;
#define NET_FILTER_DIRECTED 1
#define NET_FILTER_BROADCAST 2
#define NET_FILTER_MULTICAST 4
#define NET_FILTER_PROMISCUOUS 8
	// statistics
	UINT	uRecOverflows;
	UINT	uSendOverflows;

	// Head of buffer (first allocated packet)
	UINT	uHead;
	// Tail of buffer
	UINT	uTail;
	// Packet(s) present into buffer
	UINT	bRecPacketPresent;
	// statistics
	UINT	uRecPackets;

	// Head of buffer (first allocated packet)
	UINT	uSendHead;
	// Tail of buffer
	UINT	uSendTail;
	// Packet(s) present in send buffer
	UINT	bSendPacketPresent;
	// statistics
	UINT	uSentPackets;

	// align to 1Kb boundary here
	UCHAR	pad[1024 - 48];

	// data buffers allow 1.5Kb tailroom to ease wraparound
	UCHAR	aBuffer[IO_NET_SIZE + 1536];
	UCHAR	aSendBuffer[IO_NET_SIZE + 1536];
} STRUCT_NAME(NET_BUFFER);
typedef NET_BUFFER *PNET_BUFFER;

struct AssertIoNetSize {
	char assert[IO_NET_SIZE & (IO_NET_SIZE-1) ? -1 : 1];
};
struct AssertNetBufferStructure {
	char assert[sizeof(NET_BUFFER) != 4096 + IO_NET_SIZE * 2 ? -1 : 1];
};

/////////////////////////////////////////////////////////////////////////////
//
// STRUCTURE: NET_PACKET
//
// DESCRIPTION:
//		Received Packet information
//
/////////////////////////////////////////////////////////////////////////////
typedef struct _NET_PACKET
{
	// eth header happens to be 14 bytes long
	// align IP payload on 16 byte boundary
	USHORT	uDataLen;
	// Packed data start
	UCHAR	aDataBuf[1];
} STRUCT_NAME(NET_PACKET);

struct AssertIPHeaderAlignment {
	char assert[(sizeof(NET_PACKET) - 1 + 14) & 15 ? -1 : 1];
};

// align packets to 16 byte boundary
#define FULL_PACKET_SIZE(uPacketDataLen) \
	((sizeof(NET_PACKET) - 1 + (uPacketDataLen) + 15) & ~15)


// Packet reception constants. (emulation part)
typedef enum
{
	// Buffer empty, no any packet
	NETPACK_NO_PACKET				= 0,
	// Device is not initialised yet
	NETPACK_DEV_NOT_INIT			= 0x01,
	// device not ready to receive packet (OS specific, may be overrun of buffer or smth. like it)
	NETPACK_DEV_NOT_READY			= 0x02,
	// Buffer overrun
	NETPACK_OVERFLOW				= 0x04,
	// We do not need such packet (not our)
	NETPACK_CANCELLED				= 0x08,
	// Packet ready for reception
	NETPACK_RECEIVE					= 0x10,
	// Resubmit packet
	NETPACK_RETRY					= 0x20,
	// Break reception loop (NIC buffer full)
	NETPACK_MASK_PAUSE_RX_LOOP		= NETPACK_DEV_NOT_READY,
} NETPACK_STATE;

// Packet packing restoring
#if !defined(_LINUX_) && !defined(_BSD_)
#pragma pack(pop, save_pack)
#endif

typedef struct	_SET_NET_PARAMETERS
{
	PUCHAR	pBufferAddress;
	UINT	uiSize;
	UINT	uiAdapterType;
	UCHAR	ucAddress[6];
	PVOID	pEvent;
}
NET_PARAMETERS, *PNET_PARAMETERS;

#endif // API_NET_H
// Note: gcc needs empty line after code.
