//
// $Id: AtomicOps.h,v 1.577.2.42 2006/11/22 10:16:26 integrator Exp $
//
#ifndef ATOMICOPS_H_INCLUDED
#define ATOMICOPS_H_INCLUDED

//
// all AtomicOps() return *dst value fetched before executing op
// NOTE: the caller MUST align operands properly
//
// generic case uses CMPXCHG/CMPXCHG8B instructions
// 32bit AtomicAdd/Inc/Dec() are optimized using XADD
// 32bit AtomicWrite/Swap() use XCHG
//
// no explicit memory barriers are defined here,
// however all operations behave as a strong memory barrier
// for both compiler and processor
//
// TODO: better ways to 64bit read/write atomicity may exist
//

#if defined(_MSC_VER)
// Microsoft Visual C path

#pragma warning(push)
// suppress "no return value"
// assemply inlines do return [edx:]eax per register convention
#pragma warning(disable:4035)

//
// NOTE: volatile guarantees that value is not cached by the compiler
//
#define defineAtomicRead32(name, op_ignored, argtype) \
__inline argtype \
Atomic##name(const argtype *src) \
{ \
	return *(volatile argtype *)src; \
}

//
// NOTE: __fastcall interferes with inline assembly in VC6
//
// NOTE: xchg is always locking
//
#define defineAtomicSwap32(name, op_ignored, argtype) \
__inline argtype __stdcall \
Atomic##name(argtype *dst, argtype swap) \
{ \
	__asm mov ecx, dst \
	__asm mov eax, swap \
	__asm xchg [ecx], eax \
}

#define defineAtomicCompareSwap32(name, op_ignored, argtype) \
__inline argtype __stdcall \
Atomic##name(argtype *dst, argtype compare, argtype swap) \
{ \
	__asm mov ecx, dst \
	__asm mov eax, compare \
	__asm mov edx, swap \
	__asm lock cmpxchg [ecx], edx \
}

#define defineAtomicOp32(name, op, argtype) \
__inline argtype __stdcall \
Atomic##name(argtype *dst, argtype delta) \
{ \
	__asm mov ecx, dst \
	__asm mov eax, [ecx] \
__asm retry: \
	__asm mov edx, delta \
	__asm op edx, eax \
	__asm lock cmpxchg [ecx], edx \
	__asm jnz retry \
}

#define defineAtomicAdd32(name, op_ignored, argtype) \
__inline argtype __stdcall \
Atomic##name(argtype *dst, argtype delta) \
{ \
	__asm mov ecx, dst \
	__asm mov eax, delta \
	__asm lock xadd [ecx], eax \
}

#define defineAtomicInc32(name, op_delta, argtype) \
__inline argtype __stdcall \
Atomic##name(argtype *dst) \
{ \
	__asm mov ecx, dst \
	__asm mov eax, op_delta \
	__asm lock xadd [ecx], eax \
}

#define defineAtomicDec32(name, op_delta, argtype) \
defineAtomicInc32(name, op_delta, argtype)

//
// base 64bit type
// not a typedef because we'll be using `unsigned`
// modifier with it
//
#define atomicInt64 __int64

//
// if *dst happens to be equal to edx:eax the same value
// will be swapped in breaking nothing
//
#define defineAtomicRead64asm(src) \
	__asm mov esi, src \
	__asm mov ebx, eax \
	__asm mov ecx, edx \
	__asm lock cmpxchg8b [esi]

#define defineAtomicRead64(name, op_ignored, argtype) \
__inline argtype __stdcall \
Atomic##name(const argtype *src) \
{ \
	defineAtomicRead64asm(src) \
}

#define defineAtomicSwap64(name, op_ignored, argtype) \
__inline argtype __stdcall \
Atomic##name(argtype *dst, argtype swap) \
{ \
	defineAtomicRead64asm(dst) \
	__asm mov ebx, dword ptr swap \
	__asm mov ecx, dword ptr swap+4 \
__asm retry: \
	__asm lock cmpxchg8b [esi] \
	__asm jnz retry \
}

#define defineAtomicCompareSwap64(name, op_ignored, argtype) \
__inline argtype __stdcall \
Atomic##name(argtype *dst, argtype compare, argtype swap) \
{ \
	__asm mov esi, dst \
	__asm mov eax, dword ptr compare \
	__asm mov edx, dword ptr compare+4 \
	__asm mov ebx, dword ptr swap \
	__asm mov ecx, dword ptr swap+4 \
	__asm lock cmpxchg8b [esi] \
}

#define defineAtomicOp64(name, op, argtype) \
__inline argtype __stdcall \
Atomic##name(argtype *dst, argtype delta) \
{ \
	defineAtomicRead64asm(dst) \
__asm retry: \
	__asm mov ebx, dword ptr delta \
	__asm mov ecx, dword ptr delta+4 \
	__asm op ebx, eax \
	__asm op ecx, edx \
	__asm lock cmpxchg8b [esi] \
	__asm jnz retry \
}

//
// need to handle carry
//
#define defineAtomicAdd64(name, op_ignored, argtype) \
__inline argtype __stdcall \
Atomic##name(argtype *dst, argtype delta) \
{ \
	defineAtomicRead64asm(dst) \
__asm retry: \
	__asm mov ebx, dword ptr delta \
	__asm mov ecx, dword ptr delta+4 \
	__asm add ebx, eax \
	__asm adc ecx, edx \
	__asm lock cmpxchg8b [esi] \
	__asm jnz retry \
}

#define defineAtomicInc64(name, op_ignored, argtype) \
__inline argtype __stdcall \
Atomic##name(argtype *dst) \
{ \
	defineAtomicRead64asm(dst) \
__asm retry: \
	__asm mov ebx, eax \
	__asm mov ecx, edx \
	__asm add ebx, 1 \
	__asm adc ecx, 0 \
	__asm lock cmpxchg8b [esi] \
	__asm jnz retry \
}

#define defineAtomicDec64(name, op_ignored, argtype) \
__inline argtype __stdcall \
Atomic##name(argtype *dst) \
{ \
	defineAtomicRead64asm(dst) \
__asm retry: \
	__asm mov ebx, eax \
	__asm mov ecx, edx \
	__asm sub ebx, 1 \
	__asm sbb ecx, 0 \
	__asm lock cmpxchg8b [esi] \
	__asm jnz retry \
}

#elif defined(__GNUC__)
// GCC path

//
// NOTE: volatile guarantees that value is not cached by the compiler
//
#define defineAtomicRead32(name, op, argtype) \
__inline argtype \
Atomic##name(const argtype *src) \
{ \
	return *(volatile argtype *)src; \
}

//
// NOTE: earlyclobber outputs to avoid allocation clashes
// NOTE: output "=&r"(*dst) because it is used in a loop
// NOTE: "memory" clobber to mark memory barrier for compiler
//
// NOTE: xchg is always locking
//
#define defineAtomicSwap32(name, op, argtype) \
__inline argtype \
Atomic##name(argtype *dst, argtype swap) \
{ \
	argtype oldval; \
\
	__asm__ __volatile__ ( \
		"xchg %2, %1\n\t" \
		: "=a"(oldval) \
		: "m"(*dst), "r"(swap) \
		: "memory"); \
	return oldval; \
}

#define defineAtomicCompareSwap32(name, op, argtype) \
__inline argtype \
Atomic##name(argtype *dst, argtype compare, argtype swap) \
{ \
	argtype oldval; \
\
	__asm__ __volatile__ ( \
		"lock cmpxchg %3, %2\n\t" \
		: "=a"(oldval) \
		: "0"(compare), "m"(*dst), "r"(swap) \
		: "memory"); \
	return oldval; \
}

#define defineAtomicOp32(name, op, argtype) \
__inline argtype \
Atomic##name(argtype *dst, argtype delta) \
{ \
	argtype oldval, newval; \
\
	__asm__ __volatile__ ( \
		"mov %2, %0\n" \
	"1:\n\t" \
		"mov %3, %1\n\t" \
		#op " %0, %1\n\t" \
		"lock cmpxchg %1, %2\n\t" \
		"jnz 1b\n\t" \
		: "=&a"(oldval), "=&r"(newval) \
		: "m"(*dst), "g"(delta) \
		: "memory"); \
	return oldval; \
}

#define defineAtomicAdd32(name, op, argtype) \
__inline argtype \
Atomic##name(argtype *dst, argtype delta) \
{ \
	argtype oldval; \
\
	__asm__ __volatile__ ( \
		"lock xadd %0, %2\n\t" \
		: "=a"(oldval) \
		: "0"(delta), "m"(*dst) \
		: "memory"); \
	return oldval; \
}

#define defineAtomicInc32(name, op, argtype) \
__inline argtype \
Atomic##name(argtype *dst) \
{ \
	argtype oldval; \
\
	__asm__ __volatile__ ( \
		"lock xadd %0, %2\n\t" \
		: "=a"(oldval) \
		: "0"(op), "m"(*dst) \
		: "memory"); \
	return oldval; \
}

#define defineAtomicDec32(name, op_delta, argtype) \
defineAtomicInc32(name, op_delta, argtype)

//
// base 64bit type
// not a typedef because we'll be using `unsigned`
// modifier with it
//
#define atomicInt64 long long

//
// if *dst happens to be equal to edx:eax the same value
// will be swapped in breaking nothing
//
#define defineAtomicRead64asm(src, val) \
	__asm__ __volatile__ ( \
		"movl %%eax, %%ebx\n\t" \
		"movl %%edx, %%ecx\n\t" \
		"lock cmpxchg8b %1\n\t" \
		: "=&A"(val) \
		: "m"(*src) \
		: "ebx", "ecx")

#define defineAtomicRead64(name, op, argtype) \
__inline argtype \
Atomic##name(const argtype *src) \
{ \
	argtype val; \
\
	defineAtomicRead64asm(src, val); \
	__asm__ __volatile__ ("" : : : "memory"); \
	return val; \
}

#define defineAtomicSwap64(name, op, argtype) \
__inline argtype \
Atomic##name(argtype *dst, argtype swap) \
{ \
	argtype oldval; \
\
	defineAtomicRead64asm(dst, oldval); \
	__asm__ __volatile__ ( \
	"1:\n\t" \
		"lock cmpxchg8b %2\n\t" \
		"jnz 1b\n\t" \
		: "=&A"(oldval) \
		: "0"(oldval), "m"(*dst), "b"((unsigned)swap), "c"((unsigned)(swap >> 32)) \
		: "memory"); \
	return oldval; \
}

#define defineAtomicCompareSwap64(name, op_ignored, argtype) \
__inline argtype \
Atomic##name(argtype *dst, argtype compare, argtype swap) \
{ \
	argtype oldval; \
\
	__asm__ __volatile__ ( \
		"lock cmpxchg8b %2\n\t" \
		: "=A"(oldval)  \
		: "0"(compare), "m"(*dst), "b"((unsigned)swap), "c"((unsigned)(swap >> 32)) \
		: "memory"); \
	return oldval; \
}

#define defineAtomicOp64(name, op, argtype) \
__inline argtype \
Atomic##name(argtype *dst, argtype delta) \
{ \
	argtype oldval; \
\
	defineAtomicRead64asm(dst, oldval); \
	__asm__ __volatile__ ( \
	"1:\n\t" \
		"movl %3, %%ebx\n\t" \
		"movl %4, %%ecx\n\t" \
		#op " %%eax, %%ebx\n\t" \
		#op " %%edx, %%ecx\n\t" \
		"lock cmpxchg8b %2\n\t" \
		"jnz 1b\n\t" \
		: "=&A"(oldval) \
		: "0"(oldval), "m"(*dst), "g"((unsigned)delta), "g"((unsigned)(delta >> 32)) \
		: "ebx", "ecx", "memory"); \
	return oldval; \
}

#define defineAtomicAdd64(name, op_ignored, argtype) \
__inline argtype \
Atomic##name(argtype *dst, argtype delta) \
{ \
	argtype oldval; \
\
	defineAtomicRead64asm(dst, oldval); \
	__asm__ __volatile__ ( \
	"1:\n\t" \
		"movl %3, %%ebx\n\t" \
		"movl %4, %%ecx\n\t" \
		"addl %%eax, %%ebx\n\t" \
		"adcl %%edx, %%ecx\n\t" \
		"lock cmpxchg8b %2\n\t" \
		"jnz 1b\n\t" \
		: "=&A"(oldval) \
		: "0"(oldval), "m"(*dst), "g"((unsigned)delta), "g"((unsigned)(delta >> 32)) \
		: "ebx", "ecx", "memory"); \
	return oldval; \
}

#define defineAtomicInc64(name, op_ignored, argtype) \
__inline argtype \
Atomic##name(argtype *dst) \
{ \
	argtype oldval; \
\
	defineAtomicRead64asm(dst, oldval); \
	__asm__ __volatile__ ( \
	"1:\n\t" \
		"movl %%eax, %%ebx\n\t" \
		"movl %%edx, %%ecx\n\t" \
		"addl $1, %%ebx\n\t" \
		"adcl $0, %%ecx\n\t" \
		"lock cmpxchg8b %2\n\t" \
		"jnz 1b\n\t" \
		: "=&A"(oldval) \
		: "0"(oldval), "m"(*dst) \
		: "ebx", "ecx", "memory"); \
	return oldval; \
}

#define defineAtomicDec64(name, op_ignored, argtype) \
__inline argtype \
Atomic##name(argtype *dst) \
{ \
	argtype oldval; \
\
	defineAtomicRead64asm(dst, oldval); \
	__asm__ __volatile__ ( \
	"1:\n\t" \
		"movl %%eax, %%ebx\n\t" \
		"movl %%edx, %%ecx\n\t" \
		"subl $1, %%ebx\n\t" \
		"sbbl $0, %%ecx\n\t" \
		"lock cmpxchg8b %2\n\t" \
		"jnz 1b\n\t" \
		: "=&A"(oldval) \
		: "0"(oldval), "m"(*dst) \
		: "ebx", "ecx", "memory"); \
	return oldval; \
}

#endif // __GNUC__


struct AssertAtomic32bitArchitecture {
	char assert[
		sizeof(int) == 4 && sizeof(unsigned) == 4
			&& sizeof(long) == 4 && sizeof(unsigned long) == 4 ? 1 : -1];
};

#ifdef __cplusplus
extern "C" {
#endif

//
// NOTE: AtomicWrite() is implemented as swap
//
#define defineAtomicFamily(bits, argtype, suffix) \
defineAtomicRead##bits(Read##suffix, mov, argtype) \
defineAtomicSwap##bits(Write##suffix, mov, argtype) \
defineAtomicSwap##bits(Swap##suffix, xchg, argtype) \
defineAtomicCompareSwap##bits(CompareSwap##suffix, cmpxchg, argtype) \
defineAtomicOp##bits(And##suffix, and, argtype) \
defineAtomicOp##bits(Or##suffix, or, argtype) \
defineAtomicOp##bits(Xor##suffix, xor, argtype) \
defineAtomicAdd##bits(Add##suffix, xadd, argtype) \
defineAtomicInc##bits(Inc##suffix, +1, argtype) \
defineAtomicDec##bits(Dec##suffix, -1, argtype)

#define defineAtomicFamily2(bits, argtype) \
defineAtomicRead##bits(Read, mov, argtype) \
defineAtomicSwap##bits(Write, mov, argtype) \
defineAtomicSwap##bits(Swap, xchg, argtype) \
defineAtomicCompareSwap##bits(CompareSwap, cmpxchg, argtype) \
defineAtomicOp##bits(And, and, argtype) \
defineAtomicOp##bits(Or, or, argtype) \
defineAtomicOp##bits(Xor, xor, argtype) \
defineAtomicAdd##bits(Add, xadd, argtype) \
defineAtomicInc##bits(Inc, +1, argtype) \
defineAtomicDec##bits(Dec, -1, argtype)


defineAtomicFamily2(32, int)
defineAtomicFamily(32, unsigned, U)
defineAtomicFamily(32, long, L)
defineAtomicFamily(32, unsigned long, Ul)
//
// limited set of operations for `void *`
//
defineAtomicSwap32(SwapPv, xchg, void *)
defineAtomicCompareSwap32(CompareSwapPv, cmpxchg, void *)


defineAtomicFamily(64, atomicInt64, 64)
defineAtomicFamily(64, unsigned atomicInt64, 64U)

#ifdef __cplusplus
} // extern "C"

//
// C++ overloads for types other than int
//
defineAtomicFamily2(32, unsigned)
defineAtomicFamily2(32, long)
defineAtomicFamily2(32, unsigned long)

//
// long modifier can not be used with 64bit types
//
defineAtomicFamily(64, unsigned atomicInt64, 64)

//
// object pointer templated ops
//
template<class T>
defineAtomicRead32(Read, mov, T *)
template<class T>
defineAtomicSwap32(Write, mov, T *)
template<class T>
defineAtomicSwap32(Swap, xchg, T *)
template<class T>
defineAtomicCompareSwap32(CompareSwap, cmpxchg, T *)

template<class T>
__inline T*
AtomicAdd(T **dst, int delta)
{
	return reinterpret_cast<T*>(AtomicAdd(
		reinterpret_cast<int *>(dst), delta * sizeof(**dst)));
}
template<class T>
__inline T*
AtomicInc(T **dst)
{
	return reinterpret_cast<T*>(AtomicAdd(
		reinterpret_cast<int *>(dst), sizeof(**dst)));
}
template<class T>
__inline T*
AtomicDec(T **dst)
{
	return reinterpret_cast<T*>(AtomicAdd(
		reinterpret_cast<int *>(dst), -sizeof(**dst)));
}

#endif // __cplusplus


#if defined(_MSC_VER)
#pragma warning(pop)
#endif

#endif // ATOMICOPS_H_INCLUDED
