/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2005 Parallels Inc. All Rights Reserved.
//		 http://www.parallelssoft.com
//
// MODULE: Spinlock.h
//
// DESCRIPTION:
//              Syncronisation object (spinlock) definition
//
/////////////////////////////////////////////////////////////////////////////
#ifndef __SPINLOCK_H__
#define __SPINLOCK_H__
#undef NEW_SPINLOCKS
/////////////////////////////////////////////////////////////////////////////
// Main idea:
// 
// We need to syncronise some operations with variables into count of 
// threads and monitor. Also we need support syncronisation onto SMP
// systems. So idea is to access SPINLOCK variable with #LOCK prefix.
/////////////////////////////////////////////////////////////////////////////

// Define types
#if (defined(_LINUX_) || defined(_BSD_)) && defined(_HOSTPART_) && !defined(MONITOR_EXPORTS)
 #ifdef NEW_SPINLOCKS
  #include "spQueue.h"
  #include "MCDLog.h"
   typedef unsigned int SPINLOCK[2];

   struct spinlock_data {
     unsigned int qlock;
     pthread_t owner;
     sp_qhead queue;
   };
 #endif
typedef unsigned int SPINLOCK;
#else
typedef unsigned int SPINLOCK;
#endif

typedef SPINLOCK *PSPINLOCK;

// Locked
#define SPINLOCK_LOCKED         1
// Unlocked
#define SPINLOCK_UNLOCKED	0

// Need to define switching function
extern void	SendHaltRequest(unsigned int uiTimeOutMs);

#if (defined(_LINUX_) || defined(_BSD_) || defined(_MACOS_)) && !defined(MONITOR_EXPORTS)
/*
 * --------------------------------------------------------------------------
 *
 * FUNCTION: lowLock
 *
 * DESCRIPTION: lock variable
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 * --------------------------------------------------------------------------
 */
static __inline unsigned int lowLock(unsigned int *lock) {

    int ret;

    __asm__
        __volatile__ (
                      "lock cmpxchgl %3, %1"
                      // output operands
                      : "=a" (ret)
                      // input operands
                      : "m" (*lock), "0" (SPINLOCK_UNLOCKED), "r" (SPINLOCK_LOCKED)
                     );

    return ret;
}

/*
 * --------------------------------------------------------------------------
 *
 * FUNCTION: lowUnlock
 *
 * DESCRIPTION: unlock variable
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 * --------------------------------------------------------------------------
 */
static __inline void lowUnlock(unsigned int *lock) {

    __asm__
        __volatile__  (
                       "lock xchgl %1, %0"
                       // output operands
                       :
                       // input operands
                       : "m" (*lock), "a" (SPINLOCK_UNLOCKED)
                      );
}
#ifdef NEW_SPINLOCKS
/*
 * --------------------------------------------------------------------------
 *
 * FUNCTION: initLock
 *
 * DESCRIPTION: initialize lock variable
 *
 * PARAMETERS:
 *
 * RETURN VALUE:
 *
 * --------------------------------------------------------------------------
 */
static __inline void initLock(PSPINLOCK *lock) {

    struct spinlock_data *data;

    *lock = 0;
    data = (struct spinlock_data *)lock + sizeof(unsigned int);
    data->qlock = 0;
    data->owner = 0;
    data->queue.count = 0;
    data->queue.head = NULL;
    data->queue.head = NULL;

}
#endif
#endif
/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: TryLock()
//
// DESCRIPTION:
//		Try to lock SPINLOCK value
//
// PARAMETERS:
//		pSL    [in] - Pointer to allocated spinlock
//
// RETURN VALUE:
//
// COMMENTS:
//
// DEVELOPER:
//		AntonZ
//
/////////////////////////////////////////////////////////////////////////////
static __inline unsigned int TryLock(PSPINLOCK pSL)
{
	unsigned int uiRetVal;

#if defined _MSC_VER
	__asm 
	{
		mov edx, pSL
		mov eax, SPINLOCK_UNLOCKED
		mov ebx, SPINLOCK_LOCKED
		lock cmpxchg [edx], ebx
		mov uiRetVal, eax
	}
#elif defined __GNUC__
 #ifdef NEW_SPINLOCKS
        struct spinlock_data *data;

        if ((int)(uiRetVal = lowLock(*pSL)) == SPINLOCK_LOCKED) {
            printf("Locked by: %lu\n", pthread_self());
            data = (struct spinlock_data *)pSL + sizeof(unsigned int);
            lowLock(&data->qlock);
            sp_queue_get(&data->queue); /* remove */
            data->owner = pthread_self();
            lowUnlock(&data->qlock);
        }
 #else
        uiRetVal = lowLock(pSL);
 #endif
#endif

	return uiRetVal;
}

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: Lock()
//
// DESCRIPTION:
//		Infinite wait of SPINLOCK value to be locked
//
// PARAMETERS:
//		pSL    [in] - Pointer to allocated spinlock
//
// RETURN VALUE:
//
// COMMENTS:
//
// DEVELOPER:
//		AntonZ
//
/////////////////////////////////////////////////////////////////////////////
static __inline unsigned int Lock(PSPINLOCK pSL)
{
	while (TryLock(pSL) == SPINLOCK_LOCKED)
	{
#if defined _MSC_VER
  #if defined _AFXDLL
		// Application
		Sleep(1);
  #endif
  #if defined MONITOR_EXPORTS
		// I think we do not need call wait cycle from monitor.
		//SendHaltRequest(1);
  #endif
#elif defined __GNUC__
  #if defined(_HOSTPART_) && defined(_LINUX_)
   #ifdef NEW_SPINLOCKS
        sigset_t mask;
        int sig;
        struct spinlock_data *data;

        printf("Unlocking wait: %lu\n", pthread_self());

        data = (struct spinlock_data *)pSL + sizeof(unsigned int);
        lowLock(&data->qlock);
        if (data->owner == pthread_self()) {
            LOG_TIME_MESSAGE(DBG_FATAL, "Deadlock in %lu thread !", pthread_self());
        } else {
            sp_queue_add(&data->queue, pthread_self()); /* remove */
        }
        lowUnlock(&data->qlock);

        pthread_sigmask(SIG_SETMASK, NULL, &mask);
        //sigemptyset(&mask);
        sigaddset(&mask, SIGUSR1);
        pthread_sigmask(SIG_BLOCK, NULL, &mask);
        printf("2 Unlocking wait: %lu\n", pthread_self());
        while(sig != SIGUSR1) {
            sigwait(&mask, &sig);
            printf("Signalled: %d\n", sig);
            switch(sig) {
            case SIGUSR1:
                printf("Wakeup received in: %lu\n", pthread_self());
                break;
            default:
                raise(sig); /* raise signal for default action */
                break;
            }
        }
   #endif
        sched_yield();
  #endif
#endif
	}

	return SPINLOCK_LOCKED;
}

/////////////////////////////////////////////////////////////////////////////
//
// FUNCTION: Unlock()
//
// DESCRIPTION:
//		Unlock SPINLOCK value
//
// PARAMETERS:
//		pSL    [in] - Pointer to allocated spinlock
//
// RETURN VALUE:
//
// COMMENTS:
//
// DEVELOPER:
//		AntonZ
//
/////////////////////////////////////////////////////////////////////////////
static __inline void Unlock(PSPINLOCK pSL)
{
#if defined _MSC_VER
	__asm
	{
		mov edx, pSL
		mov eax, SPINLOCK_UNLOCKED
		lock xchg [edx], eax
	}
#elif defined __GNUC__
 #ifdef NEW_SPINLOCKS
        pthread_t id;
        struct spinlock_data *data;

        data = (struct spinlock_data *)pSL + sizeof(unsigned int);
        lowLock(&data->qlock);
        id = sp_queue_peek(&data->queue);
        if (id > 0) {
            pthread_kill(id, SIGUSR1);
            printf("Unlock and wakeup: %lu\n", id);
        }
        lowUnlock(&data->qlock);
 #endif
        lowUnlock(pSL);
#endif
}

#endif

